/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTHTMLUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTDOMUtil"];

/**
 * DOM 操作関係ユーティリティ関数
 */
let UnMHTDOMUtil = {
  /**
   * ノードが選択範囲内かどうかチェックする
   *
   * @param  HTMLElement node
   *         対象のノード
   * @param  Selection selection
   *         対象の選択範囲
   * @return Boolean
   *         範囲内かどうか
   */
  isNodeInSelection : function (node, selection) {
    for (let i = 0; i < selection.rangeCount; i ++) {
      let range = selection.getRangeAt (i);
      
      let nodeRange = node.ownerDocument.createRange ();
      let w = node.ownerDocument.defaultView;
      nodeRange.selectNode (node);
      
      let compareSS = 0;
      let compareSE = 0;
      let compareES = 0;
      let compareEE = 0;
      try {
        compareSS
          = range.compareBoundaryPoints (w.Range.START_TO_START, nodeRange);
        compareSE
          = range.compareBoundaryPoints (w.Range.START_TO_END, nodeRange);
        compareES
          = range.compareBoundaryPoints (w.Range.END_TO_START, nodeRange);
        compareEE
          = range.compareBoundaryPoints (w.Range.END_TO_END, nodeRange);
      }
      catch (e) {
        /* 選択範囲が別のドキュメント */
        return false;
      }
        
      if (!((compareSS == -1
             && compareSE == -1
             && compareES == -1
             && compareEE == -1)
            || (compareSS == 1
                && compareSE == 1
                && compareES == 1
                && compareEE == 1))) {
        return true;
      }
    }
    
    return false;
  },
  
  /**
   * HTML ツリーを文字列に変換する
   *
   * @param  String location
   *         ドキュメントの URI
   * @param  HTMLElement node
   *         対象の要素
   * @param  nsISelection targetSelection
   *         対象の選択範囲もしくは null
   * @return String
   *         要素全体の HTML
   */
  convertToPlainText : function (location, node, targetSelection) {
    let text = "";
    
    if (!node) {
      return "";
    }
    if (!node.nodeName) {
      return "";
    }
    
    if (targetSelection) {
      if (node.nodeName.toLowerCase () == "head") {
        targetSelection = null;
      }
    }
    
    if (targetSelection) {
      if (!UnMHTDOMUtil.isNodeInSelection (node, targetSelection)) {
        return "";
      }
    }
    
    let nodeName = node.nodeName.toLowerCase ();
    if (nodeName == "#text") {
      text = UnMHTHTMLUtil.escapeEntity (node.nodeValue);
      
      if (targetSelection) {
        let textList = [];
        for (let i = 0; i < targetSelection.rangeCount; i ++) {
          let t = text;
          let isPartial = false;
          
          let range = targetSelection.getRangeAt (i);
          
          if (range.endContainer == node) {
            try {
              t = t.substr (0, range.endOffset);
              isPartial = true;
            }
            catch (e) {
            }
          }
          
          if (range.startContainer == node) {
            try {
              t = t.substr (range.startOffset);
              isPartial = true;
            }
            catch (e) {
            }
          }
          
          if (isPartial) {
            textList.push (t);
          }
        }
        
        if (textList.length > 0) {
          text = textList.join ("");
        }
      }
    }
    else if (nodeName == "#comment") {
      /* コメントをコメントとして展開する */
      text = "<!--" + node.nodeValue + "-->";
    }
    else if (nodeName == "base"
             && UnMHTAppInfo.isThunderbird ()) {
    }
    else {
      let attributes = "";
      
      if (node.attributes) {
        Array.forEach
          (node.attributes,
           function (attr) {
            let name = attr.name.toLowerCase ();
            let value = attr.value;
          
            if (name.match (/^__unmht_/)) {
              return;
            }
                        
            if (nodeName == "option") {
              if (name == "selected") {
                return;
              }
            }
          
            if (nodeName == "input") {
              if (node.type == "checkbox"
                  || node.type == "radio") {
                if (name == "checked") {
                  return;
                }
              }
              else {
                if (name == "value") {
                  return;
                }
              }
            }
            
            if ((nodeName == "frame"
                 || nodeName == "iframe")
                && name == "src") {
              value = node.contentDocument.location.href;
            }
            
            if ((name == "src" || name == "href")
                && value.match
                (/^(imap|mailbox|news|nntp|pop|snews):\/\//)) {
              value
                = value
                .replace (/^(imap|mailbox|news|nntp|pop|snews):\/\/[^\/]*\//, "")
                .replace (location
                          .replace (/^(imap|mailbox|news|nntp|pop|snews):\/\/[^\/]*\//, ""),
                          "file:///mail/attachment/");
              if (value == "file:///mail/attachment/") {
                value = "file:///mail/index.html";
              }
            }
                        
            value = UnMHTHTMLUtil.escapeEntity (value);
                        
            attributes += " " + name +  "=" + "\"" + value + "\"";
          });
      }
      
      if (nodeName == "option") {
        if (node.selected) {
          let name = "selected";
          let value = "selected";
          attributes += " " + name +  "=" + "\"" + value + "\"";
        }
      }
        
      if (nodeName == "input") {
        if (node.type == "checkbox"
            || node.type == "radio") {
          if (node.checked) {
            let name = "checked";
            let value = "checked";
            attributes += " " + name +  "=" + "\"" + value + "\"";
          }
        }
        else if (node.type == "file") {
        }
        else {
          let name = "value";
          let value = UnMHTHTMLUtil.escapeEntity (node.value);
          attributes += " " + name +  "=" + "\"" + value + "\"";
        }
      }
      
      text += "<" + nodeName + attributes + ">";
                
      if (nodeName == "noscript"
          && node.firstChild
          && node.firstChild.nodeName.toLowerCase () == "#text"
          && node.firstChild.nextSibling == null) {
        /* NOSCRIPT 要素が展開されていない場合 */
        text += node.firstChild.nodeValue;
      }
      else if (nodeName == "noembed"
               && node.firstChild
               && node.firstChild.nodeName.toLowerCase () == "#text"
               && node.firstChild.nextSibling == null) {
        /* NOEMBED 要素が展開されていない場合 */
        text += node.firstChild.nodeValue;
      }
      else if (nodeName == "style"
               || nodeName == "script") {
        let node2 = node.firstChild;
        while (node2) {
          text
            += UnMHTHTMLUtil.unescapeEntity
            (this.convertToPlainText (location, node2, targetSelection));
                        
          node2 = node2.nextSibling;
        }
      }
      else if (nodeName == "textarea") {
        text += UnMHTHTMLUtil.escapeEntity (node.value);
      }
      else {
        let node2 = node.firstChild;
        while (node2) {
          text
            += this.convertToPlainText (location, node2, targetSelection);
                        
          node2 = node2.nextSibling;
        }
      }
                
      if (nodeName != "br"
          && nodeName != "img"
          && nodeName != "imput"
          && nodeName != "hr"
          && nodeName != "meta") {
        text += "</" + nodeName + ">";
      }
    }
    
    return text;
  },
  
  /**
   * ノードを親方向に探す
   * 対象のノードも含む
   *
   * @param  XULElement/HTMLElement node
   *         対象のノード
   * @param  String nodeName
   *         探すノード名
   * @return XULElement/HTMLElement
   *         見付かったノード
   *         見付からなかった場合 null
   */
  findParentNode : function (node, nodeName) {
    while (node) {
      if (node.nodeName.toLowerCase () == nodeName) {
        return node;
      }
      node = node.parentNode;
    }
    
    return null;
  },
  
  /**
   * テキストが選択されているかどうかを返す
   *
   * @param  Window targetWindow
   *         対象のウィンドウ
   * @param  Boolean usePopupNode
   *         コンテキストメニューから情報を取得するか
   * @return Boolean
   *         テキストが選択されているかどうか
   */
  isTextSelected : function (targetWindow, usePopupNode) {
    if (usePopupNode && "gContextMenu" in targetWindow) {
      return targetWindow.gContextMenu.isTextSelected;
    }
    else {
      let targetDocument = null;
      if (usePopupNode && targetWindow.document.popupNode) {
        targetDocument = targetWindow.document.popupNode.ownerDocument;
      }
      else {
        targetDocument = UnMHTWindowUtil.getCurrentDocument (targetWindow);
      }
      
      let targetSelection = targetDocument.defaultView.getSelection ();
      for (let i = 0; i < targetSelection.rangeCount; i ++) {
        let range = targetSelection.getRangeAt (i);
        if (range.startContainer != range.endContainer
            || range.startOffset != range.endOffset) {
          return true;
        }
      }
      
      return false;
    }
  }
};
