/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

/**
 * Defines a Document Loader Factory to handle the MHT document type.
 *
 * Currently, this implementation just fires the standard MHT loading process,
 *  and refuses to create a content viewer and a stream listener.
 *
 * For more information on the document loading process, see
 *  <http://www.mozilla.org/newlayout/doc/webwidget.html> and
 *  <https://developer.mozilla.org/En/The_life_of_an_HTML_HTTP_request>
 *  (retrieved 2008-10-07).
 */

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");

let EXPORTED_SYMBOLS = ["UnMHTDocumentLoaderFactory"];

/**
 * Construct the UnMHTDocumentLoaderFactory object.
 *
 * Document loader factories are XPCOM services, so they should be constructed
 *  as singletons, for example as explained in
 *  <http://weblogs.mozillazine.org/weirdal/archives/019620.html> (retrieved
 *  2008-10-07). However, since this object doesn't store any state information,
 *  we might as well create it as a normal XPCOM object.
 *
 * Document loader factories must be registered with the
 *  "Gecko-Content-Viewers" category. For this extension, this is done
 *  dynamically on startup, because the list of MIME types that this document
 *  loader factory will handle is not known in advance.
 */
function UnMHTDocumentLoaderFactory () {
}
UnMHTDocumentLoaderFactory.prototype = {
  classID:          Components.ID ("{fa22337c-baf4-4c05-ab92-5e1fb4a97aa9}"),
  contractID:       "@unmht.org/unmht-document-loader-factory;1",
  classDescription: "UnMHT Document Loader Factory JS Component",
  QueryInterface:   XPCOMUtils.generateQI ([Ci.nsIDocumentLoaderFactory]),
  
  /* ==== nsIDocumentLoaderFactory ==== */
  
  /**
   * リフレッシュページのコンテントビューアを作成する
   *
   * @param  String aCommand
   *         実行するコマンド
   * @param  nsIChannel aChannel
   *         入力チャネル
   * @param  nsILoadGroup aLoadGroup
   *         不明
   * @param  String aContentType
   *         表示するコンテンツの MIME-Type
   * @param  nsISupport aContainer
   *         不明
   * @param  nsISupport aExtraInfo
   *         不明
   * @param  Object aDocListenerResult
   *         入力チャネルのストリームのリスナ (出力)
   *         {
   *           value: nsIStreamListener
   *         }
   * @return nsIContentViewer
   *         リフレッシュページのコンテントビューア
   */
  createInstance: function (aCommand, aChannel, aLoadGroup, aContentType,
                            aContainer, aExtraInfo, aDocListenerResult) {
    if (aCommand != "view") {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }
    
    if (aChannel.URI.spec.match (/^(imap|mailbox|news|nntp|pop|snews):\/\//)) {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }
    
    let refreshChannel
    = Services.io.newChannelFromURI
    (Services.io.newURI
     (UnMHTScheme.enRefreshURI
      (UnMHTScheme.enUnMHTURI (aChannel.URI.spec)), "utf-8", null));
    
    let listener = {};
    
    let viewer
    = Cc ["@mozilla.org/content/document-loader-factory;1"]
    .getService (Ci.nsIDocumentLoaderFactory)
    .createInstance ("view", refreshChannel, aLoadGroup, "text/html",
                     aContainer, aExtraInfo, listener);
    
    refreshChannel.asyncOpen
    (listener.value.QueryInterface (Ci.nsIStreamListener), null);
    
    /* Return the content viewer, which is already receiving data from our
     *  channel, and a dummy stream listener, to throw away data from the
     *  original channel (aChannel).
     * ==== FIXME: aChannel からデータを取得しないと
     *             サーバへのリクエストが二重になる可能性がある ==== */
    aDocListenerResult.value = {
      QueryInterface: XPCOMUtils.generateQI ([Ci.nsIRequestObserver,
                                              Ci.nsIStreamListener]),
  
      onStartRequest: function () {},
      onStopRequest: function () {},
      onDataAvailable: function () {}
    };
    
    return viewer;
  },
  
  /**
   * ドキュメントに対応するコンテントビューアの作成
   *
   * @param  nsISupport aContainer
   *         不明
   * @param  nsIDocument aDocment
   *         対象のドキュメント
   * @param  String aCommand
   *         実行するコマンド
   */
  createInstanceForDocument: function (aContainer, aDocument, aCommand) {
    /* This function should never have been called. As of 2008-10-07, it may
     *  be called only from nsDocShell::CreateAboutBlankContentViewer, that
     *  doesn't handle exceptions, so we return null instead of throwing
     *  NS_ERROR_NOT_IMPLEMENTED. */
    return null;
  },
  
  /**
   * 空のドキュメントの作成
   *
   * @param  nsILoadGroup aLoadGroup
   *         不明
   * @param  nsIPrincipal aPrincipal
   *         不明 (オプショナル)
   */
  createBlankDocument: function (aLoadGroup, aPrincipal) {
    /* See the comment inside createInstanceForDocument. */
    return null;
  }
};
