/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractor.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTShutdown.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");
Cu.import ("resource://unmht/modules/UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTExtractManager",
                        "UnMHTExtractHook"];

let PERM_IMAGE_ALLOW_ALL = 1;
let PERM_IMAGE_BLOCK_ALL = 2;
let PERM_IMAGE_BLOCK_THIRD_PIRTY = 3;

/**
 * 展開用フック
 *
 * @param  Window window
 *         対象のウィンドウ
 */
function UnMHTExtractHook (window) {
  this.window = window;
  
  this.onDOMContentLoaded = this.onDOMContentLoaded.bind (this);
}
UnMHTExtractHook.prototype = {
  window : null,
  tabbrowser : null,
  appcontent : null,
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    this.tabbrowser = null;
    this.appcontent = null;
  },
  
  /**
   * 監視する
   */
  onLoad : function () {
    if (!UnMHTAppInfo.isThunderbird ()) {
      this.tabbrowser = this.window.document.getElementById ("content");
      if (!this.tabbrowser) {
        return;
      }
      
      this.appcontent = this.window.document.getElementById ("appcontent");
      this.appcontent.addEventListener
      ("DOMContentLoaded", this.onDOMContentLoaded, false);
    }
  },
  
  /**
   * 監視をやめる
   */
  onUnload : function () {
    if (!UnMHTAppInfo.isThunderbird ()) {
      this.appcontent.removeEventListener
      ("DOMContentLoaded", this.onDOMContentLoaded);
    }
  },
  
  /**
   * コンテンツがロードされたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onDOMContentLoaded : function (event) {
    UnMHTExtractManager.onDOMContentLoaded (this.tabbrowser, event);
  }
};

/**
 * 展開の管理
 */
let UnMHTExtractManager = {
  extractors : new Object (),    /* Obejct  展開の情報を保持しておく
                                  * <String 正規化した展開したファイル名,
                                  *  UnMHTExtractor 展開の情報> */
  
  openMultiple : "",            /* String  複数のタブを含む MHT を開く時
                                 *   tabs: 全てのページをタブとして開く
                                 *   index: インデックスページを表示する
                                 *   first: 最初のタブを開く */
  
  loadImageType : PERM_IMAGE_ALLOW_ALL, /* Number  Firefox の画像読み込み設定 */
  
  enableOffline : false, /* Boolean  オフラインモード */
  
  /**
   * 初期化処理
   */
  init : function () {
    let f;
    
    f = function () {
      this.enableOffline = UnMHTPref.getPref ("open.offline");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "open.offline", f, this);
    f ();
    
    f = function () {
      this.openMultiple = UnMHTPref.getPref ("open.multiple.action");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "open.multiple.action", f, this);
    f ();
    
    f = function () {
      this.loadImageType
      = UnMHTPref.prefBranch.getIntPref ("permissions.default.image");
    }.bind (this);
    UnMHTPref.addListener ("permissions.default.image", f, this);
    f ();
  },
  
  /**
   * 終了処理
   */
  term : function () {
    UnMHTPref.removeAllListeners (this);
    
    this.clearCache (0);
  },
  
  /**
   * ファイル展開器を設定する
   *
   * @param  String uri
   *         unmht: 形式のファイルの URI
   * @param  UnMHTExtractor extractor
   *         ファイル展開器
   */
  addExtractor : function (uri, extractor) {
    this.extractors [UnMHTExtractUtil.normalizeName (uri)] = extractor;
  },
  
  /**
   * ファイル展開器を取得する
   *
   * @param  String uri
   *         unmht: 形式のファイルの URI
   * @return UnMHTExtractor
   *         ファイル展開器
   *         見付からなければ null
   */
  getExtractor : function (uri) {
    uri = UnMHTExtractUtil.normalizeName (uri);
    if (uri in this.extractors) {
        return this.extractors [uri];
    }
    return null;
  },
  
  /**
   * 古いファイル展開器を削除する
   *
   * @param  Number expire
   *         有効期限 [ms]
   */
  removeOldExtractor : function (expire) {
    let now = (new Date ()).getTime ();
    for (let uri in this.extractors) {
      let extractor = this.extractors [uri];
      
      if (now < extractor.lastAccessTime + expire) {
        continue;
      }
      
      extractor.destruct ();
      
      delete this.extractors [uri];
    }
  },
  
  /**
   * 展開したパートの情報を取得する
   *
   * @param  String uri
   *         unmht: 形式のファイルの URI
   * @return Array
   *         [UnMHTExtractor, UnMHTMIMEPart]
   *         パートの情報
   *         見付からなかった場合はそれぞれ null
   */
  getExtractorAndPart : function (uri) {
    if (!UnMHTScheme.isUnMHTURI (uri)) {
      return [null, null];
    }
    
    let part = null;
    
    let param = UnMHTScheme.getUnMHTURIParam (uri);
    let baseURI = param.baseURI;
    let refName = param.refName;
    
    let extractor = this.getExtractor (baseURI);
    if (extractor) {
      extractor.lastAccessTime = (new Date ()).getTime ();
      
      if (refName) {
        if (refName in extractor.refParts) {
          part = extractor.refParts [refName];
        }
      }
      else {
        part = extractor.startPart;
      }
    }      
    
    return [extractor, part];
  },

  /**
   * unmht 形式の URI からの相対パスを展開する
   *
   * @param  String uri
   *         unmht: 形式のファイルの URI
   * @param  String spec
   *         相対パス
   * @return String
   *         ファイルの URI
   *         見付からなかった場合は null
   */
  resolveURI : function (uri, spec) {
    if (!spec) {
      return uri;
    }
    
    let [extractor, part] = this.getExtractorAndPart (uri);
    if (extractor && part) {
      extractor.lastAccessTime = (new Date ()).getTime ();
      
      let [targetPart, fragment]
        = extractor.checkPart (part, spec, false);
      if (targetPart) {
        return extractor.baseURI + targetPart.eParam.refName + fragment;
      }
      
      if (!part.eParam.location.match (/^file/)) {
        try {
          return Services.io.newURI (part.eParam.location, null, null)
            .resolve (spec);
        }
        catch (e) {
        }
      }
    }
  
    return null;
  },
  
  /**
   * mht から展開したファイルの URI を取得する
   * 他の拡張から呼び出される
   *
   * @param  nsIURI contentLocation
   *         対象の URI
   * @param  nsIURI requestOrigin
   *         呼び出し元の URI
   *         アドレスに UnMHT と入っているもの
   * @return nsIURI
   *         展開したファイルの URI
   *         存在しなかった場合は null
   */
  getMHTFileURI : function (contentLocation, requestOrigin) {
    let [extractor, part] = this.getExtractorAndPart (requestOrigin.spec);
    
    if (extractor && part) {
      extractor.lastAccessTime = (new Date ()).getTime ();
      
      let [targetPart, fragment]
        = extractor.checkPart (part, contentLocation.spec, false);
      if (targetPart) {
        return Services.io.newURI
          (param.baseURI + targetPart.eParam.refName + fragment,
           null, null);
      }
    }
    
    return null;
  },
  
  /**
   * ファイル展開器の配列を返す
   *
   * @return Array
   *         ファイル展開器の配列を返す
   *         [UnMHTExtractor, ...]
   */
  getExtractors : function () {
    return [extractor for ([uri, extractor] in Iterator (this.extractors))]
    .sort (function (x, y) {
        return y.lastAccessTime - x.lastAccessTime;
      });
  },
  
  /**
   * mht への参照のチェックおよび
   * mht 内の URI を置き換える
   *
   * @param  Number contentType
   *         コンテントの種類
   * @param  nsIURI contentLocation
   *         対象の URI
   * @param  nsIURI requestOrigin
   *         参照元の URI
   * @param  Document/Element context
   *         ロード先
   * @return Boolean
   *         許可するかどうか
   */
  shouldLoad : function (contentType, contentLocation, requestOrigin, context) {
    if (!requestOrigin) {
      /* 呼び出し元が無い場合チェック不可能なので許可する */
      return true;
    }
    
    if (contentLocation.scheme == "chrome"
        || contentLocation.scheme == "data") {
      /* chrome, data への参照は対象外 */
      return true;
    }
    
    if (contentLocation.scheme == "unmht"
        && requestOrigin.scheme != "file"
        && requestOrigin.scheme != "unmht"
        && requestOrigin.scheme != "chrome"
        && requestOrigin.scheme != "moz-nullprincipal") {
      let param = UnMHTScheme.getUnMHTURIParam (contentLocation.spec);
      if (param.scheme == "file") {
        /* file, unmht, chrome 以外からの
         * unmht を経由した file への参照は拒否する */
        return false;
      }
    }
    
    if (requestOrigin.scheme != "unmht") {
      /* unmht 以外からの参照は対象外 */
      return true;
    }
    /* ---- ここから unmht からの参照 ---- */
    
    if (!context) {
      /* ロード先が無い場合 URI の置換えができないので許可する */
      return true;
    }
    
    if (context instanceof Ci.nsIDOMXULElement) {
      /* タブを開く場合対象外 */
      return true;
    }
    /* ---- ここからドキュメント内へのロード ---- */
    
    if (this.loadImageType == PERM_IMAGE_BLOCK_ALL
        && contentType == Ci.nsIContentPolicy.TYPE_IMAGE) {
      /* 画像の読み込みの拒否は Firefox の設定に従う */
      return false;
    }
    
    if (context instanceof Ci.nsIDOMHTMLDocument
        || context instanceof Ci.nsIDOMHTMLStyleElement) {
      /* CSS によるロードの場合 */
      return this.checkExternal (contentLocation);
    }
    /* ---- ここから HTML 要素へのロード ---- */
    
    let [extractor, originPart]
    = UnMHTExtractManager.getExtractorAndPart (requestOrigin.spec);
    if (!(extractor && originPart)) {
      /* 参照元のファイルが未展開、もしくは存在しないので拒否 */
      return false;
    }
    
    if (UnMHTScheme.isUnMHTURI (contentLocation.spec)) {
      let param = UnMHTScheme.getUnMHTURIParam (contentLocation.spec);
      if (param.original == extractor.original) {
        /* 既に同じ mht ファイル内への参照になっているので許可 */
        return true;
      }
    }
    
    /**
     * 属性を変更する関数
     *
     * @param  HTMLElement node
     *         対象の要素
     * @param  String attr
     *         属性名
     * @param  String value
     *         属性値
     */
    let setAttr = function (node, attr, value) {
      if (node instanceof Ci.nsIDOMHTMLEmbedElement
          || node instanceof Ci.nsIDOMHTMLObjectElement) {
        /* EMBED, OBJECT は属性を変えても効果が無いので
         * 要素を置き換える */
        node.ownerDocument.defaultView.setTimeout
        (function () {
          try {
            let newNode = node.cloneNode (true);
            newNode.setAttribute ("__unmht_" + attr,
                                  contentLocation.spec);
            newNode.setAttribute (attr, value);
            node.parentNode.replaceChild (newNode, node);
          }
          catch (e) {
          }
        }, 100);
      }
      else {
        node.setAttribute ("__unmht_" + attr, contentLocation.spec);
        node.setAttribute (attr, value);
          
        node.ownerDocument.defaultView.setTimeout
        (function () {
          try {
            node.setAttribute (attr, value);
          }
          catch (e) {
          }
        }, 100);
      }
    }.bind (this);
    
    /**
     * 属性を削除する関数
     *
     * @param  HTMLElement node
     *         対象の要素
     * @param  String attr
     *         属性名
     */
    let removeAttr = function (node, attr) {
      node.removeAttribute (attr);
      node.ownerDocument.defaultView.setTimeout
      (function () {
        try {
          node.removeAttribute (attr);
        }
        catch (e) {
        }
      }, 100);
    };
    
    let attr = "src";
    if (context instanceof Ci.nsIDOMHTMLObjectElement) {
      attr = "data";
    }
    
    let [targetPart, fragment]
    = extractor.checkPart (originPart, contentLocation.spec, false);
    if (targetPart) {
      setAttr (context, attr,
               extractor.baseURI + targetPart.eParam.refName + fragment); 
      return false;
    }
    
    if (contentLocation.scheme == "file") {
      /* file への読み込みは削除 */
      removeAttr (context, attr);
      return false;
    }
    
    if (contentLocation.spec.indexOf (requestOrigin.spec) == 0) {
      /* JavaScript 等で生成された参照元からの相対アドレス */
      /* **** FIXME: パスを遡るものはチェック不可 **** */
      let path = contentLocation.spec.substr (requestOrigin.spec.length);
      
      let origSpec;
      if (path.match (/^\?/)) {
        /* クエリのみ */
        origSpec
          = originPart.eParam.location.replace (/\?.*$/, "") + path;
      }
      else {
        origSpec
          = UnMHTFileUtil.jointPath (originPart.eParam.baseDir, path);
      }
      
      let origContentLocation;
      try {
        origContentLocation = Services.io.newURI (origSpec, null, null)
      }
      catch (e) {
        return this.checkExternal (contentLocation);
      }
      
      if (origContentLocation.spec != contentLocation.spec) {
        /* 本来のアドレスでチェックし直す */
        return this.shouldLoad (contentType, origContentLocation,
                                requestOrigin, context);
      }
    }
    
    return this.checkExternal (contentLocation);
  },
  
  /**
   * 外部読み込みをチェックする
   *
   * @param  nsIURI contentLocation
   *         対象の URI
   * @return Boolean
   *         許可するかどうか
   */
  checkExternal : function (contentLocation) {
    if (contentLocation.scheme == "unmht") {
      return true;
    }
    
    return !this.enableFilterLoadOffline;
  },
  
  /**
   * コンテンツがロードされたイベント
   * mht ファイルの展開、アドレスの変換を行う
   *
   * @param  XULElement tabbrowser
   *         対象のタブブラウザ
   * @param  Event event
   *         対象のイベント
   */
  onDOMContentLoaded : function (tabbrowser, event) {
    let targetWindow = event.target.defaultView;
    let targetDocument = targetWindow.document;
    let href = targetDocument.location.href;
    
    if (href.match (/[^\/]\.(mht(ml)?|eml)$/i)) {
      /* ハンドルしていない mht ファイルだった場合 */
      
      if (href.match (/^chrome:\/\/ietab\//)) {
        /* IETab がハンドル済み */
        return;
      }
        
      this.loadMHT (Services.io.newURI (href, null, null), targetWindow,
                    true,  /* 開いているドキュメントを使用する */
                    true,  /* 履歴を置き換える */
                    true,  /* 展開する */
                    "",    /* 開始ページを開く */
                    true); /* キャッシュを使用する */
    }
    if (UnMHTScheme.isUnMHTURI (href)) {
      let href = targetDocument.location.href;
      
      let param = UnMHTScheme.getUnMHTURIParam (href);
      let original = param.original;
      let scheme = param.scheme;
      
      if (scheme != "cache"
          && scheme != "cid"
          && scheme != "dummy"
          && scheme != "error"
          && scheme != "refresh") {
        /* 通常のファイル */
        let [extractor, part] = this.getExtractorAndPart (href);
        if (extractor) {
          if (part) {
            if (part == extractor.startPart) {
              this.hookReload (targetWindow);
            }
            
            this.replaceLink (part, targetDocument);
          }
        }
        else {
          /* 未展開のファイルだった場合 */
          let uri;
          try {
            uri = Services.io.newURI (original, null, null);
          }
          catch (e) {
            return;
          }
        
          this.loadMHT (uri, targetWindow,
                        false,    /* ドキュメント無し */
                        true,     /* 履歴を置き換える */
                        true,     /* 展開する */
                        original, /* 指定されたファイルを開く */
                        false);   /* キャッシュは使用しない */
        }
      }
      
      if (this.openMultiple == "tabs"
          || this.openMultiple == "first") {
        let [extractor, part] = this.getExtractorAndPart (href);
        if (extractor && part && part == extractor.startPart) {
          if ("x-unmht-tabs" in extractor.topPart.fields) {
            /* 複数のファイルを含む MHT の
             * インデックスを開いた */
            this.replaceTabs (tabbrowser, targetDocument, targetWindow);
          }
        }
      }
    }
    else {
      let re = href.match (/^unmht:\/\/([^\/]*)\/(.*)$/);
      if (re) {
        /* Ver.5.7.10 以前の URI の可能性がある */
        let [, host, path] = re;
        let newURI = "unmht://" + host + ".unmht" + "/" + path;
        
        if (UnMHTScheme.isUnMHTURI (newURI)) {
          targetDocument.location = newURI;
        }
      }
    }
  },
  
  /**
   * 保存元へのリンクを置き換える
   *
   * @param  UnMHTMIMEPart part
   *         対象のドキュメントのパート
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   */
  replaceLink : function (part, targetDocument) {
    Array.forEach
    (targetDocument.getElementsByTagName ("a"),
     function (node) {
      if ("className" in node
          && node.className == "unmht_link_to_original") {
        node.setAttribute ("href", part.contentLocation);
      }
    });
  },
  
  /**
   * インデックスページをタブに置き換える
   * 
   * @param  XULElement tabbrowser
   *         対象のタブブラウザ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Window targetWindow
   *         対象のウィンドウ
   */
  replaceTabs : function (tabbrowser, targetDocument, targetWindow) {
    let nodes = targetDocument.getElementsByTagName ("a");
    
    Array.forEach
    ((nodes.length && this.openMultiple == "first") ? [nodes [0]] : nodes,
     function (node) {
      if ("className" in node
          && node.className == "unmht-tabs") {
        tabbrowser.addTab (node.href);
      }
    });
    
    for (let i = 0; i < tabbrowser.tabs.length; i ++) {
      let tab = tabbrowser.tabs [i];
      if (tab.linkedBrowser
          && tab.linkedBrowser.contentWindow == targetWindow) {
        tabbrowser.removeTab (tab);
        break;
      }
    }
  },
  
  /**
   * リロードをフックする
   * 
   * @param  Window targetWindow
   *         対象のウィンドウ
   */
  hookReload : function (targetWindow) {
    let nav = null;
    
    try {
      nav
        = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
        .getInterface (Ci.nsIWebNavigation);
    }
    catch (e) {
      return;
    }
    
    let listener = {
      QueryInterface : XPCOMUtils.generateQI ([Ci.nsISupportsWeakReference,
                                               Ci.nsISHistoryListener]),
      
      OnHistoryReload : function (reloadURI, reloadFlags) {
        if (UnMHTScheme.isUnMHTURI (reloadURI.spec)) {
          let param = UnMHTScheme.getUnMHTURIParam (reloadURI.spec);
          let uri = Services.io.newURI (param.original, null, null);
      
          UnMHTExtractManager.loadMHT
          (uri, nav,
           false,         /* ドキュメント無し */
           true,          /* 履歴は置き換える */
           true,          /* 展開する */
           param.refName, /* 現在開いているファイルを開く */
           false);        /* キャッシュは使用しない */
          return false;
        }
        
        return true;
      },
  
      OnHistoryGoBack : function () { return true; },
      OnHistoryGoForward : function () { return true; },
      OnHistoryGotoIndex : function () { return true; },
      OnHistoryNewEntry : function () { return true; },
      OnHistoryPurge : function () { return true; }
    };
    
    try {
      nav.sessionHistory.addSHistoryListener (listener);
    }
    catch (e) {
      return;
    }
    
    targetWindow.addEventListener
    ("unload",
     function () {
      try {
        nav.sessionHistory.removeSHistoryListener (listener);
      }
      catch (e) {
      }
    }, true);
  },
  
  
  /**
   * mht ファイルを展開して開く
   *
   * @param  nsIURI uri
   *         開く URI
   * @param  Window targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   * @param  Boolean useCache
   *         キャッシュを使用するか
   */
  loadMHT : function (uri, targetWindow,
                      hasDocument, replace, extract, refName,
                      useCache) {
    if (uri.scheme == "file") {
      /* file プロトコルの場合、そのファイルから開く */
      let filename = UnMHTFileUtil.getFilenameFromURLSpec (uri.spec);
      this.loadMHTFromFile (uri, targetWindow,
                            hasDocument,
                            replace,
                            extract,
                            refName,
                            filename);
    }
    else {
      /* それ以外の場合、ウェブから開く */
      this.loadMHTFromWeb (uri, targetWindow,
                           hasDocument,
                           replace,
                           extract,
                           refName,
                           useCache);
    }
    
    if (UnMHTPref.getPref ("open.cache.expire")) {
      this.clearCache (UnMHTPref.getPref ("open.cache.expire.time") * 60000);
    }
  },
  
  /**
   * ファイルから mht ファイルを展開して開く
   *
   * @param  nsIURI uri
   *         開く URI (本来の URI)
   * @param  Window targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String filename
   *         開くファイル名 (一旦キャッシュに保存したものの場合もある)
   * @param  String refName
   *         対象のファイル
   */
  loadMHTFromFile : function (uri, targetWindow,
                              hasDocument, replace, extract, refName,
                              filename) {
    let text = UnMHTFileUtil.readTextData (UnMHTFileUtil.initFile (filename));
    if (text == "") {
      if (hasDocument) {
        this.loadMHTFromDocument (uri, targetWindow,
                                  true,
                                  replace,
                                  extract,
                                  refName);
      }
      else {
        this.loadErrorPage (targetWindow, uri.spec);
      }
      return;
    }
    
    let extractor = new UnMHTExtractor ();
    let urlspec
    = extractor.extractMHT (uri.spec, text, extract);
    
    if (urlspec) {
      this.addExtractor (extractor.baseURI, extractor);
      
      let nav
        = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
        .getInterface (Ci.nsIWebNavigation);
            
      let flag = 0;
      if (replace) {
        flag = nav.LOAD_FLAGS_REPLACE_HISTORY;
      }
            
      if (refName
          && refName in extractor.refParts) {
        urlspec += refName;
      }
      
      nav.loadURI (urlspec, flag, null, null, null);
    }
  },
  
  /**
   * ドキュメントから mht ファイルを展開して開く
   *
   * @param  nsIURI uri
   *         開く URI
   * @param  Window targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   */
  loadMHTFromDocument : function (uri, targetWindow,
                                  hasDocument, replace, extract, refName) {
    let text = "";
    Array.forEach
    (targetWindow.document.getElementsByTagName ("pre"),
     function (node) {
      node = node.firstChild;
      while (node) {
        text += node.nodeValue;
        node = node.nextSibling;
      }
    });
    
    if (text == "") {
      this.loadErrorPage (targetWindow, uri.spec);
      return;
    }
    
    text = UnMHTUnicodeConverter.convertFromUnicode (text, "utf-8");
    
    let extractor = new UnMHTExtractor ();
    let urlspec
    = extractor.extractMHT (uri.spec, text, extract);
    
    if (urlspec) {
      this.addExtractor (extractor.baseURI, extractor);
      
      let nav
        = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
        .getInterface (Ci.nsIWebNavigation);
            
      let flag = 0;
      if (replace) {
        flag = nav.LOAD_FLAGS_REPLACE_HISTORY;
      }
            
      if (refName
          && refName in extractor.refParts) {
        urlspec += refName;
      }
            
      nav.loadURI (urlspec, flag, null, null, null);
    }
  },
    
  /**
   * ウェブのファイルから mht ファイルを展開して開く
   *
   * @param  nsIURI uri
   *         開く URI
   * @param  Window targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   * @param  Boolean useCache
   *         キャッシュを使用するか
   */
  loadMHTFromWeb : function (uri, targetWindow,
                             hasDocument, replace, extract, refName,
                             useCache) {
    let leafName 
    = (new Date ()).getTime () + "_" + Math.floor (Math.random () * 1000);
    let filename
    = UnMHTFileUtil.cacheDirectory + UnMHTFileUtil.separator + leafName;
    
    let targetFile = UnMHTFileUtil.initFile (filename);
        
    let listener = {
      QueryInterface : XPCOMUtils.generateQI ([Ci.nsIWebProgressListener]),
  
      onStateChange : function (webProgress, request, stateFlags, status) {
        let httpStatus = 200;
        try {
          /* HTTP の場合ステータスをチェックする */
          httpStatus
            = request.QueryInterface (Ci.nsIHttpChannel).responseStatus;
        }
        catch (e) {
          /* HTTP でない場合はデフォルトの 200 のまま */
        }
    
        if (stateFlags & Ci.nsIWebProgressListener.STATE_STOP) {
          if (httpStatus == 200) {
            /* 転送が終了したら
             * (HTTP の場合は 200 ならば) */
            UnMHTExtractManager.loadMHTFromFile
            (uri, targetWindow,
             hasDocument,
             replace,
             extract,
             refName,
             filename);
            let targetFile = UnMHTFileUtil.initFile (filename);
            if (targetFile.exists ()) {
              targetFile.remove (true);
            }
          }
          else {
            /* ファイルが見付からなかった場合 */
            if (hasDocument) {
              UnMHTExtractManager.loadMHTFromDocument
              (uri, targetWindow,
               true,
               replace,
               extract,
               refName);
            }
            else {
              UnMHTExtractManager.loadErrorPage (targetWindow, uri.spec);
            }
          }
        }
      },
  
      onLocationChange : function () {},
      onProgressChange : function () {},
      onSecurityChange : function () {},
      onStatusChange : function () {}
    };
    
    /* ファイルを保存する */
    let webBrowserPersist
    = Cc ["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
    .createInstance (Ci.nsIWebBrowserPersist);
    if (useCache) {
      webBrowserPersist.persistFlags = 0;
    }
    else {
      webBrowserPersist.persistFlags
      = webBrowserPersist.PERSIST_FLAGS_BYPASS_CACHE;
    }
    webBrowserPersist.progressListener = listener;
    
    try {
      webBrowserPersist.saveURI (uri, null, null, null, null,
                                 targetFile);
      return;
    }
    catch (e) {
    }
    
    if (hasDocument) {
      this.loadMHTFromDocument (uri, targetWindow,
                                true,
                                replace,
                                extract,
                                refName);
      return;
    }
    
    this.loadErrorPage (targetWindow, uri.spec);
  },
  
  /**
   * フォーカスのあるウィンドウから mht ファイルを開く
   *
   * @param  Window targetWindow
   *         実行中のウィンドウ
   * @param  Boolean extract
   *         展開するか
   */
  loadMHTFromFocusedWindow : function (window, extract) {
    let targetWindow = window.document.commandDispatcher.focusedWindow;
    if (targetWindow == window) {
      targetWindow = content;
    }
    
    let uri
    = Services.io.newURI (targetWindow.document.location.href, null, null);
    
    if (UnMHTScheme.isUnMHTURI (uri.spec)) {
      uri.spec = UnMHTScheme.deUnMHTURI (uri.spec);
    }
        
    this.loadMHT (uri, targetWindow,
                  true,    /* 開いているドキュメントを使用する */
                  false,   /* 履歴はそのまま */
                  extract, /* 展開は引数次第 */
                  "",      /* 開始ページを開く */
                  true);   /* キャッシュを使用する */
  },
  
  /**
   * エラーメッセージを表示する
   *
   * @param  Window targetWindow
   *         対象のウィンドウ
   * @param  String location
   *         ファイルの URI
   */
  loadErrorPage : function (targetWindow, location) {
    let nav
    = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
    .getInterface (Ci.nsIWebNavigation);
    
    let flag = nav.LOAD_FLAGS_REPLACE_HISTORY;
    nav.loadURI (UnMHTScheme.enErrorURI (location), flag, null, null, null);
  },
  
  /**
   * キャッシュを削除する
   *
   * @param  Number expire
   *         ミリ秒単位の経過時間
   *         0 なら即削除
   */
  clearCache : function (expire) {
    this.removeOldExtractor (expire);
    
    if (expire == 0) {
      UnMHTFileUtil.clearCacheDirectory (expire);
    }
  }
};
UnMHTExtractManager.init ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTExtractManager.term ();
  });
