/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTArrayUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTHTMLUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTMIMEUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTMaxRunTime.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTExtractor"];

/**
 * パートの展開情報
 */
function UnMHTExtractParam () {
  let randomHex = function () {
    return Math.floor (Math.random () * 15).toString (16).toUpperCase ();
  };
  
  this.cid = [randomHex() for (i in UnMHTArrayUtil.range (1, 32))].join ("");
}
UnMHTExtractParam.prototype = {
  charset : "",         /* String  キャラクタセット */
  mimetype : "",        /* String  MIME-Type */
  
  isHTML : false,       /* Boolean  HTML かどうか */
  isCSS : false,        /* Boolean  CSS かどうか */
  
  isMixed : false,      /* Boolean  multipart/mixed かどうか */
  
  cid : "",             /* String  Content-ID 代わり */
  
  location : "",        /* String  ファイルのアドレス */
  normalLocation : "",  /* String  正規化したファイルのアドレス */
  
  contentBase : "",     /* String  Content-Base の値 */
  
  baseDir : "",         /* String  親ディレクトリ */
  referredBaseDir : "", /* String  参照元のファイルの親ディレクトリ */
  /*
   * referredBaseDir は
   * IE で保存された MHT ファイルにおける以下の問題を回避するためのもの
   *
   * [状況]
   *   ファイル
   *     http://unmht.org/x.html
   *     http://unmht.org/style/a.css
   *     http://unmht.org/style/b.css
   *   x.html 中の記述
   *     <link href="./style/a.css" ... />
   *   a.css 中の記述
   *     @import url("./b.css");
   *
   * [期待される結果]
   *   相対パス "./b.css" を a.css の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/style/b.css
   *
   *   これにより a.css から b.css が参照できる
   *
   * [実際の結果]
   *   相対パス "./b.css" を x.html の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/b.css
   *
   *   これにより a.css から b.css が参照できない
   *
   * [対策]
   *   1. すべての X から参照されているすべての Y について
   *      Y.referredBaseDir を X.baseDir とする
   *   2. Z.referredBaseDir が空でない Z について
   *      Z 内の相対パス R による参照について
   *      Z.baseDir から R を解決した URI を A とする
   *      Z.referredBaseDir から R を解決した URI を B とする
   *   3. F.contentLocation が A に一致するような F が存在しない、かつ
   *      G.contentLocation が B に一致するような G が存在する場合
   *      R による参照先を G とする
   *      この時、G.referredBaseDir は F.baseDir となり 2 から再度適用する
   *
   * [例]
   *   0. {x.html}.baseDir = "http://unmht.org/"
   *      {a.css}.baseDir = "http://unmht.org/style/"
   *      {b.css}.contentLocation = "http://unmht.org/b.css"
   *   1. {a.css}.referredBaseDir = {x.html}.baseDir
   *                                "http://unmht.org/"
   *   2. Z = {a.css}
   *      R = "./b.css"
   *      A = {a.css}.baseDir + R
   *        = "http://unmht.org/style/b.css"
   *      B = {a.css}.referredBaseDir + R
   *        = "http://unmht.org/b.css"
   *   3. A != {b.css}.contentLocation
   *      B == {b.css}.contentLocation
   *      # a.css から b.css が参照できる
   *      {b.css}.referredBaseDir = {a.css}.baseDir
   *                                "http://unmht.org/style/"
   *   2. Z = {b.css}
   *      A = {b.css}.baseDir + R
   *        = "http://unmht.org/" + R
   *      B = {b.css}.referredBaseDir + R
   *        = "http://unmht.org/style/" + R
   *      # b.css から "./i.png" で参照されたファイル i.png が
   *      #   {i.png}.contentLocation = "http://unmht.org/i.png"
   *      # となっていても
   *      #   {i.png}.contentLocation = "http://unmht.org/style/i.png"
   *      # となっていても参照できる
   *
   * [問題]
   *   誤って生成された Content-Location に一致するファイルが
   *   同 MHT 内に存在した場合これらを区別する事はできない
   */
  leafName : "",        /* String  ファイル名 */
  normalLeafName : "",  /* String  正規化したファイル名 */
  
  content : "",         /* String  修正済みのファイルの中身 */
  
  refName : ""          /* String  参照用のファイル名 */
};

/*
 * ファイル展開器
 */
function UnMHTExtractor () {
}
UnMHTExtractor.prototype = {
  original : "",         /* String  展開したファイル名の URI */
  baseURI : "",          /* String  開始ページの URI */
  
  startPart : null,      /* UnMHTMIMEPart 開始ページのパート */
  
  topPart : null,        /* UnMHTMIMEPart  トップレベルのパート */
  subject : "",          /* String  Subject の値 */
  date : "",             /* String  Date の値 */
  
  parts : null,          /* Array  mht ファイルの中にあるパート
                          *   [UnMHTMIMEPart パート,
                          *    ...] */
  refParts : null,       /* Object  参照できるパート
                          *   <String 参照用のファイル名,
                          *    UnMHTMIMEPart パート> */
  
  extractTime : 0,       /* Number  展開時刻 */
  lastAccessTime : 0,    /* Number  最終アクセス時刻 */
  
  size : 0,              /* Number  全キャッシュのサイズ */
  
  /**
   * データを開放する
   */
  destruct : function () {
    this.parts
    .forEach (function (part) {
        part.eParam.content = null;
        part.eParam = null;
      });
    this.parts = null;
    
    if (this.topPart) {
      UnMHTMIMEUtil.destructPart (this.topPart);
      this.topPart = null;
    }
    
    this.startPart = null;
    this.refParts = null;
  },
  
  /* ==== 展開 ==== */
    
  /**
   * mht ファイルを展開する
   *
   * @param  String originalURISpec
   *         展開したファイル名の URI 表記
   * @param  String text
   *         mht ファイルの内容
   * @param  Boolean extract
   *         展開するかどうか
   *           false の場合、text で与えられた内容のファイルを作る
   * @return String
   *         開始ページの URL
   */
  extractMHT : function (originalURISpec, text, extract) {
    this.original
    = originalURISpec
    .replace
    (/[\(\)\'\"\[\] ]/g, 
     function (matched) {
      let hexString = "0" + matched.charCodeAt (0).toString (16).toUpperCase ();
      return "%" + hexString.slice (-2);
    });
    
    this.baseURI = UnMHTScheme.enUnMHTURI (this.original);
    
    this.extractTime = (new Date ()).getTime ();
    
    UnMHTMaxRunTime.setExtractTime ();
    
    this.parts = [];
    this.refParts = {};
    
    if (extract) {
      this.topPart
        = UnMHTMIMEUtil.decodeMessage (text.replace (/\r\n|\r|\n/g, "\r\n"));
    }
    
    if (this.topPart) {
      if ("subject" in this.topPart.fields) {
        let value = this.topPart.fields ["subject"];
        this.subject = UnMHTMIMEUtil.decodeWord (value);
      }
      if ("date" in this.topPart.fields) {
        this.date = this.topPart.fields ["date"];
      }
      
      this.startPart = UnMHTMIMEUtil.findStartPart (this.topPart);
      
      this.createExtractParam (this.topPart, null);
      
      this.setRefName ();
      
      this.modifyAttribute ();
      
      this.modifyCSS ();
      
      this.createMixedDocument ();
    }
    else {
      let part = new UnMHTMIMEPart ();
      part.contentLocation = originalURISpec;
      part.eParam = new UnMHTExtractParam ();
      
      if (extract) {
        part.eParam.refName = "index.html";
        part.eParam.mimetype = "text/html";
      }
      else {
        part.eParam.refName = "index.txt";
        part.eParam.mimetype = "text/plain";
      }
      
      part.eParam.location = originalURISpec;
      part.eParam.charset = "utf-8";
      
      part.eParam.content = text;
      
      this.getPathInfo (part);
      
      this.topPart = part;
      this.startPart = part;
      this.parts.push (part);
    }
    
    this.getPartsInfo ();
    
    UnMHTMaxRunTime.restoreTime ();
    
    this.lastAccessTime = (new Date ()).getTime ();
    
    return this.baseURI;
  },
  
  /**
   * パートの情報を作成する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  UnMHTMIMEPart parentPart
   *         親のパート
   */
  createExtractParam : function (part, parentPart) {
    let isMixed = (part.isMultipart && part.contentSubType == "mixed");
    
    if (part.isMultipart) {
      part.parts.forEach
        (function (p) {
          this.createExtractParam (p, part);
        }.bind (this));
      
      if (!isMixed) {
        return;
      }
    }
    
    let eParam = new UnMHTExtractParam ();
    if (part.contentID) {
      eParam.cid = part.contentID;
    }
    if (isMixed) {
      eParam.isMixed = true;
      eParam.content = "";
      
      eParam.mimetype = "text/html";
      eParam.charset = "utf-8";
      
      eParam.isHTML = true;
      
      eParam.location = "mixed-" + (new Date ()).getTime () + ".html";
    }
    else {
      eParam.content = part.body;
    
      eParam.mimetype = part.mimetype;
      eParam.charset = part.charset;
      if (part.contentType == "text") {
        eParam.isHTML = true;
        if (part.contentSubType == "css") {
          eParam.isCSS = true;
        }
      }
    
      if ("content-base" in part.fields) {
        let value = part.fields ["content-base"];
        try {
          value = UnMHTUnicodeConverter.convertToUnicode (value, "utf-8");
        }
        catch (e) {
          try {
            value = UnMHTUnicodeConverter.convertToUnicode (value, "shift_jis");
          }
          catch (e) {
          }
        }
        eParam.contentBase = value;
      }
    
      if (eParam.contentBase) {
        if (part.contentLocation) {
          eParam.location
          = UnMHTFileUtil.jointPath (eParam.contentBase, part.contentLocation);
        }
        else {
          eParam.location = eParam.contentBase;
        }
      }
      else {
        eParam.location = part.contentLocation;
      }
      
      if (parentPart) {
        eParam.location
        = UnMHTFileUtil.jointPath (parentPart.contentLocation, eParam.location);
      }
      
      if (!eParam.location
          && eParam.mimetype == "text/html") {
        /* 開始ページの Content-Location が
         * 指定されない場合があるので index.html とする */
        eParam.location = "index.html";
      }
    }
    part.eParam = eParam;
    
    this.getPathInfo (part);
    
    this.parts.push (part);
    
    if (eParam.mimetype == "application/octet-stream"
        && part.body.substr (0, 32).indexOf ("<") != -1) {
      /* CGI 生成のページ等が application/octet-stream として
       * 保存されている場合がある */
      if (part.body.search (/<html/i) != -1) {
        eParam.mimetype = "text/html";
        eParam.isHTML = true;
      }
    }
  },
  
  /**
   * パスの情報を取得する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   */
  getPathInfo : function (part) {
    let leafNameRe = null;
    let eParam = part.eParam;
    
    let location = eParam.location;
    
    let re = location.match (/^([^#\?]*)(?:\?[^#]*)?(?:#.*)?$/);
    if (re) {
      [, location] = re;
    }
    
    if (location.indexOf ("mhtml:") == 0) {
      /* mhtml で始まる場合 (IE のプロトコル ?)
       * ! の前後に分割する */
      leafNameRe = /^(.*)\!([^\!]+)$/;
    }
    else if (location.indexOf ("\\") != -1) {
      leafNameRe = /^(.*)\\([^\\]*)$/;
    }
    else if (location.indexOf ("/") != -1) {
      leafNameRe = /^(.*)\/([^\/]*)$/;
    }
    else if (location.indexOf (":") != -1) {
      leafNameRe = /^(.*):([^:]*)$/;
    }
    
    if (leafNameRe != null) {
      re = location.match (leafNameRe);
      if (re) {
        [, eParam.baseDir, eParam.leafName] = re;
      }
      else {
        eParam.baseDir = location;
        eParam.leafName = "";
      }
    }
    else {
      eParam.baseDir = "";
      eParam.leafName = location;
    }
    
    if (part.contentDispositionFilename) {
      eParam.leafName = part.contentDispositionFilename;
    }
    
    eParam.referredBaseDir = "";
    
    eParam.normalLocation = UnMHTExtractUtil.normalizeName (eParam.location);
    eParam.normalLeafName = UnMHTExtractUtil.normalizeName (eParam.leafName);
  },
  
  /**
   * 参照名を重複しないように設定する
   */
  setRefName : function () {
    let indices = {};
    
    this.parts.forEach (function (part, i) {
        let refBase;
        let extension = "";
        
        let re = part.eParam.leafName.match (/^(.+)\.([A-Za-z0-9]+)$/);
        if (re) {
          [, refBase, extension] = re;
        }
        else if (part.eParam.leafName) {
          refBase = part.eParam.leafName;
        }
        else if (part.contentID) {
          refBase = part.contentID;
        }
        else if (part.eParam.mimetype == "text/html") {
          refBase = "index";
          extension = "html";
        }
        else {
          refBase = i;
        }
          
        refBase = encodeURIComponent (refBase).replace (/\//g, "_");
          
        if (refBase in indices) {
          part.eParam.refName = refBase + "." + indices [refBase];
          indices [refBase] ++;
        }
        else {
          indices [refBase] = 1;
          part.eParam.refName = refBase;
        }
          
        if (extension) {
          part.eParam.refName += "." + extension;
        }
      }.bind (this));
  },
  
  cssRe : /(@import\s*(?:url\s*\(\s*)?|[^a-zA-Z0-9_]url\s*\(\s*)(?:(\"|&quot;|&#x22;|&#34;)((?:\\\"|[^\"])*)\2|(\'|&apos;|&#x27;|&#39;)((?:\\\'|[^\'])*)\4)/ig,
  /**
   * CSS 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  String matched
   *         マッチした全体
   * @param  String prev
   * @param  String dquote
   * @param  String dqpath
   * @param  String quote
   * @param  String qpath
   *         以下参照
   *
   *         dquote
   *         -
   * @import "12345.jpg";
   * -------- ---------
   * prev     dqpath
   *
   *      quote
   *      -
   * url ('12345.jpg');
   * ----- ---------
   * prev  qpath
   *
   * @return String
   *         変換した結果
   */
  cssFunc : function (part,
                      matched, prev, dquote, dqpath, quote, qpath) {
    let path = dqpath || qpath;
    quote = quote || dquote || "";
    if (!prev.match (/\(/)) {
      quote = quote || "\'";
    }
    
    if (!path.match (/^[A-Za-z0-9\-]+:/)) {
      /* 相対パスは UnMHTProtocolHandler.newURI の方で変える */
      return matched;
    }
    
    let [targetPart, fragment]
    = this.checkPart (part, path, part.eParam.isCSS);
    if (targetPart) {
      targetPart.eParam.referredBaseDir = part.eParam.baseDir;
      return prev + quote
        + this.baseURI + targetPart.eParam.refName + quote;
    }
    
    if (path.match (/^file:/)) {
      return prev + quote + quote;
    }
    
    return matched;
  },
  /**
   * CSS 中の unmht: 形式の URI を解除する
   * replace に渡す関数
   *
   * 引数と返り値は part が無い以外は cssFunc と同じ
   */
  unCssFunc : function (matched, prev, dquote, dqpath, quote, qpath) {
    let path = dqpath || qpath;
    quote = quote || dquote || "";
    if (!prev.match (/\(/)) {
      quote = quote || "\'";
    }
    
    if (!path.match (/^unmht:/)) {
      return matched;
    }
    
    let param = UnMHTScheme.getUnMHTURIParam (path);
    let baseURI = param.baseURI;
    let refName = param.refName;
    let part = null;
    if (refName) {
      if (refName in this.refParts) {
        part = this.refParts [refName];
      }
    }
    else {
      part = this.startPart;
    }
    if (part) {
      let location = part.contentLocation;
      if (!location) {
        location = "cid:" + part.eParam.cid;
      }
      
      return prev + quote + location + quote;
    }
    
    return matched;
  },
  
  /**
   * CSS 中の URI を unmht: 形式に変換する
   */
  modifyCSS : function () {
    this.parts.forEach (function (part) {
        if (part.eParam.isHTML
            || part.eParam.isCSS) {
          part.eParam.content
          = part.eParam.content
          .replace (this.cssRe,
                    function (matched, prev,
                              dquote, dqpath, quote, qpath) {
                      return this.cssFunc
                      (part, matched, prev,
                       dquote, dqpath, quote, qpath);
                    }.bind (this));
        }
      }.bind (this));
  },
  /**
   * CSS 中の unmht: 形式の URI を解除する
   *
   * @param  String content
   *         対象のドキュメント
   * @return String
   *         変換したドキュメント
   */
  unmodifyCSS : function (content) {
    return content
    .replace (this.cssRe,
              function (matched, prev,
                         dquote, dqpath, quote, qpath) {
                return this.unCssFunc
                (matched, prev,
                 dquote, dqpath, quote, qpath);
              }.bind (this));
  },
  
  /**
   * multipart/mixed なドキュメントを作成する
   */
  createMixedDocument : function () {
    let createBody = function (part) {
      let toUTF8 = function (text) {
        return UnMHTUnicodeConverter.convertFromUnicode
        (text, "utf-8");
      };
        
      let getUTF8Content = function (part) {
        let content = part.eParam.content;
          
        if (!part.eParam.charset.match (/^utf-8$/i)) {
          content
            = UnMHTUnicodeConverter.convertFromUnicode
            (UnMHTUnicodeConverter.convertToUnicode
             (content, part.eParam.charset),
             "utf-8");
        }
          
        return content;
      };
        
      if (part.eParam.mimetype == "text/html") {
        let content = getUTF8Content (part);
          
        return content
          .replace (/<\/?(html|body)(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig, "")
          .replace (/<![\?A-Za-z0-9_:\-]+(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig, "");
      }
      else if (part.eParam.mimetype == "text/plain") {
        let content = UnMHTHTMLUtil.escapeEntity (getUTF8Content (part));
          
        return content
        .replace (/[ \t]/g, "&nbsp;")
        .replace (/\r\n|\r|\n/g, "<br />\n");
      }
      else {
        let href
        = UnMHTHTMLUtil.escapeEntity
        (this.baseURI + part.eParam.refName);
        let filename
        = UnMHTHTMLUtil.escapeEntity
        (part.eParam.leafName || part.contentDispositionFilename || "???");
          
        let content;
        if (part.contentType == "image") {
          content = "<img src=\"" + href + "\" alt=\"" + filename + "\" />";
        }
        else {
          content = "<a href=\"" + href + "\">" + filename + "</a>";
        }
          
        return toUTF8 (content);
      }
    }.bind (this);
    
    this.parts.forEach (function (part) {
        if (part.eParam.isMixed) {
          part.eParam.content
            = part.parts
            .map (function (p) createBody (UnMHTMIMEUtil.findStartPart (p)))
            .join ("");
        }
      }.bind (this));
  },
  
  elementRe : /(<([\?A-Za-z0-9_:\-]+))(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)(>)/g,
  attributeRe : /(\s(src|href|background|action|data)\s*=\s*)(?:(\")((?:\\\"|[^\"])*)\"|(\')((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))/ig,
  /**
   * HTML 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  String matched
   *         マッチした全体
   * @param  String prev
   * @param  String attrs
   * @param  String next
   * @return String
   *         変換した結果
   */
  elementFunc : function (part,
                          matched, prev, name, attrs, next) {
    if (name.match (/base/i)) {
      return matched;
    }
    let isObject = false;
    if (name.match (/object/i) || name.match (/embed/i)) {
      isObject = true;
    }
    
    attrs
    = attrs.replace
    (this.attributeRe,
     function (matched, prev, attr,
                dquote, dqpath, quote, qpath, path) {
      path = dqpath || qpath || path;
      quote = quote || dquote || "\"";
    
      if (path.match (/^mailto:/)) {
        return matched;
      }
       
      if (!isObject && attr.match (/data/i)) {
        return matched;
      }
       
      if (!path.match (/^[A-Za-z0-9\-]+:/)) {
        /* 相対パスは UnMHTProtocolHandler.newURI の方で変える */
        return matched;
      }
    
      let [targetPart, fragment]
      = this.checkPart (part, path, false);
      if (targetPart) {
        if (name == "link") {
          targetPart.eParam.referredBaseDir = part.eParam.baseDir;
        }
        
        return prev + quote
          + this.baseURI + targetPart.eParam.refName + fragment
          + quote;
      }
    
      if (path.match (/^file:/)
          || path.match (/^cid:/)) {
        return prev + quote + quote;
      }
    
      return matched;
    }.bind (this));
    
    return prev + attrs + next;
  },
  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   * replace に渡す関数
   *
   * 引数と返り値は part が無い以外は elementFunc と同じ
   */
  unElementFunc : function (matched, prev, name, attrs, next) {
    let isBase = false;
    if (name.match (/base/i)) {
      isBase = true;
    }
    
    attrs
    = attrs.replace
    (this.attributeRe,
     function (matched, prev, attr,
                dquote, dqpath, quote, qpath, path) {
      path = dqpath || qpath || path;
      quote = quote || dquote || "\"";
       
      if (!path.match (/^unmht:/)) {
        return matched;
      }
       
      let param = UnMHTScheme.getUnMHTURIParam (path);
      let baseURI = param.baseURI;
      let refName = param.refName;
      let part = null;
      if (refName) {
        if (refName in this.refParts) {
          part = this.refParts [refName];
        }
      }
      else {
        part = this.startPart;
      }
      if (part) {
        let location = part.contentLocation;
        if (isBase) {
          location = part.eParam.baseDir;
        }
        if (!location && !isBase) {
          location = "cid:" + part.eParam.cid;
        }
         
        return prev + quote + location + quote;
      }
       
      return matched;
    }.bind (this));
    
    return prev + attrs + next;
  },
  
  baseRe : /<base\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?href\s*=\s*(?:\"((?:\\\"|[^\"])*)\"|\'((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*>/ig,
  /**
   * HTML 中の BASE の値を収集して削除する
   * replace に渡す関数
   *
   * 引数と返り値は elementFunc と同じ
   */
  baseFunc : function (part,
                       matched, dqpath, qpath, path) {
    path = dqpath || qpath || path;
    
    part.eParam.baseDir
    = UnMHTFileUtil.jointPath
    (part.eParam.baseDir,
     UnMHTUnicodeConverter.convertToUnicode (path, part.eParam.charset)
     /* BASE 要素にディレクトリではなくファイルのパスが入っている場合が
      * あるので、末尾を削除する */
     .replace (/\/[^\/]+$/, ""))
    
    return "";
  },
  
  /**
   * HTML 中の URI を unmht: 形式に変換する
   * また、BASE 要素と Office 用の要素を削除し
   * 設定されていれば PowerPoint の警告ページをスキップする
   */
  modifyAttribute : function () {
    let isPPT = false;
    let framePart = null;
    
    this.parts.forEach (function (part) {
        if (part.eParam.refName == "frame.htm") {
          framePart = part;
        }
          
        if (!part.eParam.isCSS && part.eParam.isHTML) {
          part.eParam.content
            = part.eParam.content
            .replace (this.baseRe,
                      function (matched, dqpath, qpath, path) {
                        return this.baseFunc
                        (part, matched, dqpath, qpath, path);
                      }.bind (this))
            .replace (this.elementRe,
                      function (matched, prev, name, attrs, next) {
                        return this.elementFunc
                        (part, matched, prev, name, attrs, next);
                      }.bind (this));
            
          if (part.eParam.content.indexOf
              ("xmlns:v=\"urn:schemas-microsoft-com:vml\"") != -1) {
            part.eParam.content
              = part.eParam.content
              .replace (/<\/?(o|p|oa|v):(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?>/g, "");
            isPPT = true;
          }
        }
      }.bind (this));
    
    if (isPPT && framePart
        && UnMHTPref.getPref ("open.ppt.skip_warning")) {
      this.startPart = framePart;
    }
  },

  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   *
   * @param  String content
   *         対象のドキュメント
   * @return String
   *         変換したドキュメント
   */
  unmodifyAttribute : function (content) {
    return content
    .replace (this.elementRe,
              function (matched, prev, name, attrs, next) {
                return this.unElementFunc
                (matched, prev, name, attrs, next);
              }.bind (this));
  },
  
  /**
   * パートの情報を収集する
   */
  getPartsInfo : function () {
    this.parts.forEach (function (part) {
        this.refParts [part.eParam.refName] = part;
          
        this.size += part.eParam.content.length;
      }.bind (this));
  },
  
  /* ==== パート取得 ==== */
  
  /**
   * mht に含まれるパートを探す
   *
   * @param  UnMHTMIMEPart originPart
   *         呼び出し元のパート
   * @param  String path
   *         対象のパス
   * @param  Boolean referred
   *         referredBaseDir をチェックするか
   * @return Array
   *         [UnMHTMIMEPart 対象のパート,
   *          String アンカー文字列]
   *         見付からなければパートが null
   */
  checkPart : function (originPart, path, referred) {
    if (path == "about:blank") {
      return [null, ""];
    }
    
    let fragment = "";
    let re = path.match (/^([^#]*)(#.+)$/);
    if (re) {
      [, path, fragment] = re;
    }
    if (fragment && path == "") {
      return [originPart, fragment];
    }
    
    let normalPath = UnMHTExtractUtil.normalizeName (path);
    
    let part;
    
    part
    = UnMHTArrayUtil.find
    (this.parts,
     function (part) normalPath == part.eParam.normalLocation,
     null);
    if (part) { return [part, fragment]; }
    
    if (originPart.eParam.baseDir) {
      /* 相対パスでの参照をチェック */
      let resolvedPath
      = UnMHTExtractUtil.normalizeName
      (UnMHTFileUtil.jointPath (originPart.eParam.baseDir, normalPath));
      
      part
      = UnMHTArrayUtil.find
      (this.parts,
       function (part) resolvedPath == part.eParam.normalLocation,
       null);
      if (part) { return [part, fragment]; }
    }
    
    re = normalPath.match (/^cid:(.+)$/);
    if (re) {
      let cid = re [1];
      
      part
      = UnMHTArrayUtil.find
      (this.parts,
       function (part) cid == part.contentID,
       null);
      if (part) { return [part, fragment]; }
    }
    
    
    if (UnMHTScheme.isUnMHTURI (normalPath)) {
      let param = UnMHTScheme.getUnMHTURIParam (normalPath);
      let refName = param.refName;
      
      if (param.original == this.original) {
        if (!refName) {
          return [this.startPart, fragment];
        }
        
        if (refName in this.refParts) {
          return [this.refParts [refName], fragment];
        }
        
        return [null, ""]
      }
      
      /* JavaScript 等で生成された参照元からの絶対アドレスとしてチェック */
      /* **** FIXME: 相対アドレスは未対応 **** */
      
      let modPath
      = UnMHTExtractUtil.normalizeName
      (UnMHTFileUtil.jointPath (originPart.eParam.baseDir, "/" + param.path));
      
      part
      = UnMHTArrayUtil.find
      (this.parts,
       function (part) modPath == part.eParam.normalLocation,
       null);
      if (part) { return [part, fragment]; }
    }
    
    if (referred && originPart.eParam.referredBaseDir != "") {
      let referredPath
      = UnMHTExtractUtil.normalizeName
      (UnMHTFileUtil.jointPath (originPart.eParam.referredBaseDir, normalPath));
      
      part
      = UnMHTArrayUtil.find
      (this.parts,
       function (part) referredPath == part.eParam.normalLocation,
       null);
      if (part) { return [part, fragment]; }
    }
    
    /* ファイル名だけでも同じものがあれば使う
     * あまり良い方法ではない */
    let noQuery = normalPath.replace (/\?.+/, "");
    part
    = UnMHTArrayUtil.find
    (this.parts, function (part) {
      let leafName = part.eParam.normalLeafName;
      let length = leafName.length;
      return (leafName
              && noQuery.length >= length + 1
              && noQuery.substr (noQuery.length - length) == leafName
              && noQuery.substr (noQuery.length - length - 1, 1)
              .match (/[\/\?\\:]/));
    }, null);
    if (part) { return [part, fragment]; }
    
    /* 含まれない */
    return [null, ""];
  }
};
