/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTMIMEUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTSave.jsm");
Cu.import ("resource://unmht/modules/UnMHTVersion.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFileConverter"];

/**
 * 変換情報
 *
 * @implements nsITimerCallback
 */
function UnMHTFileConversion () {
}
UnMHTFileConversion.prototype = {
  listener : null, /* UnMHTFileConverterListener  イベントリスナ */
  suffix : "",     /* String  ディレクトリ名の末尾 */
  count : 0,       /* Number  変換したファイル数 */
  
  timer : null,    /* nsITimer 実行中のタイマ */
  
  QueryInterface: XPCOMUtils.generateQI ([Ci.nsITimerCallback]),
  
  /* ==== nsITimerCallback ==== */
  
  /**
   * 時間経過イベント
   *
   * @param  nsITimer timer
   *         呼び出し元のタイマ
   */
  notify : function (timer) {
    if (this.files.length == 0) {
      this.listener.onComplete (this.count);
      this.listener = null;
      return;
    }
    
    let file = this.files.shift ();
    
    if (file.path.match (/\.mht(ml)?$/i)) {
      UnMHTFileConverter.setTimer (this);
      return;
    }
    
    this.listener.onStartFile (file);
    
    let base = file.parent.path;
    let filenameBase = file.leafName.replace (/\.[^\.]+$/, "");
    
    let findDir = function (suffix) {
      let d = UnMHTFileUtil.initFile (base);
      d.appendRelativePath (filenameBase + suffix);
      if (d && d.exists () && d.isDirectory ()) {
        return d;
      }
      return null;
    };
    
    let dir = findDir (this.suffix) || findDir ("_files") || findDir (".files");
    if (!dir) {
      UnMHTFileConverter.setTimer (this);
      return;
    }
    
    let resourceFiles
    = UnMHTFileUtil.getDirectoryEntries (dir).filter (function (f) f.isFile ());
    
    let filename
    = UnMHTSave.getUniqueFilename (base, filenameBase);
    
    let mhtFile = UnMHTFileUtil.initFile (base);
    mhtFile.appendRelativePath (filename);
    
    if (UnMHTFileConverter.convertFile
        (mhtFile, file, dir.leafName, resourceFiles, file.leafName)) {
      this.count ++;
    }
    
    UnMHTFileConverter.setTimer (this);
    
    this.timer = null;
  }
};

/**
 * HTML ファイルを MHT ファイルに変換
 */
let UnMHTFileConverter = {
  /**
   * 複数の HTML ファイルを MHT ファイルに変換する
   *
   * @param  UnMHTFileConverterListener listener
   *         イベントリスナ
   *         interface UnMHTFileConverterListener {
   *           void onComplete (in int count);
   *           void onStartFile (in nsIFile file);
   *         };
   * @param  String suffix
   *         ディレクトリ名の末尾
   * @param  Array files
   *         対象の HTML ファイル
   *         [nsIFile, ...]
   */
  startConvertFiles : function (listener, suffix, files) {
    let conv = new UnMHTFileConversion ();
    conv.listener = listener;
    conv.suffix = suffix;
    conv.files = files;
    
    this.setTimer (conv);
  },
  
  /**
   * 次のタイマーをセットする
   *
   * @param  UnMHTFileConversion conv
   *         対象の変換情報
   */
  setTimer : function (conv) {
    let timer = Cc ["@mozilla.org/timer;1"].createInstance (Ci.nsITimer);
    conv.timer = timer;
    timer.initWithCallback (conv, 100, Ci.nsITimer.TYPE_ONE_SHOT);
  },
  
  /**
   * HTML ファイルを MHT に変換
   *
   * @param  nsIFile mhtFile
   *         保存先のファイル
   * @param  nsIFile htmlFile
   *         対象の HTML ファイル
   * @param  String dirname
   *         他のリソースファイルを格納したディレクトリ名
   * @param  Array resourceFiles
   *         他のリソースファイル
   *         [nsIFile, ...]
   * @param  String subject
   *         Subject の値
   * @return Boolean
   *         変換に成功したか
   */
  convertFile : function (mhtFile, htmlFile, dirname, resourceFiles, subject) {
    let fstream = UnMHTFileUtil.openOutputStream (mhtFile);
    
    let topPart = new UnMHTMIMEPart ();
    topPart.isMultipart = true;
    topPart.boundary = UnMHTMIMEUtil.createUniqueBoundary ();
    topPart.startMimetype = "text/html";
    
    UnMHTMIMEUtil.addField (topPart, "From",
                            "<Saved by UnMHT Ver." + UnMHTVersion + ">");
    UnMHTMIMEUtil.addField (topPart, "Subject", subject);
    UnMHTMIMEUtil.addField (topPart, "Date",
                            UnMHTMIMEUtil.formatDate (new Date ()));
    UnMHTMIMEUtil.addField (topPart, "MIME-Version", "1.0");
    UnMHTMIMEUtil.addField (topPart, "Content-Type",
                            "multipart/related"
                            + ";boundary=\"" + topPart.boundary + "\""
                            + ";type=\"" + topPart.startMimetype + "\"");
    topPart.parts = [];
    
    [htmlFile].concat (resourceFiles).forEach (function (file) {
        let ext = "";
        let re = file.leafName.match (/\.([^\.]+)(\?.+)?(#.+)?$/i);
        if (re) {
          ext = re [1];
        }
        let mimetype = UnMHTMIMEUtil.getTypeFromExtension (ext);
        
        let location;
        if (file == htmlFile) {
          location = file.leafName;
        }
        else {
          location = dirname + "/" + file.leafName;
        }
        
        let part = new UnMHTMIMEPart ();
        part.contentTransferEncoding = "base64";
        UnMHTMIMEUtil.addField (part, "Content-Type",
                                mimetype);
        UnMHTMIMEUtil.addField (part, "Content-Transfer-Encoding",
                                part.contentTransferEncoding);
        UnMHTMIMEUtil.addField (part, "Content-Location", location);
        
        part.body = UnMHTFileUtil.readBinaryData (file);
      
        topPart.parts.push (part);
      });
    
    UnMHTMIMEUtil.encodeMessage (topPart, fstream);
    
    UnMHTMIMEUtil.destructPart (topPart);
    
    try {
      mhtFile.lastModifiedTime = htmlFile.lastModifiedTime;
    }
    catch (e) {
    }
        
    return true;
  }
};
