/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFileUtil"];

let PR_RDONLY = 0x01;
let PR_WRONLY = 0x02;
let PR_CREATE_FILE = 0x08;
let PR_TRUNCATE = 0x20;

/**
 * ファイル操作関係ユーティリティ関数
 */
let UnMHTFileUtil = {
  /* nsIFileProtocolHandler  プロトコルの変換器 */
  fileProtocolHandler : Services.io.getProtocolHandler ("file")
  .QueryInterface (Ci.nsIFileProtocolHandler),

  separator : "\\",           /* String  ネイティブパスのセパレータ */
  systemDirectory : "",       /* String  システムディレクトリ
                               *   Profile ディレクトリ/UnMHT */
  cacheDirectory : "",        /* String  キャッシュディレクトリ
                               *   Profile ディレクトリ/UnMHT/Cache */
  
  /**
   * ファイルを初期化する
   *
   * @param  String path
   *         ファイルのパス
   * @return nsIFile
   *         指定されたパスのファイル
   *         失敗した場合は null
   */
  initFile : function (path) {
    let file
    = Cc ["@mozilla.org/file/local;1"].createInstance (Ci.nsIFile);
    try {
      file.initWithPath (path);
      return file;
    }
    catch (e) {
    }
    
    return null;
  },
  
  /**
   * ファイルの入力ストリームを開く
   *
   * @param  nsIFile file
   *         対象のファイル
   * @return nsIFileInputStream
   *         ファイルの入力ストリーム
   *         失敗したら null
   */
  openInputStream : function (file) {
    let fstream
    = Cc ["@mozilla.org/network/file-input-stream;1"]
    .createInstance (Ci.nsIFileInputStream);
    try {
      fstream.init (file, PR_RDONLY, parseInt ("0444", 8), 0);
    }
    catch (e) {
      return null;
    }
    
    return fstream;
  },
  
  /**
   * ファイルの出力ストリームを開く
   *
   * @param  nsIFile file
   *         対象のファイル
   * @param  String permission
   *         パーミッション (オプショナル)
   *         undefined の場合は 0644
   * @return nsIFileOutputStream
   *         ファイルの出力ストリーム
   *         失敗したら null
   */
  openOutputStream : function (file, permission) {
    let fstream
    = Cc ["@mozilla.org/network/file-output-stream;1"]
    .createInstance (Ci.nsIFileOutputStream);
    
    try {
      fstream.init (file, PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE,
                    parseInt (permission || "0644", 8), 0);
    }
    catch (e) {
      return null;
    }
    
    return fstream;
  },
  
  /**
   * ファイルからテキストデータを読み込む
   *
   * @param  nsIFile file
   *         対象のファイル
   * @return String
   *         ファイルのテキストデータ
   */
  readTextData : function (file) {
    let fstream = this.openInputStream (file);
    if (!fstream) {
      return "";
    }
    
    let sstream
    = Cc ["@mozilla.org/scriptableinputstream;1"]
    .createInstance (Ci.nsIScriptableInputStream);
    sstream.init (fstream);
    
    let data = sstream.read (-1);
    
    sstream.close ();
    fstream.close ();
    
    return data;
  },
  
  /**
   * ファイルからバイナリデータを読み込む
   *
   * @param  nsIFile file
   *         対象のファイル
   * @return String
   *         ファイルのバイナリデータ
   */
  readBinaryData : function (file) {
    let fstream = this.openInputStream (file);
    if (!fstream) {
      return "";
    }
    
    let bstream
    = Cc ["@mozilla.org/binaryinputstream;1"]
    .createInstance (Ci.nsIBinaryInputStream);
    bstream.setInputStream (fstream);
    
    let data = bstream.readBytes (file.fileSize);
    
    bstream.close ();
    fstream.close ();
    
    return data;
  },
  
  /**
   * ディレクトリの内容を配列で返す
   *
   * @param  nsIFile dir
   *         対象のディレクトリ
   * @return Array
   *         ディレクトリの内容のファイル
   *         [nsIFile, ...]
   */
  getDirectoryEntries : function (dir) {
    let files = [];
    
    return [file for (file in XPCOMUtils.IterSimpleEnumerator
                        (dir.directoryEntries, Ci.nsIFile))];
  },
  
  /**
   * ファイル名に使えない文字を変換する
   *
   * @param  String filename
   *         元のファイル名
   * @return String
   *         変換したファイル名
   */
  validateFileName : function (filename) {
    let runtime
    = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    
    let sepRe = /\//g;
    
    if (runtime.OS == "WINNT") {
      sepRe = /[\\\/]/g;
      
      /* ファイル名に使えない文字
       *   |
       *   *
       *   :
       *   ?
       *   "
       *   <
       *   >
       * ファイル名に使用すると IE が MHT ファイルを開けなくなる文字
       *   #
       *   !
       */
      filename = filename.replace (/\"/g, "'")
        .replace (/[\*\:\?#\!\|]/g, "_")
        .replace (/</g, "(")
        .replace (/>/g, ")");
    }
    else if (runtime.OS == "Darwin") {
      sepRe = /[\:\/]/g;
    }
    
    return filename.replace (sepRe, "_");
  },
  
  /**
   * パスを展開、連結する
   *
   * @param  String base
   *         親ディレクトリのパス
   * @param  String sub
   *         相対パス、もしくは絶対パス
   * @return String
   *         連結したパス
   */
  jointPath : function (base, sub) {
    if (!base) {
      return sub;
    }
    
    /* セパレータを検出 */
    let sep;
    let pathRe = null;
    if (base.indexOf ("\\") != -1) {
      /* Windows のパスの場合 */
      sep = "\\";
      pathRe = /[^\\]+/g;
    }
    else if (base.indexOf ("/") != -1) {
      /* Mac OS X、Linux のパス、URI の場合 */
      sep = "/";
      pathRe = /[^\/]+/g;
    }
    else if (base.indexOf (":") != -1) {
      /* Mac OS 9 以前のパスの場合 */
      sep = ":";
      pathRe = /[^:]+/g;
    }
    else {
      /* 区切りが見付からない場合連結できないので
       * サブのみ返す */
      return sub;
    }
    
    /* スキーム、ホスト名の取得 */
    let scheme = "";
    let host = "";
    if (base.match (/^([A-Za-z0-9\-]+:\/\/\/?)([^\/]+)(.*)$/)) {
      /* URI の場合 */
      scheme = RegExp.$1;
      host = RegExp.$2;
      base = RegExp.$3;
    }
    else if (base.match (/^([A-Za-z]+:)(.*)$/)) {
      /* Windows のネイティブなパスの場合 */
      scheme = RegExp.$1;
      host = "";
      base = RegExp.$2;
    }
    else if (base.match (/^(\\)(\\.*)$/)) {
      /* Windows のネットワークドライブの場合 */
      scheme = RegExp.$1;
      host = "";
      base = RegExp.$2;
    }
    
    let re = sub.match (/^\/\/\/?(.*)$/);
    if (re) {
      /* プロトコルのみ継承する */
      return scheme + re [1];
    }
    
    if (sub.match (/^([A-Za-z0-9\-]+:\/\/\/?)([^\/]+)(.*)$/)
        || sub.match (/^([A-Za-z]+:)(.*)$/)
        || sub.match (/^(\\)(\\.*)$/)) {
      /* sub が URI、 Windows のネイティブな絶対パス、
       * ネットワークドライブの場合 */
      return sub;
    }
        
    if (sub.substr (0, 1) == sep) {
      /* sub が絶対パスの場合 */
      base = "";
    }
    
    if (sub.length > 5 && sub.substr (0, 5) == "data:") {
      /* sub が data スキームの場合 */
      return sub;
    }
    
    let path = base + sep + sub;
    
    let isDir = false;
    if (path.substr (path.length - 1, 1) == sep) {
      isDir = true;
    }
    
    let newNames = [];
    path.replace
    (pathRe,
     function (matched) {
      if (matched == ".") {
      }
      else if (matched == "..") {
        newNames.pop ();
      }
      else if (matched == "") {
      }
      else {
        newNames.push (matched);
      }
      return "";
    });
    
    path = sep + newNames.join (sep);
    if (isDir) {
      path += sep;
    }
    
    return scheme + host + path;
  },
  
  /**
   * ディレクトリを作成する
   *
   * @param  nsIFile dir
   *         対象のディレクトリ
   */
  createDirectory : function (dir) {
    if (!dir.exists ()) {
      dir.create (Ci.nsIFile.DIRECTORY_TYPE, parseInt ("0755", 8));
    }
  },
  
  /**
   * file プロトコルをネイティブパスに変換する
   *
   * @param  String uri
   *         file プロトコル
   * @return String
   *         ネイティブパス
   */
  getFilenameFromURLSpec : function (uri) {
    return this.fileProtocolHandler.getFileFromURLSpec (uri).path;
  },
  
  /**
   * file プロトコルをファイルに変換する
   *
   * @param  String uri
   *         file プロトコル
   * @return nsIFile
   *         ファイル、もしくは
   */
  getFileFromURLSpec : function (uri) {
    return this.fileProtocolHandler.getFileFromURLSpec (uri);
  },
  
  /**
   * ファイルを file プロトコルに変換する
   *
   * @param  nsIFile file
   *         ファイル
   * @return String
   *         file プロトコル
   */
  getURLSpecFromFile : function (file) {
    return this.fileProtocolHandler.getURLSpecFromFile (file);
  },
  
  /**
   * ネイティブパスを file プロトコルに変換する
   *
   * @param  String filename
   *         ネイティブパス
   * @return String
   *         file プロトコル
   */
  getURLSpecFromFilename : function (filename) {
    let targetFile = this.initFile (filename);
    return this.fileProtocolHandler.getURLSpecFromFile (targetFile);
  },
  
  /**
   * 現在のユーザの Profile ディレクトリを取得する
   *
   * @return  String
   *          現在のユーザの Profile ディレクトリ
   */
  getProfileDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("ProfD", Ci.nsIFile).path;
    }
    catch (e) {
    }
        
    return "";
  },
  
  /**
   * システムディレクトリを作成する
   */
  makeSystemDirectory : function () {
    let runtime
    = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    
    let dirname = this.getProfileDirectory ();
    if (runtime.OS == "WINNT") {
      this.separator = "\\";
    }
    else {
      this.separator = "/";
    }
    this.systemDirectory = dirname + this.separator + "UnMHT";
    
    let dir = this.initFile (this.systemDirectory);
    this.createDirectory (dir);
  },
  
  /**
   * キャッシュディレクトリを作成する
   */
  makeCacheDirectory : function () {
    this.cacheDirectory = this.systemDirectory + this.separator + "Cache";
    
    let dir = this.initFile (this.cacheDirectory);
    this.createDirectory (dir);
  },
  
  /**
   * キャッシュを削除する
   *
   * @param  Number expire
   *         ミリ秒単位の経過時間
   *         0 なら即削除
   */
  clearCacheDirectory : function (expire) {
    try {
      let dir = this.initFille (this.cacheDirectory ());
      if (!dir.exists ()) {
        return;
      }
      
      let now = (new Date ()).getTime ();
      
      let files = this.getDirectoryEntries (dir);
      files.forEach (function (file) {
          if (expire > 0) {
            let createTime
              = parseInt (file.leafName.replace (/_[0-9]*$/g, ""), 10);
            
            if (now < createTime + expire) {
              /* 作成から指定時間経過していない */
              return;
            }
          }
                    
          file.remove (true);
        });
    }
    catch (e) {
    }
  },
  
  /**
   * ファイルが存在するかどうか返す
   *
   * @param  String path
   *         ファイルのパス
   * @return Boolean
   *         ファイルが存在するか
   */
  isExist : function (path) {
    let file = this.initFile (path);
    return file && file.exists ();
  }
};

UnMHTFileUtil.makeSystemDirectory ();
UnMHTFileUtil.makeCacheDirectory ();
