/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTSave.jsm");
Cu.import ("resource://unmht/modules/UnMHTSaveManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMAFIntegration"];

/**
 * MAF 組み込みオブジェクト
 *
 * @param  Boolean useCurrentState
 *         現在の状態か
 * @param  Boolean useAdditionalFiles
 *         追加のファイルを使用するか
 * @implements nsIWebBrowserPersist
 * @implements nsICancelable
 */
function UnMHTMAFSavePersist (useCurrentState, useAdditionalFiles) {
  this._useCurrentState = useCurrentState;
  this._useAdditionalFiles = useAdditionalFiles;
  this.currentState = Ci.nsIWebBrowserPersist.PERSIST_STATE_READY;
}
UnMHTMAFSavePersist.prototype = {
  _useCurrentState : false,    /* Boolean  現在の状態を使用するか */
  _useAdditionalFiles : false, /* Boolean  追加ファイルを使用するか */
  
  QueryInterface : XPCOMUtils.generateQI ([Ci.nsIWebBrowserPersist,
                                           Ci.nsICancelable]),
  
  /* ==== nsICancelable ==== */
  
  /**
   * 保存キャンセル
   *
   * @param  nsresult reason
   *         理由
   */
  cancel : function (reason) {
    this.result = reason;
  },
    
  /* ==== nsIWebBrowserPersist ==== */
  
  persistFlags : 0,                  /* Number  データ取得方法フラグ */
  currentState : 0,                  /* Number  状態 */
  result : Cr.NS_OK,                 /* nsresult  理由 */
  progressListener : null,           /* nsIWebProgressListener
                                      *   進捗状況リスナ */
  
  /**
   * URI を保存する
   *
   * @param  nsIURI URI
   *         対象の URI
   * @param  nsISupports cacheKey
   *         キャッシュオブジェクト
   * @param  nsIURI referrer
   *         リファラ
   * @param  nsIInputStream postData
   *         POST データ
   * @param  String extraHeaders
   *         拡張ヘッダ
   * @param  nsIFile file
   *         対象のファイル
   */
  saveURI : function (URI, cacheKey, referrer, postData, extraHeaders, file) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /**
   * チャネルを保存する
   *
   * @param  nsIChannel channel
   *         対象のチャネル
   * @param  nsIFile file
   *         対象のファイル
   */
  saveChannel: function (channel, file) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /**
   * 保存開始イベント
   *
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISupports targetFile
   *         保存先のファイル (nsIFile)
   * @param  nsISupports targetDataDir
   *         保存先のディレクトリ? (nsIFile/nsIURI)
   * @param  String outputContentType
   *         保存先のファイルの MIME-Type
   * @param  Number encodingFlags
   *         エンコーダに渡すのフラグ
   * @param  Number wrapColumn
   *         折り返し文字数
   */
  saveDocument: function (targetDocument, targetFile,
                          targetDataDir, outputContentType,
                          encodingFlags, wrapColumn) {
    try {
      this.currentState = Ci.nsIWebBrowserPersist
      .PERSIST_STATE_SAVING;
      if (this.progressListener) {
        this.progressListener.onStateChange
          (null, null,
           Ci.nsIWebProgressListener.STATE_START
           | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
           Cr.NS_OK);
      }
            
      targetFile
      = targetFile.QueryInterface (Ci.nsIFileURL).file;
            
      if (!targetFile.leafName.match (/\.mht(ml)?$/i)) {
        let path = targetFile.path + ".mht";
        file
          = Cc ["@mozilla.org/file/local;1"].createInstance (Ci.nsILocalFile);
        targetFile.initWithPath (path);
      }
            
      let callback = function () {
        this.result = Cr.NS_OK;
        this._onComplete ();
      }.bind (this);
      
      let chromeWindow = Services.wm.getMostRecentWindow ("navigator:browser");
      
      UnMHTSaveManager.saveDocumentToFile
      (chromeWindow, targetDocument,
       this._useCurrentState, this._useAdditionalFiles,
       targetFile, callback);
    }
    catch (e) {
      if (e instanceof Ci.nsIXPCException) {
        this.result = e.result;
      } else {
        this.result = Cr.NS_ERROR_FAILURE;
      }
            
      this._onComplete ();
    }
  },
  
  /**
   * 保存キャンセルイベント
   */
  cancelSave : function () {
    this.cancel (Cr.NS_BINDING_ABORTED);
  },
  
  /* ==== 他 ==== */
  
  /**
   * 保存完了イベント
   */
  _onComplete : function () {
    if (this.currentState == Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED) {
      return;
    }
        
    this.currentState
    = Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED;
        
    if (this.progressListener) {
      this.progressListener.onStateChange
        (null, null,
         Ci.nsIWebProgressListener.STATE_STOP |
         Ci.nsIWebProgressListener.STATE_IS_NETWORK,
         this.result);
    }
  }
};

/**
 * MAF 組み込み
 *
 * @param  Window window
 *         対象のウィンドウ
 */
function UnMHTMAFIntegration (window) {
  this.window = window;
  this.behaviors = [];
}
UnMHTMAFIntegration.prototype = {
  behaviors : null, /* Array  追加した組み込み
                     *   [InternalSaveBehavior, ...] */
  window : null,  /* Window  対象のウィンドウ */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
  },
  
  /**
   * 組み込む
   */
  onLoad : function () {
    if (!("InternalSaveBehavior" in this.window
          && "gInternalSaveBehaviors" in this.window)) {
      return;
    }
    
    let types = [];
    types.push ({
        label : UnMHTStringBundle.get ("filetype")
          + " "
          + UnMHTStringBundle.get ("filetype_current"),
          ext : "*.mht",
          useCurrentState : true,
          useAdditionalFiles : false
          });
    types.push ({
        label : UnMHTStringBundle.get ("filetype")
          + " "
          + UnMHTStringBundle.get ("filetype_original"),
          ext : "*.mht",
          useCurrentState : false,
          useAdditionalFiles : false
          });
    types.push ({
        label : UnMHTStringBundle.get ("filetype")
          + " "
          + UnMHTStringBundle.get ("filetype_current_add"),
          ext : "*.mht",
          useCurrentState : true,
          useAdditionalFiles : true
          });
    types.push ({
        label : UnMHTStringBundle.get ("filetype")
          + " "
          + UnMHTStringBundle.get ("filetype_original_add"),
          ext : "*.mht",
          useCurrentState : false,
          useAdditionalFiles : true
          });
    
    this.behaviors
    = types.map (function (type) {
        let newSaveBehavior = new this.window.InternalSaveBehavior ();
          
        newSaveBehavior.isComplete = true;
        newSaveBehavior.mandatoryExtension = true;
        newSaveBehavior.isValidForSaveMode = function (saveMode) {
          return saveMode & 0x01;
        };
        
        newSaveBehavior.getFileFilter
        = function (contentType, fileExtension) {
          let result = {
            title: type.label,
            extensionstring: type.ext
          };
          return result;
        };
        newSaveBehavior.getPersistObject
        = function (saveBrowsers) {
          let result;
          result
          = new UnMHTMAFSavePersist
          (type.useCurrentState, type.useAdditionalFiles);
          
          return result;
        };
          
        /* 末尾は *.* なのでその前に挿入する */
        this.window.gInternalSaveBehaviors.splice
        (this.window.gInternalSaveBehaviors.length - 1, 0, newSaveBehavior);
          
        return newSaveBehavior;
      }.bind (this));
  },
  
  /**
   * 組み込みを解除する
   */
  onUnload : function () {
    if (!("InternalSaveBehavior" in this.window
          && "gInternalSaveBehaviors" in this.window)) {
      return;
    }

    this.window.gInternalSaveBehaviors
    = this.window.gInternalSaveBehaviors.filter
    ((function (b) this.behaviors.indexOf (b) == -1).bind (this));
    
    this.behaviors = [];
  }
};
