/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTArrayUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMIMEPart", "UnMHTMIMEUtil"];

/*
 * [RFC2045]
 *
 *  MIME-message-headers := entity-headers
 *                          fields
 *                          version CRLF
 *                          ; The ordering of the header
 *                          ; fields implied by this BNF
 *                          ; definition should be ignored.
 *
 *  entity-headers := [ content CRLF ]
 *                    [ encoding CRLF ]
 *                    [ id CRLF ]
 *                    [ description CRLF ]
 *                    *( MIME-extension-field CRLF )
 *
 *  content := "Content-Type" ":" type "/" subtype
 *             *(";" parameter)
 *             ; Matching of media type and subtype
 *             ; is ALWAYS case-insensitive.
 *
 *  type := discrete-type / composite-type
 *
 *  discrete-type := "text" / "image" / "audio" / "video" /
 *                   "application" / extension-token
 *
 *  extension-token := ietf-token / x-token
 *
 *  ietf-token := <An extension token defined by a
 *                 standards-track RFC and registered
 *                 with IANA.>
 *
 *  x-token := <The two characters "X-" or "x-" followed, with
 *              no  intervening white space, by any token>
 *
 *  token := 1*<any (US-ASCII) CHAR except SPACE, CTLs,
 *              or tspecials>
 *
 *  tspecials :=  "(" / ")" / "<" / ">" / "@" /
 *                "," / ";" / ":" / "\" / <">
 *                "/" / "[" / "]" / "?" / "="
 *                ; Must be in quoted-string,
 *                ; to use within parameter values
 *
 *  composite-type := "message" / "multipart" / extension-token
 *
 *  subtype := extension-token / iana-token
 *
 *  iana-token := <A publicly-defined extension token. Tokens
 *                 of this form must be registered with IANA
 *                 as specified in RFC 2048.>
 *
 *  ; parameter is updated by [RFC2231]
 *
 *  encoding := "Content-Transfer-Encoding" ":" mechanism
 *
 *  mechanism := "7bit" / "8bit" / "binary" /
 *               "quoted-printable" / "base64" /
 *               ietf-token / x-token
 *
 *  id := "Content-ID" ":" msg-id
 *
 *  ; msg-id is defined in [RFC5322]
 *
 *  description := "Content-Description" ":" *text
 *
 *  ; text is defined in [RFC5322]
 *
 *  MIME-extension-field := <Any RFC 822 header field which
 *                           begins with the string
 *                           "Content-">
 *
 *  ; fields is defined in [RFC5322]
 *
 *  version := "MIME-Version" ":" 1*DIGIT "." 1*DIGIT
 *
 *  MIME-part-headers := entity-headers
 *                       [fields]
 *                       ; Any field not beginning with
 *                       ; "content-" can have no defined
 *                       ; meaning and may be ignored.
 *                       ; The ordering of the header
 *                       ; fields implied by this BNF
 *                       ; definition should be ignored.
 *
 * [RFC2045:Quoted-Printable]
 *
 *  quoted-printable := qp-line *(CRLF qp-line)
 *
 *  qp-line := *(qp-segment transport-padding CRLF)
 *             qp-part transport-padding
 *
 *  qp-segment := qp-section *(SPACE / TAB) "="
 *                ; Maximum length of 76 characters
 *
 *  qp-section := [*(ptext / SPACE / TAB) ptext]
 *
 *  ptext := hex-octet / safe-char
 *
 *  hex-octet := "=" 2(DIGIT / "A" / "B" / "C" / "D" / "E" / "F")
 *               ; Octet must be used for characters > 127, =,
 *               ; SPACEs or TABs at the ends of lines, and is
 *               ; recommended for any character not listed in
 *               ; RFC 2049 as "mail-safe".
 *
 *  safe-char := <any octet with decimal value of 33 through
 *               60 inclusive, and 62 through 126>
 *               ; Characters not listed as "mail-safe" in
 *               ; RFC 2049 are also not recommended.
 *
 *  transport-padding := *LWSP-char
 *                       ; Composers MUST NOT generate
 *                       ; non-zero length transport
 *                       ; padding, but receivers MUST
 *                       ; be able to handle padding
 *                       ; added by message transports.
 *
 *  qp-part := qp-section
 *             ; Maximum length of 76 characters
 *
 * [RFC2046]
 *
 *  boundary := 0*69<bchars> bcharsnospace
 *
 *  bchars := bcharsnospace / " "
 *
 *  bcharsnospace := DIGIT / ALPHA / "'" / "(" / ")" /
 *                   "+" / "_" / "," / "-" / "." /
 *                   "/" / ":" / "=" / "?"
 *
 *
 *  multipart-body := [preamble CRLF]
 *                    dash-boundary transport-padding CRLF
 *                    body-part *encapsulation
 *                    close-delimiter transport-padding
 *                    [CRLF epilogue]
 *
 *  preamble := discard-text
 *
 *  discard-text := *(*text CRLF)
 *                  ; May be ignored or discarded.
 *
 *  dash-boundary := "--" boundary
 *                   ; boundary taken from the value of
 *                   ; boundary parameter of the
 *                   ; Content-Type field.
 *
 *  transport-padding := *LWSP-char
 *                       ; Composers MUST NOT generate
 *                       ; non-zero length transport
 *                       ; padding, but receivers MUST
 *                       ; be able to handle padding
 *                       ; added by message transports.
 *
 *  body-part := <"message" as defined in RFC 822, with all
 *                header fields optional, not starting with the
 *                specified dash-boundary, and with the
 *                delimiter not occurring anywhere in the
 *                body part.  Note that the semantics of a
 *                part differ from the semantics of a message,
 *                as described in the text.>
 *
 *  encapsulation := delimiter transport-padding
 *                   CRLF body-part
 *
 *  delimiter := CRLF dash-boundary
 *
 *  close-delimiter := delimiter "--"
 *
 *  epilogue := discard-text
 *
 * [RFC2047]
 *
 *   encoded-word := "=?" charset "?" encoding "?" encoded-text "?="
 *                   ; RFC2231 is updated by RFC2231
 *
 *   charset      := token    ; see section 3
 *
 *   encoding     := token   ; see section 4
 *
 *   token        := 1*<Any CHAR except SPACE, CTLs, and especials>
 *
 *   especials    := "(" / ")" / "<" / ">" / "@" / "," / ";" / ":" / "
 *                   <"> / "/" / "[" / "]" / "?" / "." / "="
 *
 *   encoded-text := 1*<Any printable ASCII character other than "?"
 *                     or SPACE>
 *                   ; (but see "Use of encoded-words in message
 *                   ; headers", section 5)
 *
 *   comment      := "(" *(ctext / quoted-pair / comment / encoded-word) ")"
 *
 *   phrase := 1*( encoded-word / word )
 *
 * [RFC2110] (obsoleted)
 *
 *   content-base := "Content-Base:" absoluteURI
 *                   ; Content-Base header has been removed by RFC2557
 *
 * [RFC2183]
 *
 *   disposition := "Content-Disposition" ":"
 *                  disposition-type
 *                  *(";" disposition-parm)
 *
 *   disposition-type := "inline"
 *                     / "attachment"
 *                     / extension-token
 *                     ; values are not case-sensitive
 *
 *   disposition-parm := filename-parm
 *                     / creation-date-parm
 *                     / modification-date-parm
 *                     / read-date-parm
 *                     / size-parm
 *                     / parameter
 *
 *   filename-parm := "filename" "=" value
 *
 *   creation-date-parm := "creation-date" "=" quoted-date-time
 *
 *   modification-date-parm := "modification-date" "=" quoted-date-time
 *
 *   read-date-parm := "read-date" "=" quoted-date-time
 *
 *   size-parm := "size" "=" 1*DIGIT
 *
 *   quoted-date-time := quoted-string
 *                    ; contents MUST be an RFC 822 `date-time'
 *                    ; numeric timezones (+HHMM or -HHMM) MUST be used
 *
 * [RFC2231]
 *
 *   parameter := regular-parameter / extended-parameter
 *
 *   regular-parameter := regular-parameter-name "=" value
 *
 *   regular-parameter-name := attribute [section]
 *
 *   attribute := 1*attribute-char
 *
 *   attribute-char := <any (US-ASCII) CHAR except SPACE, CTLs,
 *                     "*", "'", "%", or tspecials>
 *
 *   section := initial-section / other-sections
 *
 *   initial-section := "*0"
 *
 *   other-sections := "*" ("1" / "2" / "3" / "4" / "5" /
 *                          "6" / "7" / "8" / "9") *DIGIT
 *                     ; removed trailing right paren
 *
 *   extended-parameter := (extended-initial-name "="
 *                          extended-value) /
 *                         (extended-other-names "="
 *                          extended-other-values)
 *
 *   extended-initial-name := attribute [initial-section] "*"
 *
 *   extended-other-names := attribute other-sections "*"
 *
 *   extended-initial-value := [charset] "'" [language] "'"
 *                             extended-other-values
 *
 *   extended-other-values := *(ext-octet / attribute-char)
 *
 *   ext-octet := "%" 2(DIGIT / "A" / "B" / "C" / "D" / "E" / "F")
 *
 *   charset := <registered character set name>
 *
 *   language := <registered language tag [RFC-1766]>
 *
 *   encoded-word := "=?" charset ["*" language] "?" encoded-text "?="
 *
 * [RFC2387]
 *
 *     related-param   := [ ";" "start" "=" cid ]
 *                        [ ";" "start-info"  "="
 *                           ( cid-list / value ) ]
 *                        [ ";" "type"  "=" type "/" subtype ]
 *                        ; order independent
 *
 *     cid-list        := cid cid-list
 *
 *     cid             := msg-id     ; c.f. [822]
 *
 *     value           := token / quoted-string    ; c.f. [MIME]
 *                           ; value cannot begin with "<"
 *
 * [RFC2557]
 *
 *   quoted-pair := ("\" text)
 *
 *   text             =   %d1-9 / ; Characters excluding CR and LF
 *                        %d11-12 /
 *                        %d14-127
 *
 *   WSP              =   SP / HTAB ; Whitespace characters
 *
 *   FWS              =   ([*WSP CRLF] 1*WSP) ; Folding white-space
 *
 *   ctext            =   NO-WS-CTL / ; Non-white-space controls
 *                        %d33-39 / ; The rest of the US-ASCII
 *                        %d42-91 / ; characters not including "(",
 *                        %d93-127 ; ")", or "\"
 *
 *   comment          =  "(" *([FWS] (ctext / quoted-pair / comment))
 *                        [FWS] ")"
 *
 *   CFWS             =   *([FWS] comment) (([FWS] comment) / FWS)
 *
 *   content-location =   "Content-Location:" [CFWS] URI [CFWS]
 *
 *   URI              =   absoluteURI | relativeURI
 *
 * [RFC4288]
 *
 *       type-name = reg-name
 *       subtype-name = reg-name
 *
 *       reg-name = 1*127reg-name-chars
 *       reg-name-chars = ALPHA / DIGIT / "!" /
 *                       "#" / "$" / "&" / "." /
 *                       "+" / "-" / "^" / "_"
 *
 *       parameter-name = reg-name
 *
 * [RFC5324] Core Rules
 *
 *        ALPHA          =  %x41-5A / %x61-7A   ; A-Z / a-z
 *
 *         BIT            =  "0" / "1"
 *
 *         CHAR           =  %x01-7F
 *                                ; any 7-bit US-ASCII character,
 *                                ;  excluding NUL
 *
 *
 *         CR             =  %x0D
 *                                ; carriage return
 *
 *         CRLF           =  CR LF
 *                                ; Internet standard newline
 *
 *         CTL            =  %x00-1F / %x7F
 *                                ; controls
 *
 *         DIGIT          =  %x30-39
 *                                ; 0-9
 *
 *         DQUOTE         =  %x22
 *                                ; " (Double Quote)
 *
 *         HEXDIG         =  DIGIT / "A" / "B" / "C" / "D" / "E" / "F"
 *
 *         HTAB           =  %x09
 *                                ; horizontal tab
 *
 *         LF             =  %x0A
 *                                ; linefeed
 *
 *         LWSP           =  *(WSP / CRLF WSP)
 *                                ; Use of this linear-white-space rule
 *                                ;  permits lines containing only white
 *                                ;  space that are no longer legal in
 *                                ;  mail headers and have caused
 *                                ;  interoperability problems in other
 *                                ;  contexts.
 *                                ; Do not use when defining mail
 *                                ;  headers and use with caution in
 *                                ;  other contexts.
 *
 *         OCTET          =  %x00-FF
 *                                ; 8 bits of data
 *
 *         SP             =  %x20
 *
 *         VCHAR          =  %x21-7E
 *                                ; visible (printing) characters
 *
 *         WSP            =  SP / HTAB
 *                                ; white space
 *
 * [RFC5322]
 *
 *   quoted-pair     =   ("\" (VCHAR / WSP)) / obs-qp
 *
 *   FWS             =   ([*WSP CRLF] 1*WSP) /  obs-FWS
 *                                          ; Folding white space
 *
 *   ctext           =   %d33-39 /          ; Printable US-ASCII
 *                       %d42-91 /          ;  characters not including
 *                       %d93-126 /         ;  "(", ")", or "\"
 *                       obs-ctext
 *
 *   ccontent        =   ctext / quoted-pair / comment
 *                       ; added encoded-word by [RFC2047]
 *
 *   comment         =   "(" *([FWS] ccontent) [FWS] ")"
 *
 *   CFWS            =   (1*([FWS] comment) [FWS]) / FWS
 *
 *   atext           =   ALPHA / DIGIT /    ; Printable US-ASCII
 *                       "!" / "#" /        ;  characters not including
 *                       "$" / "%" /        ;  specials.  Used for atoms.
 *                       "&" / "'" /
 *                       "*" / "+" /
 *                       "-" / "/" /
 *                       "=" / "?" /
 *                       "^" / "_" /
 *                       "`" / "{" /
 *                       "|" / "}" /
 *                       "~"
 *
 *   atom            =   [CFWS] 1*atext [CFWS]
 *
 *   dot-atom-text   =   1*atext *("." 1*atext)
 *
 *   dot-atom        =   [CFWS] dot-atom-text [CFWS]
 *
 *   specials        =   "(" / ")" /        ; Special characters that do
 *                       "<" / ">" /        ;  not appear in atext
 *                       "[" / "]" /
 *                       ":" / ";" /
 *                       "@" / "\" /
 *                       "," / "." /
 *                       DQUOTE
 *
 *   qtext           =   %d33 /             ; Printable US-ASCII
 *                       %d35-91 /          ;  characters not including
 *                       %d93-126 /         ;  "\" or the quote character
 *                       obs-qtext
 *
 *   qcontent        =   qtext / quoted-pair
 *
 *   quoted-string   =   [CFWS]
 *                       DQUOTE *([FWS] qcontent) [FWS] DQUOTE
 *                       [CFWS]
 *
 *   word            =   atom / quoted-string
 *
 *   phrase          =   1*word / obs-phrase
 *                       ; added encoded-word by [RFC2047]
 *
 *   unstructured    =   (*([FWS] VCHAR) *WSP) / obs-unstruct
 *
 *   date-time       =   [ day-of-week "," ] date time [CFWS]
 *
 *   day-of-week     =   ([FWS] day-name) / obs-day-of-week
 *
 *   day-name        =   "Mon" / "Tue" / "Wed" / "Thu" /
 *                       "Fri" / "Sat" / "Sun"
 *
 *   date            =   day month year
 *
 *   day             =   ([FWS] 1*2DIGIT FWS) / obs-day
 *
 *   month           =   "Jan" / "Feb" / "Mar" / "Apr" /
 *                       "May" / "Jun" / "Jul" / "Aug" /
 *                       "Sep" / "Oct" / "Nov" / "Dec"
 *
 *   year            =   (FWS 4*DIGIT FWS) / obs-year
 *
 *   time            =   time-of-day zone
 *
 *   time-of-day     =   hour ":" minute [ ":" second ]
 *
 *   hour            =   2DIGIT / obs-hour
 *
 *   minute          =   2DIGIT / obs-minute
 *
 *   second          =   2DIGIT / obs-second
 *
 *   zone            =   (FWS ( "+" / "-" ) 4DIGIT) / obs-zone
 *
 *   address         =   mailbox / group
 *
 *   mailbox         =   name-addr / addr-spec
 *
 *   name-addr       =   [display-name] angle-addr
 *
 *   angle-addr      =   [CFWS] "<" addr-spec ">" [CFWS] /
 *                       obs-angle-addr
 *
 *   group           =   display-name ":" [group-list] ";" [CFWS]
 *
 *   display-name    =   phrase
 *
 *   mailbox-list    =   (mailbox *("," mailbox)) / obs-mbox-list
 *
 *   address-list    =   (address *("," address)) / obs-addr-list
 *
 *   group-list      =   mailbox-list / CFWS / obs-group-list
 *
 *   addr-spec       =   local-part "@" domain
 *
 *   local-part      =   dot-atom / quoted-string / obs-local-part
 *
 *   domain          =   dot-atom / domain-literal / obs-domain
 *
 *   domain-literal  =   [CFWS] "[" *([FWS] dtext) [FWS] "]" [CFWS]
 *
 *   dtext           =   %d33-90 /          ; Printable US-ASCII
 *                       %d94-126 /         ;  characters not including
 *                       obs-dtext          ;  "[", "]", or "\"
 *
 *   message         =   (fields / obs-fields)
 *                       [CRLF body]
 *
 *   body            =   (*(*998text CRLF) *998text) / obs-body
 *
 *   text            =   %d1-9 /            ; Characters excluding CR
 *                       %d11 /             ;  and LF
 *                       %d12 /
 *                       %d14-127
 *
 *   fields          =   *(trace
 *                         *optional-field /
 *                         *(resent-date /
 *                          resent-from /
 *                          resent-sender /
 *                          resent-to /
 *                          resent-cc /
 *                          resent-bcc /
 *                          resent-msg-id))
 *                       *(orig-date /
 *                       from /
 *                       sender /
 *                       reply-to /
 *                       to /
 *                       cc /
 *                       bcc /
 *                       message-id /
 *                       in-reply-to /
 *                       references /
 *                       subject /
 *                       comments /
 *                       keywords /
 *                       optional-field)
 *
 *   orig-date       =   "Date:" date-time CRLF
 *
 *   from            =   "From:" mailbox-list CRLF
 *
 *   sender          =   "Sender:" mailbox CRLF
 *
 *   reply-to        =   "Reply-To:" address-list CRLF
 *
 *   to              =   "To:" address-list CRLF
 *
 *   cc              =   "Cc:" address-list CRLF
 *
 *   bcc             =   "Bcc:" [address-list / CFWS] CRLF
 *
 *   message-id      =   "Message-ID:" msg-id CRLF
 *
 *   in-reply-to     =   "In-Reply-To:" 1*msg-id CRLF
 *
 *   references      =   "References:" 1*msg-id CRLF
 *
 *   msg-id          =   [CFWS] "<" id-left "@" id-right ">" [CFWS]
 *
 *   id-left         =   dot-atom-text / obs-id-left
 *
 *   id-right        =   dot-atom-text / no-fold-literal / obs-id-right
 *
 *   no-fold-literal =   "[" *dtext "]"
 *
 *   subject         =   "Subject:" unstructured CRLF
 *
 *   comments        =   "Comments:" unstructured CRLF
 *
 *   keywords        =   "Keywords:" phrase *("," phrase) CRLF
 *
 *   resent-date     =   "Resent-Date:" date-time CRLF
 *
 *   resent-from     =   "Resent-From:" mailbox-list CRLF
 *
 *   resent-sender   =   "Resent-Sender:" mailbox CRLF
 *
 *   resent-to       =   "Resent-To:" address-list CRLF
 *
 *   resent-cc       =   "Resent-Cc:" address-list CRLF
 *
 *   resent-bcc      =   "Resent-Bcc:" [address-list / CFWS] CRLF
 *
 *   resent-msg-id   =   "Resent-Message-ID:" msg-id CRLF
 *
 *   trace           =   [return]
 *                       1*received
 *
 *   return          =   "Return-Path:" path CRLF
 *
 *   path            =   angle-addr / ([CFWS] "<" [CFWS] ">" [CFWS])
 *
 *   received        =   "Received:" *received-token ";" date-time CRLF
 *
 *   received-token  =   word / angle-addr / addr-spec / domain
 *
 *   optional-field  =   field-name ":" unstructured CRLF
 *
 *   field-name      =   1*ftext
 *
 *   ftext           =   %d33-57 /          ; Printable US-ASCII
 *                       %d59-126           ;  characters not including
 *                                          ;  ":".
 *
 *   obs-NO-WS-CTL   =   %d1-8 /            ; US-ASCII control
 *                       %d11 /             ;  characters that do not
 *                       %d12 /             ;  include the carriage
 *                       %d14-31 /          ;  return, line feed, and
 *                       %d127              ;  white space characters
 *
 *   obs-ctext       =   obs-NO-WS-CTL
 *
 *   obs-qtext       =   obs-NO-WS-CTL
 *
 *   obs-utext       =   %d0 / obs-NO-WS-CTL / VCHAR
 *
 *   obs-qp          =   "\" (%d0 / obs-NO-WS-CTL / LF / CR)
 *
 *   obs-body        =   *((*LF *CR *((%d0 / text) *LF *CR)) / CRLF)
 *
 *   obs-unstruct    =   *((*LF *CR *(obs-utext *LF *CR)) / FWS)
 *
 *   obs-phrase      =   word *(word / "." / CFWS)
 *
 *   obs-phrase-list =   [phrase / CFWS] *("," [phrase / CFWS])
 *
 *   obs-FWS         =   1*WSP *(CRLF 1*WSP)
 *
 *   obs-day-of-week =   [CFWS] day-name [CFWS]
 *
 *   obs-day         =   [CFWS] 1*2DIGIT [CFWS]
 *
 *   obs-year        =   [CFWS] 2*DIGIT [CFWS]
 *
 *   obs-hour        =   [CFWS] 2DIGIT [CFWS]
 *
 *   obs-minute      =   [CFWS] 2DIGIT [CFWS]
 *
 *   obs-second      =   [CFWS] 2DIGIT [CFWS]
 *
 *   obs-zone        =   "UT" / "GMT" /     ; Universal Time
 *                                          ; North American UT
 *                                          ; offsets
 *                       "EST" / "EDT" /    ; Eastern:  - 5/ - 4
 *                       "CST" / "CDT" /    ; Central:  - 6/ - 5
 *                       "MST" / "MDT" /    ; Mountain: - 7/ - 6
 *                       "PST" / "PDT" /    ; Pacific:  - 8/ - 7
 *                                          ;
 *
 *   obs-angle-addr  =   [CFWS] "<" obs-route addr-spec ">" [CFWS]
 *
 *   obs-route       =   obs-domain-list ":"
 *
 *   obs-domain-list =   *(CFWS / ",") "@" domain
 *                       *("," [CFWS] ["@" domain])
 *
 *   obs-mbox-list   =   *([CFWS] ",") mailbox *("," [mailbox / CFWS])
 *
 *   obs-addr-list   =   *([CFWS] ",") address *("," [address / CFWS])
 *
 *   obs-group-list  =   1*([CFWS] ",") [CFWS]
 *
 *   obs-local-part  =   word *("." word)
 *
 *   obs-domain      =   atom *("." atom)
 *
 *   obs-dtext       =   obs-NO-WS-CTL / quoted-pair
 *
 *   obs-fields      =   *(obs-return /
 *                       obs-received /
 *                       obs-orig-date /
 *                       obs-from /
 *                       obs-sender /
 *                       obs-reply-to /
 *                       obs-to /
 *                       obs-cc /
 *                       obs-bcc /
 *                       obs-message-id /
 *                       obs-in-reply-to /
 *                       obs-references /
 *                       obs-subject /
 *                       obs-comments /
 *                       obs-keywords /
 *                       obs-resent-date /
 *                       obs-resent-from /
 *                       obs-resent-send /
 *                       obs-resent-rply /
 *                       obs-resent-to /
 *                       obs-resent-cc /
 *                       obs-resent-bcc /
 *                       obs-resent-mid /
 *                       obs-optional)
 *
 *   obs-orig-date   =   "Date" *WSP ":" date-time CRLF
 *
 *   obs-from        =   "From" *WSP ":" mailbox-list CRLF
 *
 *   obs-sender      =   "Sender" *WSP ":" mailbox CRLF
 *
 *   obs-reply-to    =   "Reply-To" *WSP ":" address-list CRLF
 *
 *   obs-to          =   "To" *WSP ":" address-list CRLF
 *
 *   obs-cc          =   "Cc" *WSP ":" address-list CRLF
 *
 *   obs-bcc         =   "Bcc" *WSP ":"
 *                       (address-list / (*([CFWS] ",") [CFWS])) CRLF
 *
 *   obs-message-id  =   "Message-ID" *WSP ":" msg-id CRLF
 *
 *   obs-in-reply-to =   "In-Reply-To" *WSP ":" *(phrase / msg-id) CRLF
 *
 *   obs-references  =   "References" *WSP ":" *(phrase / msg-id) CRLF
 *
 *   obs-id-left     =   local-part
 *
 *   obs-id-right    =   domain
 *
 *   obs-subject     =   "Subject" *WSP ":" unstructured CRLF
 *
 *   obs-comments    =   "Comments" *WSP ":" unstructured CRLF
 *
 *   obs-keywords    =   "Keywords" *WSP ":" obs-phrase-list CRLF
 *
 *   obs-resent-from =   "Resent-From" *WSP ":" mailbox-list CRLF
 *
 *   obs-resent-send =   "Resent-Sender" *WSP ":" mailbox CRLF
 *
 *   obs-resent-date =   "Resent-Date" *WSP ":" date-time CRLF
 *
 *   obs-resent-to   =   "Resent-To" *WSP ":" address-list CRLF
 *
 *   obs-resent-cc   =   "Resent-Cc" *WSP ":" address-list CRLF
 *
 *   obs-resent-bcc  =   "Resent-Bcc" *WSP ":"
 *                       (address-list / (*([CFWS] ",") [CFWS])) CRLF
 *
 *   obs-resent-mid  =   "Resent-Message-ID" *WSP ":" msg-id CRLF
 *
 *   obs-resent-rply =   "Resent-Reply-To" *WSP ":" address-list CRLF
 *
 *   obs-return      =   "Return-Path" *WSP ":" path CRLF
 *
 *   obs-received    =   "Received" *WSP ":" *received-token CRLF
 *
 *   obs-optional    =   field-name *WSP ":" unstructured CRLF
 *
 */

function UnMHTMIMEPart () {
  this.fieldList = [];
  this.fields = {};
}
UnMHTMIMEPart.prototype = {
  fieldList : null, /* Array  フィールドの配列
                     *   [[String フィールド名, String フィールド値], ...]
                     *   **encode 時に必要** */
  fields : null,    /* Object  フィールドのマップ
                     *   解析用 (フィールド名は小文字)
                     *   <String フィールド名, String フィールド値>
                     *   **encode 時に必要** */
  
  /* Content-Type: (content[RFC2045]) */
  contentTypeMap : null, /* UnMHTMIMEParam  解析結果 */
  contentType : "",      /* String  type[RFC2045] */
  contentSubType : "",   /* String  subtype[RFC2045] */
  mimetype : "",         /* String  MIME-Type */
  charset : "",          /* String  charset[???] */
  isMultipart : false,   /* Boolean  マルチパートか
                          *   **encode 時に必要** */
  boundary : "",         /* String  boundary[RFC2046]
                          *   **encode 時に必要** */
  start : "",            /* String  "start"[RFC2387] */
  startMimetype : "",    /* String  "type"[RFC2387] */
  
  /* Content-Location: (content-location[RFC2557]) */
  contentLocation : "", /* String  URI[RFC2557] */
  
  /* Content-Disposition: (disposition[RFC2138]) */
  contentDispositionMap : null,     /* UnMHTMIMEParam  解析結果 */
  contentDispositionType : "",      /* String  disposition-type[RFC2138] */
  contentDispositionFilename : "",  /* String  filename-parm[RFC2138] */
  
  /* Content-ID: (id[RFC2045]) */
  contentID : "", /* String  msg-id[RFC2045] */
  
  /* Content-Transfer-Encoding: (encoding[RFC2045]) */
  contentTransferEncoding : "", /* String  mechanism[RFC2045]
                                 *   **encode 時に必要** */
  
  body : "", /* String  body[RFC5322]
              *   **encode 時に必要** */
  
  parts : null, /* Array  マルチパート
                 *   [UnMHTMIMEPart  body-part[RFC2046], ...]
                 *   **encode 時に必要** */
  
  eParam : null, /* UnMHTExtractParam  展開で使用するパラメータ */
  sParam : null  /* UnMHTSaveParam  保存で使用するパラメータ */
};

/**
 * MIME 関係ユーティリティ関数
 */
let UnMHTMIMEUtil = {
  /**
   * quoted-printable[RFC2045] にエンコードする
   *
   * @param  String text
   *         エンコードする文字列
   * @param  Boolean wrap
   *         76 文字で折り返す
   * @return String
   *         エンコードした文字列
   */
  encodeQ : function (text, wrap) {
    text
    = text.replace
    (/[^\t\r\n \x21-\x3C\x3E-\x7E]/g,
     function (matched) {
      let hexString = "0" + matched.charCodeAt (0).toString (16).toUpperCase ();
      return "=" + hexString.slice (-2);
    });
    
    if (!wrap) {
      return text;
    }
    
    return (text + "\r\n").replace
    (/([^\r\n]{0,73})(?:(=..|[ \t]|[^\r\n=]{2}[^\t\r\n =])($|\r?\n|\r)|([^\t\r\n =]{0,2})(\r?\n?))/g,
     function (matched, aMain,
               aLastThree, aLastThreeEOL,
               aLastTwo, aLastTwoEOL) {
      let line = aMain + aLastThree + aLastTwo;
      if (aLastThreeEOL || aLastTwoEOL) {
        if (line) {
          let lastChar = line [line.length - 1];
          if (lastChar === " "
              || lastChar === "\t") {
            line
              = line.slice (0, -1)
              + (lastChar === " " ? "=20" : "=09");
          }
        }
        return line + "\r\n";
      }
      return line ? (line + "=\r\n") : "";
    }).slice (0, -2);
  },
    
  /**
   * quoted-printable[RFC2045] をデコードする
   *
   * @param  String text
   *         デコードする文字列
   * @return String
   *         デコードした文字列
   */
  decodeQ : function (text) {
    return text.replace
    (/=(?:\r?\n|\r|([A-Fa-f0-9]{2}))/g,
     function (matched, octet) {
      return (octet ? String.fromCharCode (parseInt (octet, 16)) : "");
    });
  },
  
  /**
   * encoded-word にエンコードする
   *
   * @param  String text
   *         エンコードする文字列 (Unicode)
   * @return String
   *         エンコードした文字列
   */
  encodeWord : function (charset, type, text) {
    let converter
    = Cc ["@mozilla.org/intl/scriptableunicodeconverter"]
    .getService (Ci.nsIScriptableUnicodeConverter);
    converter.charset = charset;
    text = converter.ConvertFromUnicode (text);
    
    if (type == "B") {
      text = this.encodeBase64 (text, false);
    }
    else if (type == "Q") {
      text
      = this.encodeQ (text, false)
      .replace (/\?/g, "=3F")
      .replace (/ /g, "=20");
    }
    
    return ("=?" + charset + "?" + type + "?" + text + "?=");
  },
  
  /**
   * encoded-word をデコードする
   *
   * @param  String text
   *         デコードする文字列
   * @return String
   *         デコードした文字列 (Unicode)
   */
  decodeWord : function (text) {
    return  text.replace
    (/=\?([^\?]+)\?([^\?])\?([^\?]+)\?=/g,
     function (matched, charset, type, v) {
      if (type == "Q" || type == "q") {
        v = this.decodeQ (v);
      }
      else if (type == "B" || type == "b") {
        v = this.decodeBase64 (v);
      }
      
      return UnMHTUnicodeConverter.convertToUnicode (v, charset);
    }.bind (this));
  },
  
  /**
   * boundary[RFC2046] を作成する
   *
   * @return String
   *         バウンダリ文字列
   */
  createUniqueBoundary : function () {
    let randomHex = function () {
      return Math.floor (Math.random () * 15).toString (16).toUpperCase ();
    };
    
    return ["----=_Part_"].concat
    ([randomHex() for (i in UnMHTArrayUtil.range (1, 8))],
     "_",
     [randomHex() for (i in UnMHTArrayUtil.range (1, 8))],
     ".",
     (new Date ()).getTime ()).join ("");
  },
  
  /**
   * 日時を date-time[RFC5322] 形式にフォーマットする
   *
   * @param  Date date
   *         現在の日時
   * @return String
   *         フォーマットされた日時
   */
  formatDate : function (date) {
    let weekDay = [
      "Sun",
      "Mon",
      "Tue",
      "Wed",
      "Thu",
      "Fri",
      "Sat"
      ];
    let month = [
      "Jan",
      "Feb",
      "Mar",
      "Apr",
      "May",
      "Jun",
      "Jul",
      "Aug",
      "Sep",
      "Oct",
      "Nov",
      "Dec"
      ];
    
    let pad = function (x, n) {
      return ("0000" + n).slice (-x);
    };
    
    let sign = "-";
    let zone = date.getTimezoneOffset () * 100 / 60;
    if (zone < 0) {
      zone = -zone;
      sign = "+";
    }
    
    return (weekDay [date.getDay ()] + ", "
            + pad (2, date.getDate ()) + " "
            + month [date.getMonth ()] + " "
            + (date.getFullYear ()) + " "
            + pad (2, date.getHours ()) + ":"
            + pad (2, date.getMinutes ()) + ":"
            + pad (2, date.getSeconds ()) + " "
            + sign + pad (4, zone));
  },
  
  /**
   * 拡張子に対応した MIME-Type を返す
   *
   * @param  String ext
   *         拡張子
   * @return String
   *         対応した MIME-Type
   */
  getTypeFromExtension : function (ext) {
    /* システム設定が特殊場合に供えて既知のものは先にチェック */
    if (ext.match (/^[sx]?html?$/i)) {
      /* XHTML も text/html として扱う */
      return "text/html";
    }
    else if (ext.match (/^css$/i)) {
      return "text/css";
    }
    else if (ext.match (/^mht(ml)?$/i)) {
      return "message/rfc822";
    }
    else if (ext.match (/^eml$/i)) {
      return "message/rfc822";
    }
    else if (ext.match (/^js$/i)) {
      return "text/javascript";
    }
    else if (ext.match (/^txt$/i)) {
      return "text/plain";
    }
    else if (ext.match (/^jpe?g$/i)) {
      return "image/jpeg";
    }
    else if (ext.match (/^gif$/i)) {
      return "image/gif";
    }
    else if (ext.match (/^png$/i)) {
      return "image/png";
    }
    else if (ext.match (/^bmp$/i)) {
      return "image/bmp";
    }
    else if (ext.match (/^ico$/i)) {
      return "image/x-icon";
    }
    else if (ext.match (/^svg$/i)) {
      return "image/svg+xml";
    }
    else if (ext.match (/^swf$/i)) {
      return "application/swf";
    }
    else if (ext.match (/^jar$/i)) {
      return "application/java-archive";
    }
    else if (ext.match (/^class$/i)) {
      return "application/octet-stream";
    }
    else if (ext) {
      /* 未知なのでシステム設定を返す */
      let mime = Cc ["@mozilla.org/mime;1"].getService (Ci.nsIMIMEService);
      
      try {
        return mime.getTypeFromExtension (ext);
      }
      catch (e) {
      }
    }
    
    return "application/octet-stream";
  },
  
  /**
   * Base64 エンコードする
   *
   * @param  String text
   *         エンコードする文字列
   * @param  Boolean wrap
   *         76 文字で折り返す
   * @return String
   *         エンコードされた文字列
   */
  encodeBase64 : function (text, wrap) {
    text = btoa (text);
    if (!wrap) {
      return text;
    }
    
    return text.replace (/.{76}/g, "$&\r\n");
  },

  /**
   * Base64 デコードする
   * 改行を含むものや破損したものにも対応
   *
   * @param  String text
   *         デコードする文字列
   * @return String
   *         デコードされた文字列
   */
  decodeBase64 : function (text) {
    text = text.replace (/[^A-Za-z0-9+\/=]+/g, "");
    return atob (text.substr (0, text.length - text.length % 4));
  },
  
  /**
   * message [RFC5322] にエンコードする
   *
   * @param  UnMHTMIMEPart part
   *         トップレベルのパート
   * @param  nsIOutputStream stream
   *         メッセージ出力用ストリーム
   *         改行コードは CR LF である
   */
  encodeMessage : function (part, stream) {
    this.encodeFields (part, stream);
    
    let data = "\r\n";
    stream.write (data, data.length);

    if (part.isMultipart) {
      part.parts.forEach (function (p) {
          let data = "--" + part.boundary + "\r\n";
          stream.write (data, data.length);
          this.encodeMessage (p, stream);
          data = "\r\n";
          stream.write (data, data.length);
        }.bind (this));
      
      let data = "--" + part.boundary + "--\r\n";
      stream.write (data, data.length);
      
      return;
    }
    
    if (part.contentTransferEncoding == "quoted-printable") {
      data = this.encodeQ (part.body, true);
    }
    else if (part.contentTransferEncoding == "base64") {
      data = this.encodeBase64 (part.body, true) + "\r\n";
    }
    else {
      data = part.body;
    }
    stream.write (data, data.length);
  },
  
  /**
   * fields[RFC5322] にエンコードする
   *
   * @param  UnMHTMIMEPart part
   *         トップレベルのパート
   * @param  nsIOutputStream stream
   *         メッセージ出力用ストリーム
   *         改行コードは CR LF である
   */
  encodeFields : function (part, stream) {
    let order = [
      "from",
      "to",
      "subject",
      "date",
      "mime-version",
      "content-location",
      "content-base",
      "content-id",
      "content-disposition",
      "content-type",
      "content-transfer-encoding"
      ];
    let PLAIN        = 1;
    let ENCODED_WORD = 2;
    let PARAMETER    = 3;
    let UNICODE      = 4;
    let types = {
      "from":                      PLAIN,
      "to":                        PLAIN,
      "subject":                   ENCODED_WORD,
      "date":                      PLAIN,
      "mime-version":              PLAIN,
      "content-type":              PARAMETER,
      "content-transfer-encoding": PLAIN,
      "content-location":          UNICODE,
      "content-base":              UNICODE,
      "content-disposition":       PARAMETER,
      "content-id":                PLAIN
    };
    
    part.fieldList.sort (function ([name1,], [name2,]) {
        let p1 = order.indexOf (name1.toLowerCase ());
        let p2 = order.indexOf (name2.toLowerCase ());
        if (p1 >= 0) {
          if (p2 >= 0) {
            return p1 - p2;
          }
          else {
            return -1;
          }
        }
        else {
          if (p2 >= 0) {
            return 1;
          }
          else {
            return (name1 < name2) ? -1 : ((name1 > name2) ? 1 : 0);
          }
        }
      })
    .map (function ([name, value]) {
        let type = PLAIN;
        let lname = name.toLowerCase ();
        if (lname in types) {
          type = types [lname];
        };
          
        if (type == UNICODE) {
          try {
            value = UnMHTUnicodeConverter.convertFromUnicode (value, "utf-8");
          }
          catch (e) {
          }
        }
        else if (type == ENCODED_WORD) {
          if (!value.match (/^[\x20-\x7E]*$/)) {
            value = this.encodeWord ("utf-8", "B", value);
          }
        }
        else if (type == PARAMETER) {
          /* Emacs がおかしくなるので一箇所だけダブルクォーテーションを
           * エスケープシーケンスで書く */
          value = value.replace
            (/(([^\t\r\n \*\'%\(\)<>@,;:\\\"\/\[\]\?=]+)(?:\*([0-9]+))?(\*)?=(?:([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \*\'\(\)<>@,;:\\\"\/\[\]\?=]+)|\"((?:\\.|[^\"])*)\"|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+))|([^\t\r\n \(\)<>@,;:\\\x22\/\[\]\?=]+)\/([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+));[ \t]*/g,
             function (matched, param) {
              return param + ";\r\n\t";
            });
        }
          
        return name + ": " + value + "\r\n";
      }.bind (this))
    .forEach (function (data) stream.write (data, data.length));
  },
  
  /**
   * message [RFC5322] をデコードする
   *
   * @param  String text
   *         メッセージ
   *         改行コードは CR LF でなければならない
   * @return UnMHTMIMEPart
   *         トップレベルのパート
   *         データが不正ならば null
   */
  decodeMessage : function (text) {
    let fields = "";
    let body = "";
    
    let pos = text.indexOf ("\r\n\r\n");
    if (pos != -1) {
      fields = text.substr (0, pos + 2);
      body = text.substr (pos + 4);
    }
    else {
      fields = text;
    }
    
    let part = this.decodeFields (fields);
    if (!part) {
      return null;
    }

    if (part.isMultipart) {
      part.parts = this.decodeMultipart (body, part.boundary);
    }
    else {
      if (part.contentTransferEncoding == "quoted-printable") {
        part.body = this.decodeQ (body);
      }
      else if (part.contentTransferEncoding == "base64") {
        part.body = this.decodeBase64 (body);
      }
      else {
        part.body = body;
      }
    }
    
    return part;
  },
  
  /**
   * multipart-body[RFC2046] をデコードする
   *
   * @param  String body
   *         multipart なメッセージボディ
   *         改行コードは CR LF でなければならない
   * @param  String boundary
   *         バウンダリ文字列
   * @return Array
   *         パートの配列
   *           [UnMHTMIMEPart, ...]
   */
  decodeMultipart : function (body, boundary) {
    let b = boundary
    .replace (/([\\\[\]\(\)\^\,\.\{\}\|\?\!\-\*\+\$])/g, "\\$1");
    
    let datas = ("\r\n" + body)
    .split (new RegExp ("\r\n--" + b + "(?:--)?(?:\r\n)?"));
    return datas.splice (1, datas.length - 2)
    .map ((function (data) this.decodeMessage (data)).bind (this))
    .filter (function (part) part != null);
  },
  
  /**
   * fields[RFC5322] をデコードする
   *
   * @param  String fieldsString
   *         フィールド文字列
   *         改行コードは CR LF でなければならない
   * @return UnMHTMIMEPart
   *         パート
   *         データが不正ならば null
   */
  decodeFields : function (fieldsString) {
    let part = new UnMHTMIMEPart ();
    
    let remain
    = fieldsString.replace
    (/([\x21-\x39\x3B-\x7E]+)[ \t]*:[ \t]*([^\r\n]*)\r\n((?:[\t ]+[^\r\n]*\r\n)*)/g,
     function (matched, name, value, restLines) {
      restLines.replace
      (/[\t ]*([^\r\n]*)\r\n/g,
       function (matched, v) {
        value += v;
          
        return "";
      });
       
      this.addField (part, name, value);
      
      return "";
    }.bind (this));
    if (remain) {
      return null;
    }
    
    if ("content-type" in part.fields) {
      let value = part.fields ["content-type"];
      part.contentTypeMap = this.decodeParam (value);
      part.contentType    = part.contentTypeMap.type.toLowerCase ();
      part.contentSubType = part.contentTypeMap.subtype.toLowerCase ();
      part.mimetype       = part.contentType + "/" + part.contentSubType;
      if ("charset" in part.contentTypeMap.params) {
        part.charset  = part.contentTypeMap.params ["charset"];
      }
    }
    
    if ("content-location" in part.fields) {
      part.contentLocation
      = this.tryConvertToUnicode (part.fields ["content-location"]);
    }
    
    if ("content-disposition" in part.fields) {
      let value = part.fields ["content-disposition"];
      part.contentDispositionMap  = this.decodeParam (value);
      part.contentDispositionType
      = part.contentDispositionMap.type.toLowerCase ();
      if ("filename" in part.contentDispositionMap.params) {
        part.contentDispositionFilename
          = this.tryConvertToUnicode
          (part.contentDispositionMap.params ["filename"]);
      }
    }
    
    if ("content-id" in part.fields) {
      let value = part.fields ["content-id"];
      part.contentID = value.replace (/^<([^>]+)>$/, "$1");
    }
    
    if (part.contentType == "multipart") {
      part.isMultipart = true;
      if ("boundary" in part.contentTypeMap.params) {
        part.boundary = part.contentTypeMap.params ["boundary"];
      }
      else {
        return null;
      }
      
      if ("start" in part.contentTypeMap.params) {
        let value = part.contentTypeMap.params ["start"];
        part.start = value.replace (/^<([^>]+)>$/, "$1");
      }
      
      if ("type" in part.contentTypeMap.params) {
        part.startMimetype = part.contentTypeMap.params ["type"];
      }
    }
    else {
      if ("content-transfer-encoding" in part.fields) {
        let value = part.fields ["content-transfer-encoding"];
        part.contentTransferEncoding = value.toLowerCase ();
      }
    }
    
    return part;
  },
  
  /**
   * 開始ページになるパートを探す
   *
   * @param  UnMHTMIMEPart part
   *         走査中のパート
   * @return UnMHTMIMEPart
   *         開始ページになるパート
   *         適切なパートが無ければ null
   */
  findStartPart : function (part) {
    if (part.isMultipart) {
      if (part.contentSubType == "mixed") {
        return part;
      }
      
      if (part.start) {
        let starts
        = part.parts.filter (function (p) p.contentID == part.start);
        if (starts.length) {
          return this.findStartPart (starts [0]);
        }
      }
      
      if (part.contentSubType == "related") {
        let starts
        = part.parts.filter (function (p)
                             p.mimetype == part.startMimetype);
        if (starts.length) {
          return this.findStartPart (starts [0]);
        }
      }
      
      if (part.contentSubType == "alternative") {
        let starts = part.parts.filter (function (p)
                                        p.mimetype == "text/html");
        if (starts.length) {
          return starts [0];
        }
        
        starts
        = part.parts.filter (function (p) {
            if (p.isMultipart) {
              let startPart = this.findStartPart (p);
              if (startPart && startPart.mimetype == "text/html") {
                return true;
              }
            }
          }.bind (this));
        if (starts.length) {
          return this.findStartPart (starts [0]);
        }
      }
      
      if (part.parts.length) {
        return this.findStartPart (part.parts [0]);
      }
      
      return null;
    }
    
    return part;
  },
  
  /**
   * content[RFC2045] や disposition[RFC2183] の値を
   * parameter[RFC2231] に従ってデコードする
   * 
   * ただし disposition-type[RFC2183] や type/subtype[RFC2045] の位置が
   * 先頭でない事例が存在するのでこれらのデコードも同時に行う
   *
   * @param  String value
   *         contentRFC2045] や disposition[RFC2183] の値
   * @return UnMHTMIMEParam
   *         {
   *           type:    String  disposition-type[RFC2183]/type[RFC2045],
   *           subtype: String  subtype[RFC2045],
   *           params:  Object パラメータのマップ
   *             <String パラメータ名, String パラメータ値>
   *         }
   */
  decodeParam : function (value) {
    let map = {};
    let params = {};
    let type = "";
    let subtype = "";
    
    value.replace
    /*
     * ( 1*attribute-char [section] ["*"] "="
     *   ( ( [token] "'" [token] "'" 1*<attribute-char and %> ) /
     *     quoted-string / token ) ) /
     * ( token "/" token ) /
     * token
     */
    (/([^\t\r\n \*\'%\(\)<>@,;:\\\"\/\[\]\?=]+)(?:\*([0-9]+))?(\*)?=(?:([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \*\'\(\)<>@,;:\\\"\/\[\]\?=]+)|\"((?:\\.|[^\"])*)\"|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+))|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)\/([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)/g,
     function (matched, name, section, extended,
               charset, lang, evalue,
               qvalue, value,
               ctype, csubtype, dtype) {
      /*
        dump ("@@@" + matched + "\n");
        if (name) dump ("@@@  name: " + name + "\n");
        if (section) dump ("@@@  section: " + section + "\n");
        if (extended) dump ("@@@  extended: " + extended + "\n");
        if (charset) dump ("@@@  charset: " + charset + "\n");
        if (lang) dump ("@@@  lang: " + lang + "\n");
        if (evalue) dump ("@@@  evalue: " + evalue + "\n");
        if (qvalue) dump ("@@@  qvalue: " + qvalue + "\n");
        if (value) dump ("@@@  value: " + value + "\n");
        if (ctype) dump ("@@@  ctype: " + ctype + "\n");
        if (csubtype) dump ("@@@  csubtype: " + csubtype + "\n");
        if (dtype) dump ("@@@  dtype: " + dtype + "\n");
      */
      
      if (dtype && !type) {
        type = dtype;
      }
      else if (ctype && !type) {
        type = ctype;
        subtype = csubtype;
      }
      else if (name) {
        if (!value) {
          value = qvalue;
        }
        
        if (!(name in map)) {
          map [name] = {
            values : {},
            charset : "",
            lang : ""
          };
        }
        
        if (!section) {
          section = "0";
        }
        
        if (extended && evalue) {
          map [name].values [section] = { v: evalue, e: true };
          map [name].charset = charset;
          map [name].lang = lang;
        }
        else {
          /* 実際にはここに encoded-word は使えないが使用事例がある */
          value = this.decodeWord (value);
          
          map [name].values [section] = { v: value, e: extended != "" };
        }
      }
    }.bind (this));
    
    for (let [name, param] in Iterator (map)) {
      params [name]
        = [x for (x in Iterator (param.values))]
        .sort (function (a, b) parseInt (a [0], 10) - parseInt (b [0], 10))
        .map (function ([n, v]) {
            if (v.e) {
              if (param.charset) {
                return UnMHTUnicodeConverter.convertToUnicode
                (unescape (v.v), param.charset);
              }
              else {
                return unescape (v.v);
              }
            }
            else {
              return v.v;
            }
          })
        .join ("");
    }
    
    return {
      type: type,
      subtype: subtype,
      params: params
    };
  },

  /**
   * パートにフィールドを追加する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  String name
   *         フィールド名
   * @param  String value
   *         フィールド値
   */
  addField : function (part, name, value) {
    part.fieldList.push ([name, value]);
    part.fields [name.toLowerCase ()] = value;
  },
  
  /**
   * Unicode への変換を試みる
   *
   * @param  String value
   *         対象の文字列
   * @return String
   *         成功したら変換した文字列
   *         失敗したら元の文字列
   */
  tryConvertToUnicode : function (value) {
    try {
      return UnMHTUnicodeConverter.convertToUnicode (value, "utf-8");
    }
    catch (e) {
    }
    try {
      return UnMHTUnicodeConverter.convertToUnicode (value, "shift_jis");
    }
    catch (e) {
    }
    return value;
  },
  
  /**
   * パートのデータを複製する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @return UnMHTMIMEPart
   *         複製したパート
   */
  clonePart : function (part) {
    let newPart = new UnMHTMIMEPart ();
    
    for (let [name, value] in Iterator (part)) {
      if (name == "fieldList") {
        newPart.fieldList = value.map (function ([name, value]) [name, value]);
      }
      else if (name == "fields") {
        for (let [fName, fValue] in Iterator (value)) {
          newPart.fields [fName] = fValue;
        }
      }
      else if (name == "contentTypeMap"
               || name == "contentDispositionMap") {
        let newParams = {};
        for (let [pName, pValue] in Iterator (value.params)) {
          newParams [pName] = pValue;
        }
        
        newPart [name] = {
          type : value.type,
          subtype : value.subtype,
          params : newParams
        };
      }
      else if (name == "parts") {
        newPart.parts
          = value.map ((function (p) this.clonePart (p)).bind (this));
      }
      else if (name == "eParam"
               || name == "sParam") {
        /* 複製しない */
      }
      else {
        newPart [name] = value;
      }
    }
    
    return newPart;
  },
  
  /**
   * パートのデータを削除する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   */
  destructPart : function (part) {
    if (part.parts) {
      part.parts.forEach (function (p) {
          this.destructPart (p);
        }.bind (this));
      part.parts = null;
    }
    part.fieldList = null;
    part.fields = null;
    part.contentTypeMap = null;
    part.contentDispositionMap = null;
    part.body = "";
  }
};
