/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMail"];

/**
 * メール送信
 */
let UnMHTMail = {
  thunderbird : UnMHTPref.thunderbirdArg,
  seamonkey : UnMHTPref.seamonkeyArg,
  macmail : UnMHTPref.macmailArg,
  outlook : "",
  winmail : "",
  
  /**
   * スペース区切りの引数を引数配列に変換する
   *
   * @param  String argString
   *         スペース区切りの引数
   * @return Array
   *         引数配列
   */
  argToArray : function (argString) {
    let STATE_NORMAL          = 0;
    let STATE_ESCAPED         = 1;
    let STATE_DQUOTED         = 10;
    let STATE_DQUOTED_ESCAPED = 11;
    let STATE_QUOTED          = 20;
    let STATE_QUOTED_ESCAPED  = 21;
    
    let args = [];
    let currentArg = "";
    let state = STATE_NORMAL;
    
    Array.forEach
    (argString,
     function (c) {
      if (state == STATE_NORMAL) {
        if (c == "\\") {
          state = STATE_ESCAPED;
        }
        else if (currentArg == ""
            && c == "\"") {
          state = STATE_DQUOTED;
        }
        else if (currentArg == ""
            && c == "\'") {
          state = STATE_QUOTED;
        }
        else if (c == " ") {
          if (currentArg != "") {
            args.push (currentArg);
          }
          currentArg = "";
        }
        else {
          currentArg += c;
        }
      }
      else if (state == STATE_ESCAPED) {
        currentArg += c;
      }
      else if (state == STATE_DQUOTED) {
        if (c == "\"") {
          state = STATE_NORMAL;
        }
        else if (c == "\\") {
          state = STATE_DQUOTED_ESCAPED;
        }
        else {
          currentArg += c;
        }
      }
      else if (state == STATE_DQUOTED_ESCAPED) {
        currentArg += c;
        state = STATE_DQUOTED;
      }
      else if (state == STATE_QUOTED) {
        if (c == "\'") {
          state = STATE_NORMAL;
        }
        else if (c == "\\") {
          state = STATE_QUOTED_ESCAPED;
        }
        else {
          currentArg += c;
        }
      }
      else if (state == STATE_QUOTED_ESCAPED) {
        currentArg += c;
        state = STATE_QUOTED_ESCAPE;
      }
    });
    if (currentArg != "") {
      args.push (currentArg);
    }
    
    return args;
  },
  
  /**
   * メール送信アプリケーションを起動する
   *
   * @param  nsIFile targetFile
   *         添付ファイル
   * @param  String title
   *         ページのタイトル
   * @param  Function debugFunc
   *         デバッグ情報を受け取る関数
   * @return Boolean
   *         成功したか
   */
  launch : function (targetFile, title, debugFunc) {
    let apptype = UnMHTPref.getPref ("save.mail.app.type");
    
    if (apptype == "seamonkey"
        && UnMHTAppInfo.isSeaMonkey ()) {
      return this.openComposeWindow (targetFile, title);
    }
    
    let app;
    let argtype = UnMHTPref.getPref ("save.mail.arg.type");
    let arg;
    
    if (argtype == "thunderbird") {
      arg = this.thunderbird;
    }
    if (argtype == "seamonkey") {
      arg = this.seamonkey;
    }
    else if (argtype == "macmail") {
      arg = this.macmail;
    }
    else if (argtype == "outlook") {
      arg = this.outlook;
    }
    else if (argtype == "winmail") {
      arg = this.winmail;
    }
    else {
      arg = UnMHTPref.getPref ("save.mail.arg.other");
    }
    
    let getThunderbirdPath = function () {
      let runtime
      = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    
      let list = [];
      if (runtime.OS == "Darwin") {
        list = [
          "/Applications/Thunderbird.app/Contents/MacOS/thunderbird-bin"
          ];
      }
      else if (runtime.OS == "WINNT") {
        list = [
          "C:\\Program Files\\Mozilla Thunderbird\\thunderbird.exe",
          "C:\\Program Files\\thunderbird\\thunderbird.exe"
          ];
      }
    
      let [found] = list.filter (function (path) UnMHTFileUtil.isExist (path));
    
      return found;
    };
    
    let getSeamonkeyPath = function () {
      let runtime
      = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    
      let list = [];
      if (runtime.OS == "Darwin") {
        list = [
          "/Applications/SeaMonkey.app/Contents/MacOS/seamonkey-bin"
          ];
      }
      else if (runtime.OS == "WINNT") {
        list = [
          "C:\\Program Files\\mozilla.org\\SeaMonkey\\seamonkey.exe",
          "C:\\Program Files\\SeaMonkey\\seamonkey.exe"
          ];
      }
      
      let [found] = list.filter (function (path) UnMHTFileUtil.isExist (path));
    
      return found;
    };
    
    let macmailFile = null;
    
    if (apptype == "thunderbird") {
      app = getThunderbirdPath ();
    }
    else if (apptype == "seamonkey") {
      app = getSeamonkeyPath ();
    }
    else if (apptype == "macmail") {
      let path
      = UnMHTFileUtil.cacheDirectory + UnMHTFileUtil.separator
      + "macmailscript";
      
      let data
      = "#!/usr/bin/osascript" + "\n"
      + "on run argv" + "\n"
      + "  tell application \"Mail\"" + "\n"
      + "    tell content of (make new outgoing message with properties {visible:true, subject:(item 1 of argv)})" + "\n"
      + "      make new attachment with properties {file name:(item 2 of argv)} at after the last paragraph" + "\n"
      + "    end tell" + "\n"
      + "    activate" + "\n"
      + "  end tell" + "\n"
      + "end run" + "\n";
      
      let macmailFile = UnMHTFileUtil.initFile (path);
      let fstream = UnMHTFileUtil.openOutputStream (macmailFile, "0755");
      fstream.write (data, data.length);
      fstream.close ();
      
      app = macmailFile.path;
    }
    else {
      app = UnMHTPref.getPref ("save.mail.app");
      if (!app) {
        app = getThunderbirdPath ();
        if (app) {
          arg = this.thunderbird;
        }
      }
    }
    
    if (!app) {
      return false;
    }
    
    let targetPath
    = UnMHTUnicodeConverter.convertFromUnicode (targetFile.path, "utf-8");
    
    let targetURI
    = UnMHTFileUtil.getURLSpecFromFile (targetFile)
    .replace (/,/g, "%2C");
    
    arg = arg
    .replace (/%FILE%/g, targetPath.replace (/\\/g, "\\\\"))
    .replace (/%URI%/g, targetURI.replace (/\\/g, "\\\\"))
    .replace (/%TITLE%/g, title
              .replace (/\\/g, "\\\\")
              .replace (/[\'\"]/g, ""))
    .replace (/%TITLE_NOSP%/g, title
              .replace (/\\/g, "\\\\")
              .replace (/[\'\"]/g, "")
              .replace (/[ \t\r\n]/g, "_"));
    let args = this.argToArray (arg);
    
    let file = UnMHTFileUtil.initFile (app);
    if (!file) {
      return false;
    }
    
    if (file.isDirectory ()) {
      if (app.match (/\.app$/)) {
        file = UnMHTFileUtil.initFile (app + "/Contents/MacOS/thunderbird-bin");
        if (!file) {
          file = UnMHTFileUtil.initFile (app + "/Contents/MacOS/seamonkey-bin");
        }
        if (!file) {
          return false;
        }
      }
    }
    
    try {
      let process
      = Cc ["@mozilla.org/process/util;1"].createInstance (Ci.nsIProcess);
      process.init (file);
      
      if (UnMHTPref.getPref ("save.mail.debug")) {
        debugFunc (file.path, arg, args);
      }
      
      process.run (false, args, args.length);
    }
    catch (e) {
      return false;
    }
    
    if (macmailFile) {
      macmailFile.remove ();
    }
    
    return true;
  },
  
  /**
   * SeaMonkey でメール作成画面を開く
   *
   * @param  nsIFile targetFile
   *         添付ファイル
   * @param  String title
   *         ページのタイトル
   * @return Boolean
   *         成功したか
   */
  openComposeWindow : function (targetFile, title) {
    let params
    = Cc ["@mozilla.org/messengercompose/composeparams;1"]
    .createInstance (Ci.nsIMsgComposeParams);
    
    params.composeFields
    = Cc ["@mozilla.org/messengercompose/composefields;1"]
    .createInstance (Ci.nsIMsgCompFields);
    
    params.composeFields.subject = title;
    
    let attachmentData
    = Cc ["@mozilla.org/messengercompose/attachment;1"]
    .createInstance (Ci.nsIMsgAttachment);
    
    let targetURI = UnMHTFileUtil.getURLSpecFromFile (targetFile)
    
    attachmentData.url = targetURI;
    
    params.composeFields.addAttachment (attachmentData);
    
    let composeService
    = Cc ["@mozilla.org/messengercompose;1"]
    .getService (Ci.nsIMsgComposeService);
    
    try {
      params.identity = composeService.defaultIdentity;
    }
    catch (e) {
      params.identity = null;
    }
    
    composeService.OpenComposeWindowWithParams (null, params);
    
    return true;
  }
};
