/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;
const Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTFormatter.jsm");
Cu.import ("resource://unmht/modules/UnMHTHTMLUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTProtocolHandler"];

/**
 * unmht プロトコルからチャネルを生成する
 *
 * @implements nsIProtocolHandler
 */
function UnMHTProtocolHandler () {
}
UnMHTProtocolHandler.prototype = {
  classID:          Components.ID ("{d605b959-0688-4846-887c-f6658725f7f1}"),
  contractID:       "@mozilla.org/network/protocol;1?name=unmht",
  classDescription: "UnMHT Protocol Handler JS Component",
  QueryInterface:   XPCOMUtils.generateQI ([Ci.nsIProtocolHandler]),
  
  scheme : "unmht", /* String  プロトコルスキーム */
  defaultPort : -1, /* Number  ポート番号
                     *   ネットワークベースではないので -1 */
    
  /* Number プロトコルの詳細
   *   file と同様のプロトコル */
  protocolFlags: (Ci.nsIProtocolHandler.URI_NOAUTH
                  | Ci.nsIProtocolHandler.URI_LOADABLE_BY_ANYONE),
  
  /**
   * ブラックリストのポートを上書きするか
   *   nsIProtocolHandler.allowPort
   *
   * @param  Number port
   *         ポート番号
   * @param  String scheme
   *         プロトコルスキーム
   * @return Boolean
   *         上書きするか
   */
  allowPort: function (port, scheme) {
    return false;
  },
    
  /**
   * URI を作成する
   *   nsIProtocolHandler.newURI
   *
   * @param  String spec
   *         対象の URI
   * @param  String charset
   *         対象の文字コード
   * @param  nsIURI baseURI
   *         読み込み元の URI
   * @return nsIURI
   *         作成した URI
   */
  newURI : function (spec, charset, baseURI) {
    let uri
    = Cc ["@mozilla.org/network/standard-url;1"].createInstance (Ci.nsIURI);
    
    if (baseURI) {
      if (UnMHTScheme.isUnMHTURI (baseURI.spec)) {
        let resolved = UnMHTExtractManager.resolveURI (baseURI.spec, spec);
        if (resolved) {
          uri.spec = resolved;
          return uri;
        }
      }
      
      uri.spec = baseURI.resolve (spec);
    }
    else {
      uri.spec = spec;
    }
    
    return uri;
  },
    
  /**
   * チャネルを作成する
   *   nsIProtocolHandler.newChannel
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  newChannel : function (uri) {
    let param = UnMHTScheme.getUnMHTURIParam (uri.spec);
    
    if (!("original" in param
          && param.original)) {
      /* 不正な URI */
      return this._createIllegalChannel (uri);
    }
    else if (param.scheme == "cache") {
      try {
      return this._createCacheChannel (uri);
      }
      catch (e) {
        Cu.reportError (e + "\n" + e.stack);
      }
    }
    else if (param.scheme == "cid"
             || param.scheme == "dummy") {
      return this._createEmptyChannel (uri);
    }
    else if (param.scheme == "error") {
      return this._createErrorChannel (uri);
    }
    else if (param.scheme == "refresh") {
      return this._createRefreshChannel
      (uri, UnMHTScheme.deRefreshURI (uri.spec));
    }
    else {
      /* mht 内のファイル */
      return this._createFileChannel (uri, param);
    }
        
    return this._createEmptyChannel (uri);
  },
    
  /**
   * mht 内のファイルのチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI (umht:// 形式)
   * @param  UnMHTURIParam param
   *         対象の URI の情報
   * @return nsIChannel
   *         作成したチャネル
   */
  _createFileChannel : function (uri, param) {
    let [extractor, part] = UnMHTExtractManager.getExtractorAndPart (uri.spec);
    if (!extractor) {
      try {
        /* URI が不正でないかチェック */
        Services.io.newURI (param.original, null, null);
      }
      catch (e) {
        return this._createIllegalChannel (uri);
      }
      
      return this._createRefreshChannel (uri, uri.spec); 
    }
    
    if (!part) {
      return this._createEmptyChannel (uri);
    }
    
    let pipe = Cc ["@mozilla.org/pipe;1"].createInstance (Ci.nsIPipe);
    pipe.init (false, false, part.eParam.content.length, 1, null);
    pipe.outputStream.write
    (part.eParam.content, part.eParam.content.length);
    pipe.outputStream.close ();
    
    return this._createInputStreamChannel (uri,
                                           part.eParam.charset,
                                           part.eParam.mimetype,
                                           pipe.inputStream,
                                           part.eParam.content.length);
  },
    
  /**
   * キャッシュ情報のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createCacheChannel : function (uri) {
    let tableRow = function (header, data) {
      return ("<tr>"
              + "<th>"
              + header
              + "</th>"
              + "<td>"
              + data
              + "</tr>");
    };
    let bundle = function (name) {
      return UnMHTHTMLUtil.escapeEntity (UnMHTStringBundle.get (name));
    };
    
    let bodies = [];
    let size = 0;
    
    let extractors = UnMHTExtractManager.getExtractors ();
    
    extractors.forEach (function (extractor) {
        let body = [];
        size += extractor.size;
        
        let param = UnMHTScheme.getUnMHTURIParam (extractor.baseURI);
        let original = param.original;
        if (param.scheme == "file") {
          original = UnMHTFileUtil.getFilenameFromURLSpec (original);
        }
        
        body.push ("<table border=\"0\">");
        
        body.push (tableRow ("URI: ",
                             "<a href=\""
                             + UnMHTHTMLUtil.escapeEntity (extractor.baseURI)
                             + "\">"
                             + UnMHTHTMLUtil.escapeEntity (extractor.baseURI)
                             + "</a>"));
        body.push (tableRow (bundle ("original_uri") + " ",
                             UnMHTHTMLUtil.escapeEntity (original)));
        body.push (tableRow (bundle ("opened_time") + " ",
                             UnMHTHTMLUtil.escapeEntity
                             (UnMHTFormatter.getReadableTime
                              (extractor.extractTime))));
        body.push (tableRow (bundle ("last_access_time") + " ",
                             UnMHTHTMLUtil.escapeEntity
                             (UnMHTFormatter.getReadableTime
                              (extractor.lastAccessTime))));
        body.push (tableRow (bundle ("file_count") + " ",
                             extractor.parts.length));
        
        if (extractor.date) {
          let savedTime = extractor.date;
          try {
            savedTime = UnMHTFormatter.getReadableTime
              ((new Date (savedTime)).getTime ());
          }
          catch (e) {
          }
          
          body.push (tableRow (bundle ("saved_time") + " ",
                               UnMHTHTMLUtil.escapeEntity (savedTime)));
        }
        
        if (extractor.subject) {
          body.push (tableRow ("Subject: ",
                               UnMHTHTMLUtil.escapeEntity (extractor.subject)));
        }
        
        body.push ("</table>");
        
        extractor.parts.forEach (function (part) {
            let isStart = "";
            if (extractor.startPart == part) {
              isStart = " (" + bundle ("root_document") + ")";
            }
            
            body.push ("<table border=\"0\" class=\"file\">");
            
            body.push (tableRow (bundle ("ref_name") + " ",
                                 "<a href=\""
                                 + UnMHTHTMLUtil.escapeEntity
                                 (extractor.baseURI)
                                 + UnMHTHTMLUtil.escapeEntity
                                 (part.eParam.refName)
                                 + "\">"
                                 + UnMHTHTMLUtil.escapeEntity
                                 (part.eParam.refName)
                                 + "</a>" + isStart));
            
            if (part.contentLocation) {
              let location = part.contentLocation;
              body.push (tableRow ("Content-Location: ",
                                   UnMHTHTMLUtil.escapeEntity (location)));
            }
            
            if (part.eParam.contentBase) {
              let location = part.eParam.contentBase;
              body.push (tableRow ("Content-Base: ",
                                   UnMHTHTMLUtil.escapeEntity (location)));
            }
              
            if (part.contentID) {
              body.push (tableRow ("Content-ID: ",
                                   UnMHTHTMLUtil.escapeEntity
                                   (part.contentID)));
            }
            
            if (part.contentDispositionType
                && part.contentDispositionFilename) {
              body.push (tableRow ("Content-Disposition: ",
                                   UnMHTHTMLUtil.escapeEntity
                                   (part.contentDispositionType)
                                   + "; filename="
                                   + UnMHTHTMLUtil.escapeEntity
                                   (part.contentDispositionFilename)));
            }
            
            let charset = "";
            if (part.eParam.charset) {
              charset
                = "; charset=" + UnMHTHTMLUtil.escapeEntity
                (part.eParam.charset);
            }
            body.push (tableRow ("Content-Type: ",
                                 UnMHTHTMLUtil.escapeEntity
                                 (part.eParam.mimetype)
                                 + charset));
            
            if (part.contentTransferEncoding) {
              body.push (tableRow ("Content-Encoding: ",
                                   UnMHTHTMLUtil.escapeEntity
                                   (part.contentTransferEncoding)));
            }
                
            body.push (tableRow (bundle ("file_size") + " ",
                                 UnMHTFormatter.getReadableSize
                                 (part.eParam.content.length)
                                 + " " + bundle ("bytes")));
            
            body.push ("</table>");
          });
        
        bodies.push (body);
      });
    
    let data
    = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
    + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    + "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
    + "<head>"
    + "<title>" + bundle ("cache_title") + "</title>"
    + "<style type=\"text/css\">"
    + "th {"
    + "text-align: right;"
    + "white-space: nowrap;"
    + "}"
    + "table {"
    + "margin-bottom: 1em;"
    + "}"
    + "table.file {"
    + "margin-bottom: 1em;"
    + "margin-left: 1em;"
    + "}"
    + "</style>"
    + "</head>"
    + "<body>"
    + "<h1>" + bundle ("cache_title") + "</h1>"
    
    + "<hr />"
    + "<table border=\"0\">"
        
    + tableRow (bundle ("cache_entry") + " ",
                extractors.length)
        
    + tableRow (bundle ("cache_size") + " ",
                UnMHTFormatter.getReadableSize (size) + " " + bundle ("bytes"))
    
    + "</table>"
    + "<hr />"
    + bodies.map (function (body) body.join ("")).join ("<hr />");
    + "</body>"
    + "</html>";
    
    data = UnMHTUnicodeConverter.convertFromUnicode (data, "utf-8");
    
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 data, data.length);
  },
  
  /**
   * エラー情報のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createErrorChannel : function (uri) {
    let bundle = function (name) {
      return UnMHTHTMLUtil.escapeEntity (UnMHTStringBundle.get (name));
    };
    
    let data
    = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
    + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    + "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
    + "<head>"
    + "<title>"
    + bundle ("error_title")
    + "</title>"
    + "<style type=\"text/css\">"
    + "h1 {"
    + "color: #ff0000;"
    + "background-color: inherit;"
    + "}"
    + "div.message {"
    + "font-size: 1em;"
    + "color: #ff0000;"
    + "background-color: inherit;"
    + "}"
    + "</style>"
    + "</head>"
    + "<body>"
    + "<h1>"
    + bundle ("error_title")
    + "</h1>"
        
    + "<hr />"
        
    + "<div class=\"message\">"
    + bundle ("error_file_not_found") + "<br>"
    + UnMHTHTMLUtil.escapeEntity (UnMHTScheme.deErrorURI (uri.spec))
    + "</div>"
        
    + "<hr />"
        
    + "</body>"
    + "</html>";
    
    data = UnMHTUnicodeConverter.convertFromUnicode (data, "utf-8");
    
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 data, data.length);
  },

  /**
   * 不正な URI 情報のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createIllegalChannel : function (uri) {
    let bundle = function (name) {
      return UnMHTHTMLUtil.escapeEntity (UnMHTStringBundle.get (name));
    };
    
    let data
    = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
    + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    + "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
    + "<head>"
    + "<title>"
    + bundle ("error_title")
    + "</title>"
    + "<style type=\"text/css\">"
    + "h1 {"
    + "color: #ff0000;"
    + "background-color: inherit;"
    + "}"
    + "div.message {"
    + "font-size: 1em;"
    + "color: #ff0000;"
    + "background-color: inherit;"
    + "}"
    + "</style>"
    + "</head>"
    + "<body>"
    + "<h1>"
    + bundle ("error_title")
    + "</h1>"
        
    + "<hr />"
        
    + "<div class=\"message\">"
    + bundle ("error_illegal_address") + "<br>"
    + UnMHTHTMLUtil.escapeEntity (uri.spec)
    + "</div>"
        
    + "<hr />"
        
    + "</body>"
    + "</html>";
    
    data = UnMHTUnicodeConverter.convertFromUnicode (data, "utf-8");
    
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 data, data.length);
  },
  
  /**
   * リフレッシュ用のチャネルを作成する
   *
   * @param  nsiURI uri
   *         対象の URI
   * @param  String location
   *         リフレッシュ先
   * @return nsIChannel
   *         作成したチャネル
   */
  _createRefreshChannel : function (uri, location) {
    let data = "";
    
    data = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
    + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    + "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
    + "<head>"
    + "<title>UnMHT</title>"
    + "<meta http-equiv=\"refresh\" content=\"1;URL="
    + UnMHTHTMLUtil.escapeEntity (location) + "\" />"
    + "</head>"
    + "<body>"
    + "</body>"
    + "</html>";
    
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 data, data.length);
  },
  
  /**
   * 空のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createEmptyChannel : function (uri) {
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 "", 0);
  },
  
  /**
   * 文字列を返すチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @param  String charset
   *         文字セット
   * @param  String type
   *         MIME-Type
   * @param  String data
   *         文字列
   * @param  Number length
   *         文字列の長さ
   * @return nsIChannel
   *         作成したチャネル
   */
  _createStringInputStreamChannel : function (uri, charset, type,
                                              data, length) {
    let sstream
    = Cc ["@mozilla.org/io/string-input-stream;1"]
    .createInstance (Ci.nsIStringInputStream);
    sstream.setData (data, length);
    
    return this._createInputStreamChannel (uri, charset, type,
                                           sstream, length);
  },
  
  /**
   * ストリームの内容を返すチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @param  String charset
   *         文字セット
   * @param  String type
   *         MIME-Type
   * @param  nsIInputStream stream
   *         ストリーム
   * @param  Number length
   *         ストリームの長さ
   * @return nsIChannel
   *         作成したチャネル
   */
  _createInputStreamChannel : function (uri, charset, type, stream, length) {
    let inputStreamChannel
    = Cc ["@mozilla.org/network/input-stream-channel;1"]
    .createInstance (Ci.nsIInputStreamChannel)
    .QueryInterface (Ci.nsIChannel);
    
    inputStreamChannel.setURI (uri);
    inputStreamChannel.contentCharset = charset;
    inputStreamChannel.contentType    = type;
    inputStreamChannel.contentStream  = stream;
    inputStreamChannel.contentLength  = length;
    
    return inputStreamChannel;
  }
};
