/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTSave.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveDialogClass"];

/**
 * 保存ダイアログ
 *
 * ウィンドウの引数
 * {
 *   useCurrentState: Boolean  現在の状態か,
 *   scrapBookData:   Object   ScrapBook のデータ,
 *   targetDocument:  HTMLDocument  対象のドキュメント,
 *   callback:        Function  コールバック関数
 *   targetFile:      nsIFile  保存先のファイル,
 *   targetSelection: Selection  選択範囲,
 *   additionalFiles: Array  追加ファイルの URI [String URI, ...]
 * }
 *
 * @param  Window window
 *         対象のウィンドウ
 * @implements UnMHTSaveListener
 */
function UnMHTSaveDialogClass (window) {
  this.window = window;
}
UnMHTSaveDialogClass.prototype = {
  window : null, /* Window  対象のウィンドウ */
  
  progressLabel : null,       /* XULElement  取得状況のラベル */
  progressCurrent : null,     /* XULElement  現在取得しているファイル */
  progressCurrentSize : null, /* XULElement  現在取得しているファイルのサイズ */
  progress : null,            /* XULElement  取得状況のバー */
  failedLabel : null,         /* XULElement  失敗したファイルのラベル */
  failedLabel2 : null,        /* XULElement  失敗したファイルのラベル2 */
  failedLabel3 : null,        /* XULElement  失敗したファイルのラベル3 */
  failedList : null,          /* XULElement  失敗したファイル */
  acceptButton : null,        /* XULElement  OK ボタン */
  cancelButton : null,        /* XULElement  キャンセルボタン */
  
  callback : null,            /* Function  終了時コールバック */
  
  saveContext : null,         /* UnMHTSaveContext  保存コンテキスト */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    
    this.progressLabel = null;
    this.progressCurrent = null;
    this.progressCurrentSize = null;
    this.progress = null;
    this.failedLabel = null;
    this.failedLabel2 = null;
    this.failedLabel3 = null;
    this.failedList = null;
    this.acceptButton = null;
    this.cancelButton = null;
    
    this.callback = null;
    
    this.saveContext = null;
  },
  
  /**
   * getElementById の省略
   */
  el : function (id) {
    return this.window.document.getElementById (id);
  },
  
  /* ==== UI イベント ==== */
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    let param = this.window.arguments [0];
        
    let dialog = this.el ("unmht_save_dialog");
    
    this.acceptButton = dialog.getButton ("accept");
    this.acceptButton.disabled = true;
    this.cancelButton = dialog.getButton ("cancel");
    
    this.progressLabel = this.el ("progress_label");
    this.progressLabel.value = UnMHTStringBundle.get ("save_begin");
    this.progressCurrent = this.el ("progress_current");
    this.progressCurrentSize
    = this.el ("progress_current_size");
    this.progress = this.el ("progress");
    this.progress.setAttribute ("value", "0%");
    
    this.failedLabel = this.el ("failed_label");
    this.failedLabel2 = this.el ("failed_label2");
    this.failedLabel3 = this.el ("failed_label3");
    this.failedList = this.el ("failed");
    
    let title;
    if (param.useCurrentState) {
      title = this.el ("current_label").value;
    }
    else  {
      title = this.el ("original_label").value;
    }
    dialog.setAttribute ("title", title);
    
    let subject;
    let location;
    if (param.scrapBookData) {
      subject = param.scrapBookData ["title"];
      location = param.scrapBookData ["source"];
    }
    else {
      subject = param.targetDocument.title;
      location = param.targetDocument.location.href;
    }
    
    this.callback = param.callback;
    
    UnMHTSave.setCurrentWindow (this.window);
    let saveContext = new UnMHTSaveContext ();
    saveContext.init (param.targetFile, location, subject, this);
    
    saveContext.addDocument
    (param.targetDocument, param.targetSelection,
     param.scrapBookData, param.useCurrentState,
     location, true, true);
    if (saveContext.isStopped) {
      return;
    }
    
    if (param.additionalFiles) {
      param.additionalFiles.forEach (function (uri) {
          saveContext.addFile (null, uri, false);
        });
    }
    
    this.saveContext = saveContext;
    
    saveContext.start ();
  },
  
  /**
   * アンロードイベント
   */
  onUnload : function () {
  },
  
  /**
   * OK ボタンのイベント
   */
  onAcceptDialog : function () {
    if (this.callback) {
      this.callback (false);
    }
        
    return true;
  },
  
  /**
   * キャンセルボタンのイベント
   */
  onCancelDialog : function (context) {
    if (!this.saveContext) {
      return true;
    }
    if (this.saveContext.isStopped) {
      /* 保存が停止していればダイアログを閉じる */
      
      if (this.callback) {
        this.callback (true);
      }
      return true;
    }
    
    this.saveContext.cancel ();
    this.saveContext = null;
    
    return false;
  },
  
  /* ==== UnMHTSaveListener ===== */
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String type
   *         エラーの種類
   * @param  String desc
   *         エラーの詳細
   */
  onError : function (context, type, desc) {
    if (type == "save") {
      this.progressLabel.value = UnMHTStringBundle.get ("error_save");
    }
    else if (type == "open") {
      this.progressLabel.value = UnMHTStringBundle.get ("error_open");
    }
    else {
      this.progressCurrent.value = desc;
    }
  },

  /**
   * 進捗状況変化イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String currentFile
   *         現在取得のファイル名
   * @param  String currentSize
   *         取得したバイト数 / ファイルサイズ
   * @param  Number ok
   *         取得完了した数
   * @param  Number na
   *         取得中の数
   * @param  Number ng
   *         取得失敗した数
   */
  onProgress : function (context, currentFile, currentSize, ok, na, ng) {
    this.progressCurrent.value = currentFile;
    this.progressCurrentSize.value = currentSize;
    
    if (na > 0) {
      this.progressLabel.value
        = UnMHTStringBundle.get ("save_begin")
        + " " + (ok + ng) + "/" + (ok + na + ng);
      this.progress.setAttribute
        ("value", Math.floor (ok * 100 / (ok + na + ng)) + "%");
    }
    else {
      this.progressLabel.value
      = UnMHTStringBundle.get ("save_begin")
      + " " + (ok + ng) + "/" + (ok + na + ng);
      this.progress.setAttribute ("value", "100%");
    }
  },
  
  /**
   * 保存完了
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (context, ngList) {
    this.progressLabel.value = UnMHTStringBundle.get ("save_done");
    
    if (UnMHTPref.getPref ("save.show_errors") && ngList.length > 0) {
      this.acceptButton.disabled = false;
      this.cancelButton.disabled = true;
      
      this.failedList.value = ngList.join ("\n");
      this.failedList.hidden = false;
      
      if (this.failedLabel && this.failedLabel.value != "") {
        this.failedLabel.hidden = false;
      }
      if (this.failedLabel2 && this.failedLabel2.value != "") {
        this.failedLabel2.hidden = false;
      }
      if (this.failedLabel3 && this.failedLabel3.value != "") {
        this.failedLabel3.hidden = false;
      }
      
      this.window.sizeToContent ();
    }
    else {
      UnMHTTimer.setTimeout (function () {
          if (this.callback) {
            this.callback (false);
          }
          this.window.close ();
        }.bind (this), 1000);
    }
  },
  
  /**
   * キャンセルが完了したイベント
   */
  onCancel : function () {
    if (this.callback) {
      this.callback (true);
    }
    this.window.close ();
  }
};
