/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTCustomFilename.jsm");
Cu.import ("resource://unmht/modules/UnMHTDOMUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTFilePicker.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTFormatter.jsm");
Cu.import ("resource://unmht/modules/UnMHTMail.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTSave.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTShutdown.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");
Cu.import ("resource://unmht/modules/UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveManager"];

/**
 * 保存イベントリスナ
 *
 * @param  Window window
 *         実行中のウィンドウ
 */
function UnMHTSaveProgressListener (window) {
  this.window = window;
}
UnMHTSaveProgressListener.prototype = {
  window : null,    /* Window  実行中のウィンドウ */
  callback : null,  /* Function  完了時コールバック */
  prog : null,      /* XULElement  進捗状況全体のボックス */
  progbar : null,   /* XULElement  進捗状況のプログレスバー */
  proglabel : null, /* XULElement  進捗状況のラベル */
  
  progress : 0,     /* Number  進捗状況
                     *   0-1: 保存中の取得率
                     *   2:   保存完了 */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.prog = null;
    this.progbar = null;
    this.proglabel = null;
    this.window = null;
    this.callback = null;
  },
  
  /**
   * 進捗状況変化イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String currentFile
   *         現在取得のファイル名
   * @param  String currentSize
   *         取得したバイト数 / ファイルサイズ
   * @param  Number ok
   *         取得完了した数
   * @param  Number na
   *         取得中の数
   * @param  Number ng
   *         取得失敗した数
   */
  onProgress : function (context, currentFile, currentSize, ok, na, ng) {
    if (na > 0) {
      this.progress = ok / (ok + na + ng);
    }
    else {
      this.progress = 1;
    }
    
    UnMHTSaveManager.updateProgress ();
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String type
   *         エラーの種類
   * @param  String desc
   *         エラーの詳細
   */
  onError : function (context, type, desc) {
    let result = {
      result : 0,
      applyToAll: false,
      rename : false,
      filename : ""
    };
        
    let mes = "";
    if (type == "save") {
      mes = UnMHTStringBundle.get ("error_save");
    }
    else if (type == "open") {
      mes = UnMHTStringBundle.get ("error_open");
    }
    
    let confirmURL = "chrome://unmht/content/confirm.xul";
    let features = "chrome,dialog,modal,centerscreen";
    let param = {
      result: result,
      title: UnMHTStringBundle.get ("error_title"),
      message1: mes,
      message2: desc,
      accept: UnMHTStringBundle.get ("ok"),
    };
    this.window.openDialog (confirmURL, "", features, param);
    
    if (this.callback) {
      this.callback (true);
      this.callback = null;
    }
    
    UnMHTSaveManager.removeContext (context);
  },
  
  /**
   * 保存完了
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (context, ngList) {
    this.progress = 2;
    
    if (UnMHTPref.getPref ("save.show_errors") && ngList.length > 0) {
      let nglistURL = "chrome://unmht/content/nglist.xul";
      let features = "chrome,titlebar,toolbar,centerscreen,resizable";
      let param = { ngList : ngList };
      this.window.openDialog (nglistURL, "", features, param);
    }
    
    if (this.callback) {
      this.callback (false);
      this.callback = null;
    }
    
    UnMHTSaveManager.updateProgress ();
    UnMHTSaveManager.removeContext (context);
  },
  
  /**
   * キャンセルが完了したイベント
   */
  onCancel : function (context) {
    if (this.callback) {
      this.callback (true);
      this.callback = null;
    }
    
    UnMHTSaveManager.removeContext (context);
  }
};

/**
 * 保存の管理
 */
let UnMHTSaveManager = {
  contexts : [],    /* Array 保存コンテキスト
                     *   [UnMHTSaveContext, ...] */
  
  /**
   * 初期化処理
   */
  init : function () {
  },
  
  /**
   * 終了処理
   */
  term : function () {
    if (this.contexts) {
      this.contexts
      .forEach (function (context) {
          context.destruct ();
        });
      this.contexts = [];
    }
  },
  
  /**
   * 保存状況をステータスバーに表示するリスナを作成する
   *
   * @param  Window window
   *         実行中のウィンドウ
   */
  createListener : function (window) {
    let pref
    = window.document.getElementById ("unmht-statusbarpanel-preferences");
    
    let prog
    = window.document.getElementById ("unmht-statusbarpanel-progress");
    let progbar
    = window.document.getElementById ("unmht-statusbarpanel-progressbar");
    let proglabel
    = window.document.getElementById ("unmht-statusbarpanel-progress-label");
    
    if (pref && prog && pref.parentNode) {
      pref.parentNode.insertBefore (prog, pref.nextSibling);
    }
    
    let listener = new UnMHTSaveProgressListener (window);
    
    if (prog && progbar && proglabel) {
      progbar.setAttribute ("value", "0%");
      prog.hidden = false;
      progbar.hidden = false;
      proglabel.hidden = true;
      
      listener.prog = prog;
      listener.progbar = progbar;
      listener.proglabel = proglabel;
    }
    
    return listener;
  },
  
  /**
   * ステータスバーの進捗状況を更新する
   */
  updateProgress : function () {
    let na = 0;
    let ok = 0;
    let progress = 0;
    
    let progs = [];
    let progbars = [];
    let proglabels = [];
    
    this.contexts.forEach (function (context) {
        let listener = context.listener;
        
        if (progs.indexOf (listener.prog) == -1) {
          progs.push (listener.prog);
        }
        if (progbars.indexOf (listener.progbar) == -1) {
          progbars.push (listener.progbar);
        }
        if (proglabels.indexOf (listener.proglabel) == -1) {
          proglabels.push (listener.proglabel);
        }
        
        if (listener.progress == 2) {
          progress += 1;
          ok ++;
        }
        else {
          progress += listener.progress;
          na ++;
        }
      });
    
    if (na == 0) {
      if (ok > 0) {
        /* 保存完了 */
        progs.forEach (function (prog) prog.hidden = false);
        progbars.forEach (function (progbar) progbar.hidden = true);
        proglabels.forEach (function (proglabel) {
            proglabel.hidden = false;
            proglabel.value = UnMHTStringBundle.get ("save_done");
          });
        
        if (progs.length) {
          UnMHTTimer.setTimeout
            (function () {
              progs.forEach (function (prog) prog.hidden = true);
            }.bind (this), 3000);
        }
      }
      else {
        /* 現在なにも無い */
        progs.forEach (function (prog) prog.hidden = true);
      }
    }
    else {
      /* 保存中 */
      progress /= this.contexts.length;
      
      progs.forEach (function (prog) prog.hidden = false);
      progbars.forEach (function (progbar) {
          progbar.hidden = false;
          progbar.setAttribute ("value", Math.floor (progress * 100) + "%");
        });
      proglabels.forEach (function (proglabel) proglabel.hidden = true);
    }
  },
  
  /**
   * 保存する際のファイル名を取得する
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   */
  getFilename : function (targetDocument, scrapBookData) {
    let filename;
    let title, uri;
        
    filename = "";
    
    if (scrapBookData) {
      uri = scrapBookData ["source"];
      title = scrapBookData ["title"];
    }
    else {
      uri = targetDocument.location.href;
      try {
        uri = decodeURI (uri);
      }
      catch (e) {
      }
      title = targetDocument.title;
    }
    
    if (UnMHTPref.getPref ("save.local.same_dir")) {
      let original = uri;
      if (UnMHTScheme.isUnMHTURI (original)) {
        let param = UnMHTScheme.getUnMHTURIParam (original);
        original = param.original;
      }
      if (original.match (/^file:\/\/\//)) {
        /* ローカルファイルである */
        let file = UnMHTFileUtil.getFileFromURLSpec (original);
        filename = file.leafName;
        if (filename.match (/^(.+)\.[^\.]+$/)) {
          filename = RegExp.$1;
        }
        
        let append = UnMHTPref.getPref ("save.local.same_dir.append");
        return filename + append;
      }
    }
    
    if (uri.match (/[^\?]+\/([^\/\?]+)(\?.+)?(#.+)?$/)) {
      filename = RegExp.$1;
      if (filename.match (/^(.+)\.[^\.]+$/)) {
        filename = RegExp.$1;
      }
    }
    else {
      filename = "index";
    }
        
    let saveFilename = UnMHTPref.getPref ("save.filename");
    if (UnMHTAppInfo.isThunderbird ()
        && saveFilename == "filename") {
      saveFilename = "title";
    }
    if (saveFilename == "title") {
      let tmp = UnMHTFileUtil.validateFileName (title);
            
      if (tmp) {
        filename = tmp;
      }
    }
    else if (saveFilename == "custom") {
      filename = UnMHTCustomFilename.format (title, filename, uri);
    }
    else {
      filename = UnMHTFileUtil.validateFileName (filename);
    }
    
    return filename;
  },
  
  /**
   * 現在のドキュメントを MHT で保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  save : function (window, callback) {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
    this.saveDocument (window, targetDocument, callback);
  },
  
  /**
   * 対象のドキュメントを MHT で保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveDocument : function (window, targetDocument, callback) {
    this.saveDocumentWithSelection (window, targetDocument, null, callback);
  },
  
  /**
   * 対象のドキュメントの対象の選択範囲を MHT で保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Selection targetSelection
   *         対象の選択範囲
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveDocumentWithSelection : function (window, targetDocument, targetSelection,
                                        callback) {
    this.showOptionDialogOrSave (window, targetDocument, targetSelection,
                                 null, false,
                                 callback);
  },
  
  /**
   * 現在のドキュメントの選択範囲を保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveSelection : function (window, callback) {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
    
    this.saveSelectionInDocument (window, targetDocument, callback);
  },
  
  /**
   * 現在のドキュメントの現在の選択範囲を MHT で保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveSelectionInDocument : function (window, targetDocument, callback) {
    let targetSelection = targetDocument.defaultView.getSelection ();
    if (targetSelection.rangeCount == 0) {
      return;
    }
    let targetRange = targetSelection.getRangeAt (0);
    targetDocument = targetRange.startContainer.ownerDocument;
    
    this.saveDocumentWithSelection (window, targetDocument, targetSelection,
                                    callback);
  },
  
  /**
   * リンク先を MHT で保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveLink : function (window, callback) {
    let node = window.document.popupNode;
    while (node) {
      if ("href" in node) {
        break;
      }
      node = node.parentNode;
    }
    if (!node) {
      return;
    }
    let href = node.href;
    if (!href) {
      return;
    }
    
    let iframe = window.document.createElement ("iframe");
    let wrappedCallback = function (isStopped) {
      iframe.parentNode.removeChild (iframe);
      if (callback) {
        callback (isStopped);
      }
    };
    
    iframe.hidden = true;
    iframe.setAttribute ("src", href);
    iframe.setAttribute ("type", "content");
    iframe.addEventListener
    ("load",
     function () {
      this.saveDocument (window, iframe.contentDocument, wrappedCallback);
    }.bind (this), true);
    
    let tabbrowser = window.document.getElementById ("content");
    if (tabbrowser) {
      tabbrowser.parentNode.insertBefore (iframe, tabbrowser);
    }
  },
  
  /**
   * 複数のタブを MHT で保存
   *
   * @param  Window window
   *         実行中のウィンドウ
   */
  savemulti : function (window) {
    let pageURL = "chrome://unmht/content/savemulti.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let param = {
      targetWindow : window
    };
    window.openDialog (pageURL, "", features, param);
  },
  
  /**
   * QuickSave
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  quicksave : function (window, callback) {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
    
    if (!this.isSavable (window, targetDocument)) {
      window.alert (UnMHTStringBundle.get ("not_savable"));
      return;
    }
    
    let base = UnMHTPref.getPref ("save.quick.dir");
    let useCurrentState = (UnMHTPref.getPref ("save.quick.type") == "current");
    
    let dir = base ? UnMHTFileUtil.initFile (base) : null;
    if (!dir || !dir.exists ()) {
      base
        = Cc ["@mozilla.org/file/directory_service;1"]
        .getService (Ci.nsIProperties).get ("Desk", Ci.nsIFile).path;
    }
    
    if (UnMHTPref.getPref ("save.local.same_dir")) {
      let original = targetDocument.location.href;
      if (UnMHTScheme.isUnMHTURI (original)) {
        let param = UnMHTScheme.getUnMHTURIParam (original);
        original = param.original;
      }
      
      if (original.match (/^file:\/\/\//)) {
        /* ローカルファイルである */
        base = UnMHTFileUtil.getFileFromURLSpec (original).parent.path;
      }
    }
    
    let filenameBase = this.getFilename (targetDocument, null);
    let filename = filenameBase + ".mht";
    let targetFile = UnMHTFileUtil.initFile (base);
    targetFile.appendRelativePath (filename);
    if (targetFile.exists ()) {
      let dupAction = UnMHTPref.getPref ("save.quick.dup.action");
      if (dupAction == "rename") {
        filename = UnMHTSave.getUniqueFilename (base, filenameBase);
        targetFile = UnMHTFileUtil.initFile (base);
        targetFile.appendRelativePath (filename);
      }
      else if (dupAction == "ask") {
        let result = {
          result : 0,
          applyToAll: false,
          rename : false,
          filename : ""
        };
        
        let confirmURL = "chrome://unmht/content/confirm.xul";
        let features = "chrome,dialog,modal,centerscreen";
        let param = {
          result: result,
          title: UnMHTStringBundle.get ("overwrite_title"),
          message1: UnMHTStringBundle.get ("overwrite_pre")
          + " " + filename + " "
          + UnMHTStringBundle.get ("overwrite_post"),
          message2: UnMHTStringBundle.get ("overwrite_lastaccess")
          + " " + UnMHTFormatter.getReadableTime (targetFile.lastModifiedTime),
          message3: UnMHTStringBundle.get ("overwrite_post2"),
          accept: UnMHTStringBundle.get ("replace"),
          extra2: UnMHTStringBundle.get ("abort"),
          rename: UnMHTStringBundle.get ("rename"),
          toall: null,
          filename: UnMHTSave.getUniqueFilename (base, filenameBase),
          base: base,
          exist1: UnMHTStringBundle.get ("overwrite_pre"),
          exist2: UnMHTStringBundle.get ("overwrite_post")
        };
        window.openDialog (confirmURL, "", features, param);
        
        if (result.result == "abort" || result.result == "no") {
          return;
        }
        
        if (result.result == "rename") {
          targetFile = UnMHTFileUtil.initFile (base);
          targetFile.appendRelativePath (result.filename);
        }
      }
    }
    
    let targetSelection = null;
    if (UnMHTPref.getPref ("save.quick.use_selection")) {
      let targetWindow = targetDocument.defaultView;
      if (UnMHTDOMUtil.isTextSelected (window, false)) {
        targetSelection = targetWindow.getSelection ();
      }
    }
    
    this.startSave (window, targetDocument, targetSelection,
                    null,
                    useCurrentState, null,
                    targetFile, callback);
  },
  
  /**
   * 現在のドキュメントを MHT としてメールで送信
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  sendmail : function (window, callback) {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
    this.sendmailDocument (window, targetDocument, callback);
  },

  /**
   * 対象のドキュメントを MHT としてメールで送信
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Function callback
   *         保存終了時のコールバック
   */
  sendmailDocument : function (window, targetDocument, callback) {
    this.showOptionDialogOrSave (window, targetDocument, null,
                                 null, true,
                                 callback);
  },
  
  /**
   * メール用のファイルとコールバックを作成する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  String targetFilename
   *         対象のファイル名
   * @param  Function callback
   *         保存終了時のコールバック
   * @return Array
   *         [nsIFile 対象のファイル, Function コールバック]
   */
  generateMailData : function (window, targetDocument,
                               targetFilename, callback) {
    let tmpDir
    = Cc ["@mozilla.org/file/directory_service;1"]
    .getService (Ci.nsIProperties).get ("TmpD", Ci.nsIFile);
    
    let tmpURI
    = Services.io.newURI (UnMHTFileUtil.getURLSpecFromFile (tmpDir),
                          null, null);
    
    let targetFile
    = UnMHTFileUtil.getFileFromURLSpec (tmpURI.resolve (targetFilename));
    
    let title
    = UnMHTUnicodeConverter.convertFromUnicode (targetDocument.title, "utf-8");
    
    let wrappedCallback = function (isStopped) {
      if (!isStopped) {
        if (UnMHTPref.getPref ("save.mail.zip")) {
          let zippedPath
          = targetFile.path.replace (/\.mht(ml)?$/i, "") + ".zip";
            
          var zip
          = Cc ["@mozilla.org/zipwriter;1"].createInstance (Ci.nsIZipWriter);
          let PR_WRONLY = 0x02;
          let PR_CREATE_FILE = 0x08;
          let PR_TRUNCATE = 0x20;
          zip.open
          (UnMHTFileUtil.initFile (zippedPath),
           PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE);
          zip.addEntryFile
          (targetFile.leafName,
           Ci.nsIZipWriter.COMPRESSION_DEFAULT, targetFile, false);
          zip.close ();
            
          targetFile = UnMHTFileUtil.initFile (zippedPath);
        }
          
        let debugFunc = function (app, arg, args) {
          window.alert
          ("[Debug]\n"
           + "Command line for mail\n"
           + "\n"
           + app + " " + arg + "\n"
           + "\n"
           + "Arguments will be passed as following array\n"
           + "\n"
           + args.join ("\n"));
        }.bind (this);
          
        let ret
        = UnMHTMail.launch (targetFile, title, debugFunc);
        if (!ret) {
          window.alert (UnMHTStringBundle.get ("sendmail_fail") + "\n"
                        + UnMHTStringBundle.get ("sendmail_fail2"));
        }
      }
        
      if (callback) {
        callback (isStopped);
      }
    }.bind (this);
    
    return [targetFile, wrappedCallback];
  },
  
  /**
   * 対象のドキュメントを対象のファイルに保存する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Boolean useCurrentState
   *         現在の状態を使用するか
   * @param  Array additionalFiles
   *         追加ファイル
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         終了時コールバック
   */
  saveDocumentToFile : function (window, targetDocument,
                                 useCurrentState, useAdditionalFiles,
                                 targetFile, callback) {
    if (useAdditionalFiles) {
      let saveParam = {
        useCurrentState: useCurrentState
      };
      this.showAdditionalFilesDialogAndSave (window, targetDocument, null,
                                             null, saveParam,
                                             targetFile, callback);
    }
    else {
      this.startSave (window, targetDocument, null,
                      null,
                      useCurrentState, null,
                      targetFile, callback);
    }
  },
  
  /**
   * ScrapBook を MHT で保存する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Function callback
   *         終了時コールバック
   */
  saveScrapBook : function (window, scrapBookData, callback) {
    this.showOptionDialogOrSave (window, null, null,
                                 scrapBookData, false,
                                 callback);
  },
  
  /**
   * 必要であればオプションダイアログを表示してから保存する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Selection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean isMail
   *         メールで送信するかどうか
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         保存終了時のコールバック
   */
  showOptionDialogOrSave : function (window, targetDocument, targetSelection,
                                     scrapBookData, isMail,
                                     callback) {
    let targetFilename
    = this.getFilename (targetDocument, scrapBookData);
    targetFilename += ".mht";
    
    let dialogCallback;
    
    if (isMail) {
      dialogCallback = function (saveParam) {
        let [targetFile, wrappedCallback]
        = this.generateMailData (window, targetDocument,
                                 saveParam.targetFilename,
                                 callback);
          
        this.startSave (window, targetDocument, targetSelection,
                        scrapBookData,
                        saveParam.useCurrentState, saveParam.additionalFiles,
                        targetFile, wrappedCallback);
      }.bind (this);
    }
    else {
      dialogCallback = function (saveParam) {
        this.showFilePickerOrSave (window, targetDocument, targetSelection,
                                   scrapBookData, saveParam,
                                   callback);
      }.bind (this)
    }
    
    if (UnMHTPref.getPref ("save.show_option_dialog")) {
      let showSaveType = true;
      let saveType = UnMHTPref.getPref ("save.single.type");
      let useCurrentState = (saveType == "current");
      
      if (isMail) {
        showSaveType = false;
        saveType = UnMHTPref.getPref ("save.mail.type");
        useCurrentState = (saveType == "current");
      }
      else if (scrapBookData) {
        showSaveType = false;
        useCurrentState = false;
      }
      else if (UnMHTAppInfo.isThunderbird ()) {
        showSaveType = false;
        useCurrentState = true;
      }
      else if (targetSelection) {
        showSaveType = false;
        useCurrentState = true;
      }
      
      let pageURL = "chrome://unmht/content/saveoptions.xul";
      let features = "chrome,titlebar,centerscreen,resizable";
      let param = {
        callback : function (saveParam) {
          saveParam.showAdditionalFilesDialog = false; 
          dialogCallback (saveParam);
        },
        filename : targetFilename,
        showFilename : isMail,
        showSaveType : showSaveType,
        useCurrentState : useCurrentState,
      };
      window.openDialog (pageURL, "", features, param);
    }
    else if (isMail) {
      let saveType = UnMHTPref.getPref ("save.mail.type");
      let saveParam = {
        targetFilename : targetFilename,
        useCurrentState: (saveType == "current"),
        additionalFiles: null,
      };
      dialogCallback (saveParam);
    }
    else {
      let saveParam = {
        targetFilename : targetFilename
      };
      dialogCallback (saveParam);
    }
  },
  
  /**
   * 必要であればファイル選択ダイアログを表示してから保存する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Selection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Object saveParam
   *         追加パラメータ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  showFilePickerOrSave : function (window, targetDocument, targetSelection,
                                   scrapBookData, saveParam,
                                   callback) {
    let currentIndex, originalIndex;
    let currentAddIndex, originalAddIndex;
    
    let showAdditionalFilesDialog = false;
    
    let currentFilePicker = null;
    
    let title = UnMHTStringBundle.get ("save_title");
    let mode = Ci.nsIFilePicker.modeSave;
    let modFunc = function (filePicker) {
      currentFilePicker = filePicker;
        
      if (UnMHTPref.getPref ("save.local.same_dir")) {
        let original = targetDocument.location.href;
        if (UnMHTScheme.isUnMHTURI (original)) {
          let param = UnMHTScheme.getUnMHTURIParam (original);
          original = param.original;
        }
      
        if (original.match (/^file:\/\/\//)) {
          /* ローカルファイルである */
          let file = UnMHTFileUtil.getFileFromURLSpec (original);
          filePicker.displayDirectory = file.parent;
        }
      }
        
      if ("useCurrentState" in saveParam
          && "additionalFiles" in saveParam) {
        if (saveParam.useCurrentState) {
          if (saveParam.additionalFiles.length > 0) {
            filePicker.appendFilter
            (UnMHTStringBundle.get ("filetype")
             + " " + UnMHTStringBundle.get ("filetype_current_add"),
             "*.mht");
          }
          else {
            filePicker.appendFilter
            (UnMHTStringBundle.get ("filetype")
             + " " + UnMHTStringBundle.get ("filetype_current"),
             "*.mht");
          }
        }
        else {
          if (saveParam.additionalFiles.length > 0) {
            filePicker.appendFilter
            (UnMHTStringBundle.get ("filetype")
             + " " + UnMHTStringBundle.get ("filetype_original"),
             "*.mht");
          }
          else {
            filePicker.appendFilter
            (UnMHTStringBundle.get ("filetype")
             + " " + UnMHTStringBundle.get ("filetype_original_add"),
             "*.mht");
          }
        }
      }
      else {
        let showCurrent = true;
        let showOriginal = true;
    
        if (scrapBookData) {
          showCurrent = false;
        }
        if (UnMHTAppInfo.isThunderbird ()) {
          showOriginal = false;
        }
        if (targetSelection) {
          showOriginal = false;
        }
    
        let index = 0;
        if (showCurrent) {
          filePicker.appendFilter
            (UnMHTStringBundle.get ("filetype")
             + " " + UnMHTStringBundle.get ("filetype_current"),
             "*.mht");
          currentIndex = index;
          index ++;
        }
        else {
          currentIndex = -1;
        }
        if (showOriginal) {
          filePicker.appendFilter
          (UnMHTStringBundle.get ("filetype")
           + " " + UnMHTStringBundle.get ("filetype_original"),
           "*.mht");
          originalIndex = index;
          index ++;
        }
        else {
          originalIndex = -2;
        }
        if (showCurrent) {
          filePicker.appendFilter
          (UnMHTStringBundle.get ("filetype")
           + " " + UnMHTStringBundle.get ("filetype_current_add"),
           "*.mht");
          currentAddIndex = index;
          index ++;
        }
        else {
          currentAddIndex = -1;
        }
        if (showOriginal) {
          filePicker.appendFilter
          (UnMHTStringBundle.get ("filetype")
           + " " + UnMHTStringBundle.get ("filetype_original_add"),
           "*.mht");
          originalAddIndex = index;
          index ++;
        }
        else {
          originalAddIndex = -2;
        }
          
        let lastType = UnMHTPref.getPref ("save.single.type");
          
        if (lastType == "current") {
          filePicker.filterIndex = currentIndex;
        }
        else if (lastType == "original") {
          filePicker.filterIndex = originalIndex;
        }
        else if (lastType == "current_add") {
          filePicker.filterIndex = currentAddIndex;
        }
        else if (lastType == "original_add") {
          filePicker.filterIndex = originalAddIndex;
        }
      }
        
      filePicker.defaultString = saveParam.targetFilename;
    }.bind (this);
    
    let uri = null;
    if (targetDocument) {
      uri = targetDocument.location.href;
    }
    let targetFile
    = UnMHTFilePicker.open (window, title, mode, true, "",
                            modFunc, uri);
    
    if (targetFile) {
      if (!("useCurrentState" in saveParam
            && "additionalFiles" in saveParam)) {
        if (currentFilePicker.filterIndex == currentIndex) {
          saveParam.useCurrentState = true;
          saveParam.additionalFiles = null;
          UnMHTPref.setPref ("save.single.type", "current");
        }
        else if (currentFilePicker.filterIndex == originalIndex) {
          saveParam.useCurrentState = false;
          saveParam.additionalFiles = null;
          UnMHTPref.setPref ("save.single.type", "original");
        }
        else if (currentFilePicker.filterIndex == currentAddIndex) {
          saveParam.useCurrentState = true;
          showAdditionalFilesDialog = true;
          UnMHTPref.setPref ("save.single.type", "current_add");
        }
        else if (currentFilePicker.filterIndex == originalAddIndex) {
          saveParam.useCurrentState = false;
          showAdditionalFilesDialog = true;
          UnMHTPref.setPref ("save.single.type", "original_add");
        }
      }
      
      targetFile.leafName
        = UnMHTFileUtil.validateFileName (targetFile.leafName);
      if (!targetFile.leafName.match (/\.mht(ml)?$/i)) {
        /* 拡張子 .mht が無いと UnMHT でハンドルできなくなるので追加する */
        let filenameBase = targetFile.leafName;
        let path = targetFile.path + ".mht";
        targetFile = UnMHTFileUtil.initFile (path);
        
        if (targetFile.exists ()) {
          /* .mht を追加した事でファイル名が被った */
          let base = targetFile.parent.path;
          let filename = targetFile.leafName;
          
          let mtime = new Date ();
          mtime.setTime (targetFile.lastModifiedTime);
          let result = {
            result : 0,
            applyToAll: false,
            rename : false,
            filename : ""
          };
          
          let confirmURL = "chrome://unmht/content/confirm.xul";
          let features = "chrome,dialog,modal,centerscreen";
          let param = {
            result: result,
            title: UnMHTStringBundle.get ("overwrite_title"),
            message1: UnMHTStringBundle.get ("overwrite_pre")
            + " " + filename + " "
            + UnMHTStringBundle.get ("overwrite_post"),
            message2: UnMHTStringBundle.get ("overwrite_lastaccess")
            + " " + mtime,
            message3: UnMHTStringBundle.get ("overwrite_post2"),
            accept: UnMHTStringBundle.get ("replace"),
            extra2: UnMHTStringBundle.get ("abort"),
            rename: UnMHTStringBundle.get ("rename"),
            toall: null,
            filename: UnMHTSave.getUniqueFilename (base, filenameBase),
            base: base,
            exist1: UnMHTStringBundle.get ("overwrite_pre"),
            exist2: UnMHTStringBundle.get ("overwrite_post")
          };
          window.openDialog (confirmURL, "", features, param);
          
          if (result.result == "abort" || result.result == "no") {
            return;
          }
          
          if (result.result == "rename") {
            filename = result.filename;
              
            targetFile = UnMHTFileUtil.initFile (base);
            targetFile.appendRelativePath (filename);
          }
        }
      }
      
      if (showAdditionalFilesDialog) {
        this.showAdditionalFilesDialogAndSave (window,
                                               targetDocument, targetSelection,
                                               scrapBookData, saveParam,
                                               targetFile, callback);
      }
      else {
        this.startSave (window, targetDocument, targetSelection,
                        scrapBookData,
                        saveParam.useCurrentState, saveParam.additionalFiles,
                        targetFile, callback);
      }
    }
  },
  
  /**
   * 追加ファイルのダイアログを表示してから保存する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Object saveParam
   *         追加パラメータ
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         終了時コールバック
   */
  showAdditionalFilesDialogAndSave : function (window,
                                               targetDocument, targetSelection,
                                               scrapBookData, saveParam,
                                               targetFile, callback) {
    let dialogCallback
    = function (additionalFiles) {
      this.startSave (window, targetDocument, targetSelection,
                      scrapBookData,
                      saveParam.useCurrentState, additionalFiles,
                      targetFile, callback);
    }.bind (this);
    
    let pageURL = "chrome://unmht/content/additional.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let param = {
      callback : dialogCallback
    };
    window.openDialog (pageURL, "", features, param);
  },
  
  /**
   * ファイルを保存する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean useCurrentState
   *         現在の状態を使用するか
   * @param  Array additionalFiles
   *         追加ファイル
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         終了時コールバック
   */
  startSave : function (window, targetDocument, targetSelection,
                        scrapBookData,
                        useCurrentState, additionalFiles,
                        targetFile, callback) {
    if (UnMHTPref.getPref ("save.progress.statusbar")) {
      let listener = this.createListener (window);
      listener.callback = callback;
      
      UnMHTSave.setCurrentWindow (window);
      let saveContext = new UnMHTSaveContext ();
      
      let location = "";
      let subject = "";
      if (scrapBookData) {
        subject = scrapBookData ["title"];
        location = scrapBookData ["source"];
      }
      else {
        subject = targetDocument.title;
        location = targetDocument.location.href;
      }
      saveContext.init (targetFile, location, subject, listener);
      
      saveContext.addDocument
      (targetDocument, targetSelection,
       scrapBookData, useCurrentState,
       location, true, false);
      if (saveContext.isStopped) {
        return;
      }
      
      if (additionalFiles) {
        additionalFiles.forEach (function (file) {
            saveContext.addFile (null, file, false);
          });
      }
      
      saveContext.start ();
      
      this.contexts.push (saveContext);
    }
    else {
      let pageURL = "chrome://unmht/content/save.xul";
      let features = "chrome,titlebar,centerscreen,resizable";
      let param = {
        targetFile : targetFile,
        targetDocument : targetDocument,
        useCurrentState : useCurrentState,
        scrapBookData : scrapBookData,
        targetSelection : targetSelection,
        additionalFiles : additionalFiles,
        callback : callback
      };
      window.openDialog (pageURL, "", features, param);
    }
  },
  
  /**
   * 保存コンテキストを監視対象から削除
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   */
  removeContext : function (context) {
    this.contexts = this.contexts.filter (function (c) c != context);
    context.destruct ();
  },
  
  /**
   * 保存をキャンセルする
   *
   * @param  Window window
   *         実行中のウィンドウ
   */
  cancelSave : function (window) {
    this.contexts.forEach (function (context) {
        context.cancel ();
      });
    this.contexts = new Array ();
    
    this.updateProgress ();
  },
  
  /**
   * 保存可能かどうかを調べる
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @return Boolean
   *         保存可能かどうか
   */
  isSavable : function (window, targetDocument) {
    try {
      if ("gIeTab" in window
          && window.gIeTab.isIeEngine ()) {
        return false;
      }
    }
    catch (e) {
    }
    
    if (UnMHTAppInfo.isThunderbird ()) {
      if (targetDocument.location.href == "about:blank") {
        return false;
      }
      return true;
    }
    else if ("contentType" in targetDocument
             && targetDocument.contentType
             && (targetDocument.contentType.match (/^text/)
                 || targetDocument.contentType.match (/\/x(ht)?ml(\+xml)?$/))) {
      return true;
    }
    else {
      return false;
    }
  }
};
UnMHTSaveManager.init ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTSaveManager.term ();
  });
