/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTSaveManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTScrapBookIntegration"];

/**
 * ScrapBook サイトバー組み込み
 *
 * @param  Document sidebarDocument
 *         対象のサイドバーのドキュメント
 */
function UnMHTScrapBookSidebarIntegration (sidebarDocument) {
  this.sidebarDocument = sidebarDocument;
  
  this.onPopupShowing = this.onPopupShowing.bind (this);
  this.onSave = this.onSave.bind (this);
  this.onUnload = this.onUnload.bind (this);
  
  this.sidebarDocument.addEventListener ("unload", this.onUnload, false);
  
  this.createMenuItem ();
}
UnMHTScrapBookSidebarIntegration.prototype = {
  sidebarDocument : null, /* Document  対象のサイドバーのドキュメント */
  popup : null,           /* XULElement  ScrapBook のポップアップ */
  menuitem : null,        /* XULElement  追加したメニューアイテム */
  scrapBookData : null,   /* Object  ScrapBook のデータ */
  
  parent : null,          /* UnMHTScrapBookIntegration  ScrapBook 組み込み */
  
  /**
   * 破棄
   */
  destruct : function () {
    if (this.popup) {
      try {
        this.popup.removeEventListener ("popupshowing", this.onPopupShowing);
      }
      catch (e) {
      }
      this.popup = null;
    }
    
    if (this.menuitem) {
      try {
        this.menuitem.removeEventListener ("command", this.onSave);
        this.menuitem.parentNode.removeChild (this.menuitem);
      }
      catch (e) {
      }
      this.menuitem = null;
    }
    
    if (this.sidebarDocument) {
      try {
        this.sidebarDocument.removeEventListener ("unload", this.onUnload);
      }
      catch (e) {
      }
      this.sidebarDocument = null;
    }
    
    this.scrapBookData = null;
    this.parent = null;
  },
  
  /**
   * メニューアイテムを作成する
   */
  createMenuItem : function () {
    let popup = this.sidebarDocument.getElementById ("sbPopup");
    let pos = this.sidebarDocument.getElementById ("sbPopupTools");
    
    if (popup && pos) {
      let item = this.sidebarDocument.createElement ("menuitem");
      item.id = "unmht-scrapbook-save";
      item.className = "menuitem-iconic";
      item.setAttribute ("label", UnMHTStringBundle.get ("save_title"));
      item.setAttribute ("image", "resource://unmht/res/icon_small.png");
      item.addEventListener ("command", this.onSave, false);
      pos.parentNode.insertBefore (item, pos);
      
      popup.addEventListener ("popupshowing", this.onPopupShowing, false);
      
      this.popup = popup;
      this.menuitem = item;
    }
  },
  
  /**
   * ScrapBook のポップアップを設定する
   *
   * @param  Event event
   *         対象のイベント
   */
  onPopupShowing : function (event) {
    let sidebarWindow = this.sidebarDocument.defaultView;
    
    let sbData;
    let sbUtils;
    let sbResource;
    if ("ScrapBookData" in sidebarWindow) {
      sbData = sidebarWindow.ScrapBookData;
    }
    else if ("sbDataSource" in sidebarWindow) {
      sbData = sidebarWindow.sbDataSource;
    }
    else {
      return;
    }
    if ("ScrapBookUtils" in sidebarWindow) {
      sbUtils = sidebarWindow.ScrapBookUtils;
    }
    else if ("sbCommonUtils" in sidebarWindow) {
      sbUtils = sidebarWindow.sbCommonUtils;
    }
    else {
      return;
    }
    if ("sbTreeUI" in sidebarWindow) {
      sbResource = sidebarWindow.sbTreeUI.resource;
    }
    else if ("sbTreeHandler" in sidebarWindow) {
      sbResource = sidebarWindow.sbTreeHandler.resource;
    }
    else if ("sbListHandler" in sidebarWindow) {
      sbResource = sidebarWindow.sbListHandler.resource;
    }
    else {
      return;
    }
    
    let scrapBookData = {};
    scrapBookData ["source"]  = sbData.getProperty (sbResource, "source");
    scrapBookData ["title"]   = sbData.getProperty (sbResource, "title");
    let id                    = sbData.getProperty (sbResource, "id");
    scrapBookData ["id"]      = id;
    scrapBookData ["dir"]     = sbUtils.getContentDir (id);
    scrapBookData ["charset"] = sbData.getProperty (sbResource, "chars");
    
    if (scrapBookData ["source"]) {
      this.menuitem.hidden = false;
    }
    else {
      this.menuitem.hidden = true;
    }
    
    this.scrapBookData = scrapBookData;
  },
  
  /**
   * 保存
   */
  onSave : function () {
    if (this.scrapBookData) {
      let window = this.sidebarDocument.defaultView;
      if (window.parent) {
        window = window.parent;
      }
      UnMHTSaveManager.saveScrapBook (window, this.scrapBookData);
    }
  },
  
  /**
   * ドキュメントのアンロードイベント
   */
  onUnload : function () {
    this.parent.removeIntegration (this);
  }
};

/**
 * ScrapBook 組み込み
 *
 * @param  Window window
 *         対象のウィンドウ
 */
function UnMHTScrapBookIntegration (window) {
  this.window = window;
  
  this.integs = [];
  this.onLoadSidebar = this.onLoadSidebar.bind (this);
}
UnMHTScrapBookIntegration.prototype = {
  integs : null, /* Array  組み込みリスト
                  *   [UnMHTScrapBookSidebarIntegration, ...] */
  
  window : null,
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
  },
  
  /**
   * 組み込む
   */
  onLoad : function () {
    let sidebar = this.window.document.getElementById ("sidebar");
    if (sidebar) {
      sidebar.addEventListener
        ("DOMContentLoaded", this.onLoadSidebar, false);
    }
  },
  
  /**
   * 組み込みを解除する
   */
  onUnload : function () {
    let sidebar = this.window.document.getElementById ("sidebar");
    if (sidebar) {
      sidebar.removeEventListener ("DOMContentLoaded", this.onLoadSidebar);
    }
    
    this.integs
    .forEach (function (integ) {
        integ.destruct ();
      });
    this.integs = [];
    
    this.integs = null;
  },
  
  /**
   * サイドバーの内容が ScrapBook かどうかチェック
   * ScrapBook ならばメニューの項目を追加する
   *
   * @param  Event event
   *         対象のイベント
   */
  onLoadSidebar : function (event) {
    let sidebarDocument = event.target.defaultView.document;
    if (sidebarDocument.documentURI
        != "chrome://scrapbook/content/scrapbook.xul") {
      return;
    }
    
    let integ = new UnMHTScrapBookSidebarIntegration (sidebarDocument);
    integ.parent = this;
    this.integs.push (integ);
  },
  
  /**
   * 組み込みを解除する
   *
   * @param  UnMHTScrapBookIntegration integration
   *         解除する組み込み
   */
  removeIntegration : function (integration) {
    this.integs
    = this.integs.filter
    (function (integ) {
      if (integ == integration) {
        integ.destruct ();
        return false;
      }
      return true;
    });
  }
};
