/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTUI.jsm");
Cu.import ("resource://unmht/modules/UnMHTPageInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTScrapBookIntegration.jsm");
Cu.import ("resource://unmht/modules/UnMHTMAFIntegration.jsm");
Cu.import ("resource://unmht/modules/UnMHTOverlayLoader.jsm");

let EXPORTED_SYMBOLS = ["UnMHTStartupWindowHook"];

/**
 * ウィンドウイベントリスナ
 *
 * @implements nsIWindowMediatorListener
 */
let UnMHTWindowListener = {
  QueryInterface:    XPCOMUtils.generateQI ([Ci.nsIWindowMediatorListener]),
  
  /* ==== nsIWindowMediatorListener ==== */
  
  /**
   * ウィンドウを開いたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onOpenWindow: function (window) {
    window
    = window.QueryInterface
    (Ci.nsIInterfaceRequestor).getInterface
    (Ci.nsIDOMWindowInternal || Ci.nsIDOMWindow);
    
    let loadFunc = function () {
      window.removeEventListener ("load", loadFunc, false);
      
      UnMHTStartupWindowHook.onLoad (window);
    };
    
    window.addEventListener ("load", loadFunc, false);
  },
  
  /**
   * ウィンドウを閉じたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onCloseWindow: function (window) {
    UnMHTStartupWindowHook.onUnload (window);
  },
  
  /**
   * ウィンドウを開いたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  String title
   *         タイトル
   */
  onWindowTitleChange: function (window, title) {
  }
};

let browserHooks = [
  UnMHTUIClass,
  UnMHTExtractHook,
  UnMHTScrapBookIntegration,
  UnMHTMAFIntegration
  ];

let mailHooks = [
  UnMHTUIClass,
  UnMHTExtractHook,
  UnMHTMAFIntegration
  ];

let overlays = {
  "chrome://browser/content/browser.xul" : {
    uri : "chrome://unmht/content/ui-browser.xul",
    hooks : browserHooks
  },
  "chrome://navigator/content/navigator.xul" : {
    uri : "chrome://unmht/content/ui-browser.xul",
    hooks : browserHooks
  },
  "chrome://messenger/content/messenger.xul" : {
    uri : "chrome://unmht/content/ui-messenger.xul",
    hooks : mailHooks
  },
  "chrome://browser/content/pageinfo/pageInfo.xul" : {
    uri : "",
    hooks : [
      UnMHTPageInfo
      ]
  }
};

/**
 * ウィンドウへの組み込み
 */
let UnMHTStartupWindowHook = {
  registered : [], /* 組み込みリスト
                    *   [[nsIXULWindow, Object], ...] */
  
  /**
   * スタートアップ時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onStartup : function (data, reason) {
    for (let window in XPCOMUtils.IterSimpleEnumerator
           (Services.wm.getEnumerator (null), Ci.nsIDOMWindow)) {
      this.onLoad (window);
    }
    
    Services.wm.addListener (UnMHTWindowListener);
  },
  
  /**
   * シャットダウン時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onShutdown : function (data, reason) {
    let dialogs = [
      "chrome://unmht/content/pref/pref.xul",
      "chrome://unmht/content/additional.xul",
      "chrome://unmht/content/confirm.xul",
      "chrome://unmht/content/nglist.xul",
      "chrome://unmht/content/save.xul",
      "chrome://unmht/content/savemulti.xul",
      "chrome://unmht/content/saveoptions.xul"
      ];
    
    Services.wm.removeListener (UnMHTWindowListener);
    
    for (let window in XPCOMUtils.IterSimpleEnumerator
           (Services.wm.getEnumerator (null), Ci.nsIDOMWindow)) {
      this.onUnload (window);
      if (dialogs.indexOf (window.document.documentURI) != -1) {
        window.close ();
      }
    }
  },
  
  /**
   * ウィンドウがロードされたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onLoad : function (window) {
    let location = window.document.location.href;
    if (location in overlays) {
      let callback = function () {
        let list
        = overlays [location].hooks.map (function (n) new n (window));
          
        this.registered.push ([window, list]);
          
        list.forEach (function (item) item.onLoad ());
      }.bind (this);
      
      if (overlays [location].uri) {
        UnMHTOverlayLoader.loadInto
          (window, overlays [location].uri, callback);
      }
      else {
        callback ();
      }
    }
  },
  
  /**
   * ウィンドウがアンロードされたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onUnload : function (window) {
    if (window.document) {
      let location = window.document.location.href;
      if (location in overlays) {
        UnMHTOverlayLoader.unloadFrom (window, overlays [location].uri);
      }
    }
    
    this.registered
    = this.registered.filter (function ([w, list]) {
        if (w == window) {
          list.forEach (function (item) {
              item.onUnload ();
              item.destruct ();
            });
          return false;
        }
        return true;
      });
  }
};
